/*	$Id: _context_u.S,v 1.5 2003/06/13 15:25:01 tsubai Exp $	*/

/*
 * Copyright (c) 2001 Wasabi Systems, Inc.
 * All rights reserved.
 *
 * Written by Allen Briggs for Wasabi Systems, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed for the NetBSD Project by
 *      Wasabi Systems, Inc.
 * 4. The name of Wasabi Systems, Inc. may not be used to endorse
 *    or promote products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY WASABI SYSTEMS, INC. ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL WASABI SYSTEMS, INC
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <machine/asm.h>
#include "assym.h"

/*
 * Define:
 *	int _getcontext_u(ucontext_t *ctx)
 *		Store the current context in the provided ctx structure.
 *		[only store the callee-saved registers]
 *	int _setcontext_u(const ucontext_t *ctx)
 *		Restore the current context from the provided ctx structure.
 *	int _swapcontext_u(ucontext_t *from_ctx, const ucontext_t *to_ctx)
 *		First, store the current context into from_ctx and then
 *		restore the current context from the to_ctx.
 */

/*
 * According to the SVR4 ABI for PPC, r1, r14-r31, LR, and portions of the
 * CR are non-volatile.  I.e., they need to be preserved across the function
 * call.
 *
 * If _getcontext_u is called, set the UC_USER_BIT in the UC_FLAGS.  If
 * _setcontext_u sees that bit, then it will just restore this part of the
 * context, otherwise, it will call through to setcontext(2).
 *
 * Note: 0xd032 (put in MSR) is a magic constant that keeps the kernel
 * setcontext(2) happy as it compares this value to srr1 to see if it
 * should return EINVAL.
 */
#define GETC(reg)					  \
	stw	1,(_REG_R0 + 1*4)(reg)			; \
	stmw	14,(_REG_R0 + 14*4)(reg)		; \
	mfcr	0					; \
	stw	0,_REG_CR(reg)				; \
	mflr	0					; \
	stw	0,_REG_LR(reg)				; \
	stw	0,_REG_PC(reg)				; \
							; \
	mffs	0					; \
	stfd	0,(_REG_FPSCR - 4)(reg)			; \
	stfd	14,(_REG_FP0 + 14*8)(reg)		; \
	stfd	15,(_REG_FP0 + 15*8)(reg)		; \
	stfd	16,(_REG_FP0 + 16*8)(reg)		; \
	stfd	17,(_REG_FP0 + 17*8)(reg)		; \
	stfd	18,(_REG_FP0 + 18*8)(reg)		; \
	stfd	19,(_REG_FP0 + 19*8)(reg)		; \
	stfd	20,(_REG_FP0 + 20*8)(reg)		; \
	stfd	21,(_REG_FP0 + 21*8)(reg)		; \
	stfd	22,(_REG_FP0 + 22*8)(reg)		; \
	stfd	23,(_REG_FP0 + 23*8)(reg)		; \
	stfd	24,(_REG_FP0 + 24*8)(reg)		; \
	stfd	25,(_REG_FP0 + 25*8)(reg)		; \
	stfd	26,(_REG_FP0 + 26*8)(reg)		; \
	stfd	27,(_REG_FP0 + 27*8)(reg)		; \
	stfd	28,(_REG_FP0 + 28*8)(reg)		; \
	stfd	29,(_REG_FP0 + 29*8)(reg)		; \
	stfd	30,(_REG_FP0 + 30*8)(reg)		; \
	stfd	31,(_REG_FP0 + 31*8)(reg)		; \
							; \
	li	8,1					; \
	stw	8,(_UC_FPVALID)(reg)			; \
	slwi	9,8,_UC_USER_BIT			; \
	ori	9,9,(_UC_CPU | _UC_FPU)			; \
	stw	9,(UC_FLAGS)(reg)

#define SETC						  \
	li	9,1					; \
	slwi	9,9,_UC_USER_BIT			; \
	lwz	3,(UC_FLAGS)(4)				; \
	and.	9,3,9					; \
	beq	2f					; \
							; \
	lwz	1,(_REG_R0 + 1*4)(4)			; \
	lmw	14,(_REG_R0 + 14*4)(4)			; \
	lwz	0,_REG_LR(4)				; \
	mtlr	0					; \
	lwz	0,_REG_PC(4)				; \
	mtctr	0					; \
							; \
	andi.	5,3,_UC_FPU				; \
	beq	1f					; \
							; \
	lwz	5,(_UC_FPVALID)(4)			; \
	cmpwi	5,0					; \
	beq	1f					; \
							; \
	lfd	14,(_REG_FP0 + 14*8)(4)			; \
	lfd	15,(_REG_FP0 + 15*8)(4)			; \
	lfd	16,(_REG_FP0 + 16*8)(4)			; \
	lfd	17,(_REG_FP0 + 17*8)(4)			; \
	lfd	18,(_REG_FP0 + 18*8)(4)			; \
	lfd	19,(_REG_FP0 + 19*8)(4)			; \
	lfd	20,(_REG_FP0 + 20*8)(4)			; \
	lfd	21,(_REG_FP0 + 21*8)(4)			; \
	lfd	22,(_REG_FP0 + 22*8)(4)			; \
	lfd	23,(_REG_FP0 + 23*8)(4)			; \
	lfd	24,(_REG_FP0 + 24*8)(4)			; \
	lfd	25,(_REG_FP0 + 25*8)(4)			; \
	lfd	26,(_REG_FP0 + 26*8)(4)			; \
	lfd	27,(_REG_FP0 + 27*8)(4)			; \
	lfd	28,(_REG_FP0 + 28*8)(4)			; \
	lfd	29,(_REG_FP0 + 29*8)(4)			; \
	lfd	30,(_REG_FP0 + 30*8)(4)			; \
	lfd	31,(_REG_FP0 + 31*8)(4)			; \
	lfd	0,(_REG_FPSCR - 4)(4)			; \
	mtfsf	0xff,0					; \
							; \
1:							; \
	lwz	0,_REG_CR(4)				; \
	mtcr	0					; \
	bctr						; \
2:							; \
	mr	3,4					; \
	b	PIC_PLT(_C_LABEL(setcontext))		; \
	/* NOTREACHED */

ENTRY(_getcontext_u)
	GETC(3)
	li	3,0
	blr

ENTRY(_setcontext_u)
	mr	4,3
	SETC

ENTRY(_swapcontext_u)
	GETC(3)
	SETC
