/*	$Id: abtn.c,v 1.10 2003/04/01 09:57:54 tsubai Exp $	*/

/*-
 * Copyright (C) 1999, 2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/conf.h>
#include <sys/device.h>
#include <sys/proc.h>
#include <sys/systm.h>

#include <dev/wscons/wsconsio.h>
#include <dev/ofw/openfirm.h>

#include <macppc/dev/adbvar.h>
#include <macppc/dev/pm_direct.h>

int nvram_read_brightness(void);

#include "nvram.h"
#if NNVRAM == 0
#define nvram_read_brightness() (-1)
#endif

#define NVRAM_BRIGHTNESS 0x140e
#define ABTN_HANDLER_ID 31
#define WSDISPLAYDEV	35

#define ABTN_USE_PMU	0x01
#define ABTN_USE_OF	0x02

struct abtn_softc {
	struct device sc_dev;
	int sc_flags;

	int origaddr;		/* ADB device type */
	int adbaddr;		/* current ADB address */
	int handler_id;

	int brightness;		/* backlight brightness */
	int volume;		/* speaker volume (not yet) */
};

int abtn_match __P((struct device *, struct cfdata *, void *));
void abtn_attach __P((struct device *, struct device *, void *));
void abtn_adbcomplete __P((caddr_t, caddr_t, int));
void abtn_call_pmu __P((struct abtn_softc *, u_int));
void abtn_call_OF __P((struct abtn_softc *, u_int));

struct cfattach abtn_ca = {
	"abtn", {}, sizeof(struct abtn_softc), abtn_match, abtn_attach
};

int
abtn_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct adb_attach_args *aa = aux;

	if (aa->origaddr == ADBADDR_MISC &&
	    aa->handler_id == ABTN_HANDLER_ID)
		return 1;

	return 0;
}

void
abtn_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct abtn_softc *sc = (struct abtn_softc *)self;
	struct adb_attach_args *aa = aux;
	ADBSetInfoBlock adbinfo;
	int bright;

	printf("brightness/volume button\n");

	/*
	 * Don't use pmu call on Uni-North based models.
	 */
	if (OF_finddevice("/uni-n") == -1) {
		sc->sc_flags |= ABTN_USE_PMU;
		bright = pm_read_nvram(NVRAM_BRIGHTNESS);
		if (bright != 0)
			pm_set_brightness(bright);
	} else {
		sc->sc_flags |= ABTN_USE_OF;
		bright = nvram_read_brightness();
		switch (bright) {
		case 0xff:	/* backlight off */
			bright = 0x20;
			break;
		case -1:	/* error */
			bright = 0xff;
			break;
		default:	/* normal (1-31) */
			bright <<= 3;
		}
	}

	sc->brightness = bright;

	sc->origaddr = aa->origaddr;
	sc->adbaddr = aa->adbaddr;
	sc->handler_id = aa->handler_id;

	adbinfo.siServiceRtPtr = abtn_adbcomplete;
	adbinfo.siDataAreaAddr = sc;

	SetADBInfo(&adbinfo, sc->adbaddr);
}

void
abtn_adbcomplete(buffer, data, adb_command)
	caddr_t buffer, data;
	int adb_command;
{
	struct abtn_softc *sc = (struct abtn_softc *)data;
	u_int cmd;

	cmd = buffer[1];

	if (sc->sc_flags & ABTN_USE_PMU)
		abtn_call_pmu(sc, cmd);

	if (sc->sc_flags & ABTN_USE_OF)
		abtn_call_OF(sc, cmd);
}

void
abtn_call_pmu(sc, cmd)
	struct abtn_softc *sc;
	u_int cmd;
{
	switch (cmd) {
	case 0x0a:
		sc->brightness -= 8;
		if (sc->brightness < 8)
			sc->brightness = 8;
		pm_set_brightness(sc->brightness);
		pm_write_nvram(NVRAM_BRIGHTNESS, sc->brightness);
		break;

	case 0x09:
		sc->brightness += 8;
		if (sc->brightness > 0x78)
			sc->brightness = 0x78;
		pm_set_brightness(sc->brightness);
		pm_write_nvram(NVRAM_BRIGHTNESS, sc->brightness);
		break;
	}
}

/*
 * XXX HACK ALERT XXX
 * XXX The below is very dirty hack.  But better than nothing, I think.
 */

#include "wsdisplay.h"
dev_type_ioctl(wsdisplayioctl);

#if NWSDISPLAY == 0
#define wsdisplayioctl(dev, cmd, data, flag, p)
#endif

void
abtn_call_OF(sc, cmd)
	struct abtn_softc *sc;
	u_int cmd;
{
	struct wsdisplay_param wsp;
	dev_t dev = makedev(WSDISPLAYDEV, 0);		/* XXX */

	switch (cmd) {
	case 0x0a:
		sc->brightness -= 0x10;
		break;

	case 0x09:
		sc->brightness += 0x10;
		break;

	default:
		return;
	}

	if (sc->brightness < 0x10)
		sc->brightness = 0x10;
	if (sc->brightness > 0xf0)
		sc->brightness = 0xf0;
	wsp.curval = sc->brightness;
	wsdisplayioctl(dev, WSDISPLAYIO_SETPARAM, (caddr_t)&wsp, 0, curproc);
}
