/*	$Id: battery.c,v 1.11 2003/01/24 17:50:02 tsubai Exp $	*/

/*-
 * Copyright (c) 2001 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/conf.h>
#include <sys/device.h>
#include <sys/systm.h>

#include <machine/autoconf.h>
#include <macppc/dev/pm_direct.h>

int battery_match(struct device *, struct cfdata *, void *);
void battery_attach(struct device *, struct device *, void *);

struct cfattach battery_ca = {
	"battery", {}, sizeof(struct device), battery_match, battery_attach
};

dev_type_read(batteryread);
dev_type_write(batterywrite);
dev_type_mmap(batterymmap);

const struct cdevsw battery_cdevsw = {
	nullopen, nullclose, batteryread, batterywrite, noioctl,
	nostop, notty, nopoll, nommap, nokqfilter
};

int
battery_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;

	if (strcmp(ca->ca_name, "battery") == 0)
		return 1;

	return 0;
}

void
battery_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{

	printf("\n");
}

int
batteryread(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	int ac;
	u_int len, off;
	char buf[11];	/* "1 100 100\n" */

	off = uio->uio_offset;

	if (off >= sizeof buf)
		return 0;

	len = uio->uio_resid;
	if (off + len > sizeof buf)
		len = sizeof buf - off;

	ac = (pm_read_switch() & 0x04) >> 2;

	bzero(buf, sizeof buf);
	snprintf(buf, sizeof buf, "%d %d %d\n",
	    ac, pm_battery(1), pm_battery(2));
	return uiomove(buf + off, len, uio);
}

#include <sys/callout.h>

void change_clock(void *);

int
batterywrite(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	int error, s;
	char buf[32];
	static struct callout c;

	if (uio->uio_offset != 0)
		return ENXIO;
	bzero(buf, sizeof buf);
	error = uiomove(buf, sizeof buf - 1, uio);
	if (error)
		return error;

	/*
	 * XXX ߥå (pa == va) ܤ callout
	 *     ͳ롣
	 */
	if (strncmp(buf, "power0", 6) == 0) {
		s = splhigh();
		callout_init(&c);
		callout_reset(&c, 1, change_clock, (void *)0);
		splx(s);
	}
	if (strncmp(buf, "power1", 6) == 0) {
		s = splhigh();
		callout_init(&c);
		callout_reset(&c, 1, change_clock, (void *)1);
		splx(s);
	}
	if (strncmp(buf, "suspend", 7) == 0) {
		s = splhigh();
		callout_init(&c);
		callout_reset(&c, 1, change_clock, (void *)-1);
		splx(s);
	}

	return 0;
}
