/*	$Id: daca.c,v 1.2 2003/08/11 17:39:24 tsubai Exp $	*/

/*-
 * Copyright (c) 2002,2003 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Datasheet is available from
 * http://www.micronas.com/products/documentation/multimedia/dac3550a/
 */

#include <sys/param.h>
#include <sys/audioio.h>
#include <sys/device.h>
#include <sys/systm.h>

#include <dev/audio_if.h>
#include <dev/ofw/openfirm.h>

#include <machine/autoconf.h>
#include <machine/dbdma.h>
#include <machine/uni-n.h>

#include <macppc/dev/i2svar.h>

#ifdef DACA_DEBUG
# define DPRINTF printf
#else
# define DPRINTF while (0) printf
#endif

#define daca_softc i2s_softc		/* XXX */

int daca_match(struct device *, struct cfdata *, void *);
void daca_attach(struct device *, struct device *, void *);

/* XXX */
int daca_getdev(void *, struct audio_device *);

void daca_set_volume(struct daca_softc *, int, int);

struct cfattach daca_ca = {
	"daca", {}, sizeof(struct daca_softc),
	daca_match, daca_attach
};

struct audio_hw_if daca_hw_if = {
	i2s_open,
	i2s_close,
	NULL,
	i2s_query_encoding,
	i2s_set_params,
	i2s_round_blocksize,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	i2s_halt_output,
	i2s_halt_input,
	NULL,
	daca_getdev,
	NULL,
	i2s_set_port,
	i2s_get_port,
	i2s_query_devinfo,
	NULL,
	NULL,
	i2s_round_buffersize,
	i2s_mappage,
	i2s_get_props,
	i2s_trigger_output,
	i2s_trigger_input,
	NULL
};

struct audio_device daca_device = {
	"DACA",
	"",
	"daca"
};

/* DAC3550 registers */
#define DEQ_SR		0x01	/* Sample rate control (8) */
#define DEQ_AVOL	0x02	/* Analog volume (16) */
#define DEQ_GCFG	0x03	/* Global configuration (8) */

int
daca_match(parent, match, aux)
	struct device *parent;
	struct cfdata *match;
	void *aux;
{
	struct confargs *ca = aux;
	int soundbus, soundchip;
	char compat[32];

	if (strcmp(ca->ca_name, "i2s") != 0)
		return 0;

	if ((soundbus = OF_child(ca->ca_node)) == 0 ||
	    (soundchip = OF_child(soundbus)) == 0)
		return 0;

	bzero(compat, sizeof compat);
	OF_getprop(soundchip, "compatible", compat, sizeof compat);

	if (strcmp(compat, "daca") != 0)
		return 0;

	return 1;
}

#define DEQaddr 0x9a

void
daca_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct daca_softc *sc = (struct daca_softc *)self;
	int ki2c;

	/* "set volume" callback */
	sc->sc_setvolume = daca_set_volume;

	ki2c = OF_finddevice("ki2c");
	if (ki2c == -1) {
		printf(": unable to find ki2c\n");
		return;
	}
	sc->sc_i2c = ki2c_open(ki2c, DEQaddr, I2C_STDSUBMODE);
	if (sc->sc_i2c == NULL) {
		printf(": unable to attach deq\n");
		return;
	}

	i2s_attach(sc, aux);

#if 0
	/* Enable I2C interrupts. */
	ki2c_writereg(sc->sc_i2c, IER,I2C_INT_DATA|I2C_INT_ADDR|I2C_INT_STOP);
#endif

	daca_set_volume(sc, 80, 80);

	audio_attach_mi(&daca_hw_if, sc, self);
}

int
daca_getdev(h, retp)
	void *h;
	struct audio_device *retp;
{
	*retp = daca_device;
	return 0;
}

void
daca_set_volume(sc, left, right)
	struct daca_softc *sc;
	int left, right;
{
	u_int16_t data;

	sc->sc_vol_l = left;
	sc->sc_vol_r = right;

	left >>= 2;
	right >>= 2;
	data = left << 8 | right;
	ki2c_writesub(sc->sc_i2c, DEQ_AVOL, &data, 2);
}
