/*	$Id: dbdma.c,v 1.6 2002/11/05 14:00:38 tsubai Exp $	*/

/*-
 * Copyright (c) 2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <uvm/uvm_extern.h>

#include <machine/dbdma.h>

#define SET(x)  ((x) | (x) << 16)
#define CLR(x)  ((x) << 16)
#define eieio() asm volatile ("eieio")

void
dbdma_start(d, cmdp)
	struct dbdma_reg *d;
	struct dbdma_desc *cmdp;
{
	paddr_t addr;

	addr = vtophys((vaddr_t)cmdp);
	if (addr & 0xf)
		panic("dbdma_start");

	d->dr_control =
	    htole32(CLR((DBDMA_WAKE|DBDMA_FLUSH|DBDMA_PAUSE|DBDMA_RUN)));
	eieio();

	while (d->dr_status & htole32(DBDMA_ACTIVE));
	eieio();

	d->dr_cmdptrhi = 0;
	dbdma_st32(&d->dr_cmdptrlo, addr);
	eieio();

	d->dr_control = htole32(SET(DBDMA_RUN));
	eieio();
}

void
dbdma_stop(d)
	struct dbdma_reg *d;
{
	d->dr_control = htole32(CLR(DBDMA_RUN) | SET(DBDMA_FLUSH));
	eieio();
	while (d->dr_status & htole32(DBDMA_ACTIVE|DBDMA_FLUSH));
	eieio();
}

void
dbdma_pause(d)
	struct dbdma_reg *d;
{
	d->dr_control = htole32(SET(DBDMA_PAUSE));
	eieio();
	while (d->dr_status & htole32(DBDMA_ACTIVE));
	eieio();
}

void
dbdma_flush(d)
	struct dbdma_reg *d;
{
	d->dr_control = htole32(SET(DBDMA_FLUSH));
	eieio();
	while (d->dr_status & htole32(DBDMA_FLUSH));
	eieio();
}

void
dbdma_continue(d)
	struct dbdma_reg *d;
{
#if 1
	d->dr_control =
	    htole32(SET(DBDMA_RUN|DBDMA_WAKE) | CLR(DBDMA_PAUSE));
#else
	d->dr_control = htole32(SET(DBDMA_WAKE));
#endif
	eieio();
}

void
dbdma_reset(d)
	struct dbdma_reg *d;
{
	d->dr_control =
	    htole32(CLR((DBDMA_WAKE|DBDMA_FLUSH|DBDMA_PAUSE|DBDMA_RUN)));
	eieio();
	while (d->dr_status & htole32(DBDMA_RUN));
	eieio();
}
