/*	$Id: if_bm.c,v 1.29 2002/11/15 15:30:29 tsubai Exp $	*/

/*-
 * Copyright (C) 1998-2000,2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "bpfilter.h"

#include <sys/param.h>
#include <sys/device.h>
#include <sys/ioctl.h>
#include <sys/kernel.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/systm.h>
#include <sys/callout.h>

#include <uvm/uvm_extern.h>

#include <net/if.h>
#include <net/if_ether.h>
#include <net/if_media.h>

#if NBPFILTER > 0
#include <net/bpf.h>
#endif

#include <dev/ofw/openfirm.h>

#include <dev/mii/mii.h>
#include <dev/mii/miivar.h>
#include <dev/mii/mii_bitbang.h>

#include <machine/autoconf.h>
#include <machine/dbdma.h>
#include <machine/pio.h>

#include <macppc/dev/if_bmreg.h>

#define BMAC_TXBUFS 8
#define BMAC_RXBUFS 16
#define BMAC_BUFLEN 2048

#define sc_if sc_ethercom.ec_if

struct bmac_softc {
	struct device sc_dev;
	struct ethercom sc_ethercom;
	int sc_flags;
	vaddr_t sc_regs;
	struct dbdma_reg *sc_txdma;
	struct dbdma_reg *sc_rxdma;
	struct dbdma_desc sc_txcmd[BMAC_TXBUFS * 2 + 1];
	struct dbdma_desc sc_rxcmd[BMAC_RXBUFS + 1];
	caddr_t sc_txbuf;
	struct mbuf *sc_rxbuf[BMAC_RXBUFS];
	int sc_txhead;
	int sc_rxlast;
	struct callout sc_tick_ch;
	struct mii_data sc_mii;
	u_char sc_enaddr[6];
};

#define BMAC_BMACPLUS	0x01
#define BMAC_DEBUGFLAG	0x02

extern u_int *heathrow_FCR;

static inline int bmac_read_reg __P((struct bmac_softc *, int));
static inline void bmac_write_reg __P((struct bmac_softc *, int, int));
static inline void bmac_set_bits __P((struct bmac_softc *, int, int));
static inline void bmac_reset_bits __P((struct bmac_softc *, int, int));

int bmac_match __P((struct device *, struct cfdata *, void *));
void bmac_attach __P((struct device *, struct device *, void *));
void bmac_reset_chip __P((struct bmac_softc *));
int bmac_init __P((struct ifnet *));
void bmac_init_dma __P((struct bmac_softc *));
int bmac_intr __P((void *));
int bmac_rint __P((void *));
void bmac_reset __P((struct bmac_softc *));
void bmac_stop __P((struct ifnet *, int));
void bmac_start __P((struct ifnet *));
void bmac_transmit_packet __P((struct bmac_softc *, void *, int));
int bmac_put __P((struct bmac_softc *, caddr_t, struct mbuf *));
struct mbuf *bmac_get __P((struct bmac_softc *));
void bmac_watchdog __P((struct ifnet *));
int bmac_ioctl __P((struct ifnet *, u_long, caddr_t));
int bmac_mediachange __P((struct ifnet *));
void bmac_mediastatus __P((struct ifnet *, struct ifmediareq *));
void bmac_setladrf __P((struct bmac_softc *));

int bmac_mii_readreg __P((struct device *, int, int));
void bmac_mii_writereg __P((struct device *, int, int, int));
void bmac_mii_statchg __P((struct device *));
void bmac_mii_tick __P((void *));
u_int32_t bmac_mbo_read __P((struct device *));
void bmac_mbo_write __P((struct device *, u_int32_t));

struct cfattach bm_ca = {
	"bm", {}, sizeof(struct bmac_softc), bmac_match, bmac_attach
};

struct mii_bitbang_ops bmac_mbo = {
	bmac_mbo_read, bmac_mbo_write,
	{ MIFDO, MIFDI, MIFDC, MIFDIR, 0 }
};

int
bmac_read_reg(sc, off)
	struct bmac_softc *sc;
	int off;
{
	return in16rb(sc->sc_regs + off);
}

void
bmac_write_reg(sc, off, val)
	struct bmac_softc *sc;
	int off, val;
{
	out16rb(sc->sc_regs + off, val);
}

void
bmac_set_bits(sc, off, val)
	struct bmac_softc *sc;
	int off, val;
{
	val |= bmac_read_reg(sc, off);
	bmac_write_reg(sc, off, val);
}

void
bmac_reset_bits(sc, off, val)
	struct bmac_softc *sc;
	int off, val;
{
	bmac_write_reg(sc, off, bmac_read_reg(sc, off) & ~val);
}

int
bmac_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;

	if (ca->ca_nreg < 24 || ca->ca_nintr < 12)
		return 0;

	if (strcmp(ca->ca_name, "bmac") == 0)
		return 1;
	if (strcmp(ca->ca_compat, "bmac+") == 0)
		return 1;

	return 0;
}

void
bmac_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct confargs *ca = aux;
	struct bmac_softc *sc = (void *)self;
	struct ifnet *ifp = &sc->sc_if;
	struct mii_data *mii = &sc->sc_mii;
	char *p;
	int i;
	u_char laddr[6];

#ifdef DIAGNOSTIC
	if ((vaddr_t)sc->sc_txdma & 0x0f) {
		printf(": bad alignment\n");
		return;
	}
#endif

	callout_init(&sc->sc_tick_ch);

	sc->sc_flags =0;
	if (strcmp(ca->ca_compat, "bmac+") == 0)
		sc->sc_flags |= BMAC_BMACPLUS;

	ca->ca_reg[0] += ca->ca_baseaddr;
	ca->ca_reg[2] += ca->ca_baseaddr;
	ca->ca_reg[4] += ca->ca_baseaddr;

	sc->sc_regs = (vaddr_t)mapiodev(ca->ca_reg[0], NBPG);

	bmac_write_reg(sc, INTDISABLE, NoEventsMask);

	if (OF_getprop(ca->ca_node, "local-mac-address", laddr, 6) == -1 &&
	    OF_getprop(ca->ca_node, "mac-address", laddr, 6) == -1) {
		printf(": cannot get mac-address\n");
		return;
	}
	bcopy(laddr, sc->sc_enaddr, 6);

	sc->sc_txdma = mapiodev(ca->ca_reg[2], NBPG);
	sc->sc_rxdma = mapiodev(ca->ca_reg[4], NBPG);
	p = malloc(BMAC_BUFLEN * BMAC_TXBUFS + NBPG - 1, M_DEVBUF, M_NOWAIT);
	if (p == NULL) {
		printf("cannot allocate memory\n");
		return;
	}
	p = (char *)roundup((vaddr_t)p, NBPG);
	sc->sc_txbuf = p;
	for (i = 0; i < BMAC_RXBUFS; i++)
		sc->sc_rxbuf[i] = bmac_get(sc);

	printf(" irq %d,%d: address %s\n", ca->ca_intr[0], ca->ca_intr[2],
	    ether_sprintf(laddr));

	intr_establish(ca->ca_intr[0], IST_LEVEL, IPL_NET, bmac_intr, sc);
	intr_establish(ca->ca_intr[2], IST_LEVEL, IPL_NET, bmac_rint, sc);

	bcopy(self->dv_xname, ifp->if_xname, IFNAMSIZ);
	ifp->if_softc = sc;
	ifp->if_init = bmac_init;
	ifp->if_ioctl = bmac_ioctl;
	ifp->if_start = bmac_start;
	ifp->if_stop = bmac_stop;
	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
	ifp->if_watchdog = bmac_watchdog;
	IFQ_SET_READY(&ifp->if_snd);

	mii->mii_ifp = ifp;
	mii->mii_readreg = bmac_mii_readreg;
	mii->mii_writereg = bmac_mii_writereg;
	mii->mii_statchg = bmac_mii_statchg;

	ifmedia_init(&mii->mii_media, 0, bmac_mediachange, bmac_mediastatus);
	mii_attach(self, mii, 0xffffffff, MII_PHY_ANY, MII_OFFSET_ANY, 0);

	/* Choose a default media. */
	if (sc->sc_flags & BMAC_BMACPLUS) {
		ifmedia_set(&mii->mii_media, IFM_ETHER|IFM_AUTO);
	} else {
		ifmedia_add(&mii->mii_media, IFM_ETHER|IFM_10_T, 0, NULL);
		ifmedia_set(&mii->mii_media, IFM_ETHER|IFM_10_T);
	}

	bmac_reset_chip(sc);

	if_attach(ifp);
	ether_ifattach(ifp, sc->sc_enaddr);
}

/*
 * Reset and enable bmac by heathrow FCR.
 */
void
bmac_reset_chip(sc)
	struct bmac_softc *sc;
{
	u_int v;

	dbdma_reset(sc->sc_txdma);
	dbdma_reset(sc->sc_rxdma);

	v = in32rb(heathrow_FCR);

	v |= EnetEnable;
	out32rb(heathrow_FCR, v);
	delay(50000);

	v |= ResetEnetCell;
	out32rb(heathrow_FCR, v);
	delay(50000);

	v &= ~ResetEnetCell;
	out32rb(heathrow_FCR, v);
	delay(50000);

	out32rb(heathrow_FCR, v);
}

int
bmac_init(ifp)
	struct ifnet *ifp;
{
	struct bmac_softc *sc = ifp->if_softc;
	u_int16_t *p;
	int i, tb;

	bmac_reset_chip(sc);

	if (sc->sc_flags & BMAC_BMACPLUS) {
		int phy = LIST_FIRST(&sc->sc_mii.mii_phys)->mii_phy; /* XXX */
		int bmcr;

		bmcr = bmac_mii_readreg(&sc->sc_dev, phy, MII_BMCR);
		bmcr &= ~BMCR_ISO;
		bmac_mii_writereg(&sc->sc_dev, phy, MII_BMCR, bmcr);
	}

	bmac_write_reg(sc, RXRST, RxResetValue);
	bmac_write_reg(sc, TXRST, TxResetBit);

	/* Wait for reset completion. */
	for (i = 1000; i > 0; i -= 10) {
		if ((bmac_read_reg(sc, TXRST) & TxResetBit) == 0)
			goto reset_done;
		delay(10);
	}
	printf("%s: reset timeout\n", ifp->if_xname);
reset_done:

	if (!(sc->sc_flags & BMAC_BMACPLUS))
		bmac_set_bits(sc, XCVRIF, ClkBit|SerialMode|COLActiveLow);

	asm ("mftb %0" : "=r"(tb));
	bmac_write_reg(sc, RSEED, tb);
	bmac_set_bits(sc, XIFC, TxOutputEnable);
	bmac_read_reg(sc, PAREG);

	/* Reset various counters. */
	bmac_write_reg(sc, NCCNT, 0);
	bmac_write_reg(sc, NTCNT, 0);
	bmac_write_reg(sc, EXCNT, 0);
	bmac_write_reg(sc, LTCNT, 0);
	bmac_write_reg(sc, FRCNT, 0);
	bmac_write_reg(sc, LECNT, 0);
	bmac_write_reg(sc, AECNT, 0);
	bmac_write_reg(sc, FECNT, 0);
	bmac_write_reg(sc, RXCV, 0);

	/* Set tx fifo information. */
	bmac_write_reg(sc, TXTH, 255);	/* 255 octets before tx starts */

	bmac_write_reg(sc, TXFIFOCSR, 0);
	bmac_write_reg(sc, TXFIFOCSR, TxFIFOEnable);

	/* Set rx fifo information. */
	bmac_write_reg(sc, RXFIFOCSR, 0);
	bmac_write_reg(sc, RXFIFOCSR, RxFIFOEnable);

	/* Clear status register. */
	bmac_read_reg(sc, STATUS);

	bmac_write_reg(sc, HASH3, 0);
	bmac_write_reg(sc, HASH2, 0);
	bmac_write_reg(sc, HASH1, 0);
	bmac_write_reg(sc, HASH0, 0);

	/* Set MAC address. */
	p = (u_short *)sc->sc_enaddr;
	bmac_write_reg(sc, MADD0, *p++);
	bmac_write_reg(sc, MADD1, *p++);
	bmac_write_reg(sc, MADD2, *p);

	bmac_write_reg(sc, RXCFG,
	    RxCRCEnable | RxHashFilterEnable | RxRejectOwnPackets);

	if (ifp->if_flags & IFF_PROMISC)
		bmac_set_bits(sc, RXCFG, RxPromiscEnable);

	bmac_init_dma(sc);

	/* Enable TX/RX */
	bmac_set_bits(sc, RXCFG, RxMACEnable);
	bmac_set_bits(sc, TXCFG, TxMACEnable);

	bmac_write_reg(sc, INTDISABLE, NormalIntEvents);

	ifp->if_flags |= IFF_RUNNING;
	ifp->if_flags &= ~IFF_OACTIVE;
	ifp->if_timer = 0;

	bmac_start(ifp);

	if (!(sc->sc_flags & BMAC_BMACPLUS))
		callout_reset(&sc->sc_tick_ch, hz, bmac_mii_tick, sc);

	return 0;
}

void
bmac_init_dma(sc)
	struct bmac_softc *sc;
{
	struct dbdma_desc *cmd;
	int i;

	dbdma_stop(sc->sc_txdma);
	dbdma_stop(sc->sc_rxdma);
	dbdma_reset(sc->sc_txdma);
	dbdma_reset(sc->sc_rxdma);

	bzero(sc->sc_txcmd, sizeof sc->sc_txcmd);
	bzero(sc->sc_rxcmd, sizeof sc->sc_rxcmd);

	cmd = sc->sc_rxcmd;
	for (i = 0; i < BMAC_RXBUFS; i++) {
		dbdma_cmd(cmd++, DBDMA_IN_LAST | DBDMA_INTR,
		    vtophys(mtod(sc->sc_rxbuf[i], vaddr_t)), BMAC_BUFLEN, 0);
	}
	dbdma_cmd(cmd, DBDMA_NOP | DBDMA_BR, 0, 0,
	    vtophys((vaddr_t)sc->sc_rxcmd));

	cmd = sc->sc_txcmd;
	for (i = 0; i < BMAC_TXBUFS * 2; i++) {
		dbdma_cmd(cmd, DBDMA_STOP, 0, 0, 0);
		cmd->dd_status = htole16(DBDMA_ACTIVE);
		cmd++;
	}
	dbdma_cmd(cmd, DBDMA_NOP | DBDMA_BR, 0, 0,
	    vtophys((vaddr_t)sc->sc_txcmd));

	sc->sc_rxlast = 0;
	dbdma_start(sc->sc_rxdma, sc->sc_rxcmd);
	sc->sc_txhead = 0;
	dbdma_start(sc->sc_txdma, sc->sc_txcmd);
}

int
bmac_intr(v)
	void *v;
{
	struct bmac_softc *sc = v;
	int stat;

	stat = bmac_read_reg(sc, STATUS);
	if (stat == 0)
		return 0;

#ifdef BMAC_DEBUG
	printf("bmac_intr status = 0x%x\n", stat);
#endif

	if (stat & IntFrameSent) {
		sc->sc_if.if_flags &= ~IFF_OACTIVE;
		sc->sc_if.if_timer = 0;
		sc->sc_if.if_opackets++;
		bmac_start(&sc->sc_if);
	}

	/* XXX should do more! */

	return 1;
}

int
bmac_rint(v)
	void *v;
{
	struct bmac_softc *sc = v;
	struct ifnet *ifp = &sc->sc_if;
	struct mbuf *m, *nm;
	struct dbdma_desc *cmd;
	int status, resid, len;
	int i, n;

	i = sc->sc_rxlast;
	for (n = 0; n < BMAC_RXBUFS; n++, i++) {
		if (i == BMAC_RXBUFS)
			i = 0;
		cmd = &sc->sc_rxcmd[i];
		status = dbdma_ld16(&cmd->dd_status);
		resid = dbdma_ld16(&cmd->dd_resid);

#ifdef BMAC_DEBUG
		if (status != 0 && status != 0x8440 && status != 0x9440)
			printf("bmac_rint status = 0x%x\n", status);
#endif

		if ((status & DBDMA_ACTIVE) == 0)	/* 0x9440 | 0x8440 */
			continue;
		len = BMAC_BUFLEN - resid - 2;	/* 2 == framelen */
#if 0
		/* XXX Sometimes bmac reads one extra byte. */
		if (len == ETHER_MAX_LEN + 1)
			len--;
#endif
		if (len < ETHER_MIN_LEN || len > ETHER_MAX_LEN) {
			printf("%s: packet length = %d discarded\n",
			    ifp->if_xname, len);
			ifp->if_ierrors++;
			goto next;
		}
		cmd->dd_command = htole16(DBDMA_STOP);

		nm = bmac_get(sc);
		if (nm == NULL) {
			ifp->if_ierrors++;
			goto next;
		}
		m = sc->sc_rxbuf[i];
		m->m_len = m->m_pkthdr.len = len;

#if NBPFILTER > 0
		/*
		 * Check if there's a BPF listener on this interface.
		 * If so, hand off the raw packet to BPF.
		 */
		if (ifp->if_bpf)
			bpf_mtap(ifp->if_bpf, m);
#endif
		(*ifp->if_input)(ifp, m);
		ifp->if_ipackets++;
		sc->sc_rxbuf[i] = nm;
#ifdef PMAP_UNMAP_POOLPAGE
		dbdma_st32(&cmd->dd_address,
		    PMAP_UNMAP_POOLPAGE(mtod(nm, vaddr_t)));
#else
		dbdma_st32(&cmd->dd_address, vtophys(mtod(nm, vaddr_t)));
#endif

next:
		cmd->dd_command = htole16(DBDMA_IN_LAST | DBDMA_INTR);
		cmd->dd_status = 0;
		cmd->dd_resid = 0;
		sc->sc_rxlast = i + 1;
	}
	dbdma_continue(sc->sc_rxdma);

	return 1;
}

void
bmac_reset(sc)
	struct bmac_softc *sc;
{
	int s;

	s = splnet();
	bmac_init(&sc->sc_if);
	splx(s);
}

void
bmac_stop(ifp, disable)
	struct ifnet *ifp;
	int disable;
{
	struct bmac_softc *sc = ifp->if_softc;
	int s;

	s = splnet();

	callout_stop(&sc->sc_tick_ch);
	mii_down(&sc->sc_mii);

	/* Disable TX/RX. */
	bmac_reset_bits(sc, TXCFG, TxMACEnable);
	bmac_reset_bits(sc, RXCFG, RxMACEnable);

	/* Disable all interrupts. */
	bmac_write_reg(sc, INTDISABLE, NoEventsMask);

	dbdma_stop(sc->sc_txdma);
	dbdma_stop(sc->sc_rxdma);

	ifp->if_flags &= ~(IFF_UP | IFF_RUNNING);
	ifp->if_timer = 0;

	splx(s);
}

void
bmac_start(ifp)
	struct ifnet *ifp;
{
	struct bmac_softc *sc = ifp->if_softc;
	struct dbdma_desc *cmd;
	struct mbuf *m;
	caddr_t buf;
	int tlen, status;

	if ((ifp->if_flags & (IFF_RUNNING | IFF_OACTIVE)) != IFF_RUNNING)
		return;

	for (;;) {
		if (ifp->if_flags & IFF_OACTIVE)
			break;

		cmd = &sc->sc_txcmd[sc->sc_txhead * 2 + 1];
		status = dbdma_ld16(&cmd->dd_status);
		if ((status & DBDMA_ACTIVE) == 0) {
			ifp->if_flags |= IFF_OACTIVE;
			break;
		}

		IFQ_DEQUEUE(&ifp->if_snd, m);
		if (m == 0)
			break;
#if NBPFILTER > 0
		/*
		 * If BPF is listening on this interface, let it see the
		 * packet before we commit it to the wire.
		 */
		if (ifp->if_bpf)
			bpf_mtap(ifp->if_bpf, m);
#endif

		buf = sc->sc_txbuf + sc->sc_txhead * BMAC_BUFLEN;
		tlen = bmac_put(sc, buf, m);
		m_freem(m);

		/* 5 seconds to watch for failing to transmit */
		ifp->if_timer = 5;
		ifp->if_opackets++;		/* # of pkts */

		bmac_transmit_packet(sc, buf, tlen);
		sc->sc_txhead++;
		if (sc->sc_txhead >= BMAC_TXBUFS)
			sc->sc_txhead = 0;
	}
}

void
bmac_transmit_packet(sc, buff, len)
	struct bmac_softc *sc;
	void *buff;
	int len;
{
	int cur = sc->sc_txhead;
	struct dbdma_desc *cmd = &sc->sc_txcmd[cur * 2];
	struct dbdma_desc *cmd0 = cmd;
	vaddr_t va = (vaddr_t)buff;

	cmd++;
	dbdma_cmd(cmd, DBDMA_OUT_LAST | DBDMA_INTR, vtophys(va), len, 0);

	cur++;
	if (cur >= BMAC_TXBUFS)
		cur = 0;
	cmd = &sc->sc_txcmd[cur * 2];
	dbdma_cmd(cmd, DBDMA_STOP, 0, 0, 0);

	cmd0->dd_command = htole16(DBDMA_NOP);
	asm volatile ("sync" ::: "memory");

	dbdma_continue(sc->sc_txdma);
}

int
bmac_put(sc, buff, m)
	struct bmac_softc *sc;
	caddr_t buff;
	struct mbuf *m;
{
	int len, tlen = 0;

	for (; m; m = m->m_next) {
		len = m->m_len;
		if (len == 0)
			continue;
		bcopy(mtod(m, caddr_t), buff, len);
		buff += len;
		tlen += len;
	}
	if (tlen > BMAC_BUFLEN)
		panic("%s: bmac_put packet overflow", sc->sc_dev.dv_xname);

	return tlen;
}

struct mbuf *
bmac_get(sc)
	struct bmac_softc *sc;
{
	struct mbuf *m;

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if (m == NULL)
		return NULL;
	m->m_pkthdr.rcvif = &sc->sc_if;
	m->m_flags |= M_HASFCS;

	MCLGET(m, M_DONTWAIT);
	if ((m->m_flags & M_EXT) == 0) {
		m_free(m);
		return NULL;
	}

	return m;
}

void
bmac_watchdog(ifp)
	struct ifnet *ifp;
{
	struct bmac_softc *sc = ifp->if_softc;

	bmac_reset_bits(sc, RXCFG, RxMACEnable);
	bmac_reset_bits(sc, TXCFG, TxMACEnable);

	printf("%s: device timeout\n", ifp->if_xname);
	ifp->if_oerrors++;

	bmac_reset(sc);
}

int
bmac_ioctl(ifp, cmd, data)
	struct ifnet *ifp;
	u_long cmd;
	caddr_t data;
{
	struct bmac_softc *sc = ifp->if_softc;
	struct ifreq *ifr = (struct ifreq *)data;
	int s, error = 0;

	s = splnet();

	switch (cmd) {

	case SIOCSIFADDR:
		ether_ioctl(ifp, cmd, data);
		break;

	case SIOCSIFFLAGS:
		ether_ioctl(ifp, cmd, data);
#ifdef BMAC_DEBUG
		if (ifp->if_flags & IFF_DEBUG)
			sc->sc_flags |= BMAC_DEBUGFLAG;
#endif
		break;

	case SIOCADDMULTI:
	case SIOCDELMULTI:
		error = ether_ioctl(ifp, cmd, data);
		if (error == ENETRESET) {
			/*
			 * Multicast list has changed; set the hardware filter
			 * accordingly.
			 */
			bmac_init(ifp);
			bmac_setladrf(sc);
			error = 0;
		}
		break;

	case SIOCGIFMEDIA:
	case SIOCSIFMEDIA:
		error = ifmedia_ioctl(ifp, ifr, &sc->sc_mii.mii_media, cmd);
		break;

	default:
		error = EINVAL;
	}

	splx(s);
	return error;
}

int
bmac_mediachange(ifp)
	struct ifnet *ifp;
{
	struct bmac_softc *sc = ifp->if_softc;

	return mii_mediachg(&sc->sc_mii);
}

void
bmac_mediastatus(ifp, ifmr)
	struct ifnet *ifp;
	struct ifmediareq *ifmr;
{
	struct bmac_softc *sc = ifp->if_softc;

	if (!(sc->sc_flags & BMAC_BMACPLUS))
		return;

	mii_pollstat(&sc->sc_mii);
	ifmr->ifm_status = sc->sc_mii.mii_media_status;
	ifmr->ifm_active = sc->sc_mii.mii_media_active;
}

/*
 * Set up the logical address filter.
 */
void
bmac_setladrf(sc)
	struct bmac_softc *sc;
{
	struct ifnet *ifp = &sc->sc_if;
	struct ether_multi *enm;
	struct ether_multistep step;
	u_int32_t crc;
	u_int16_t hash[4];
	int x;

	/*
	 * Set up multicast address filter by passing all multicast addresses
	 * through a crc generator, and then using the high order 6 bits as an
	 * index into the 64 bit logical address filter.  The high order bit
	 * selects the word, while the rest of the bits select the bit within
	 * the word.
	 */

	if (ifp->if_flags & IFF_PROMISC) {
		bmac_set_bits(sc, RXCFG, RxPromiscEnable);
		return;
	}

	if (ifp->if_flags & IFF_ALLMULTI) {
		hash[3] = hash[2] = hash[1] = hash[0] = 0xffff;
		goto chipit;
	}

	hash[3] = hash[2] = hash[1] = hash[0] = 0;

	ETHER_FIRST_MULTI(step, &sc->sc_ethercom, enm);
	while (enm != NULL) {
		if (bcmp(enm->enm_addrlo, enm->enm_addrhi, ETHER_ADDR_LEN)) {
			/*
			 * We must listen to a range of multicast addresses.
			 * For now, just accept all multicasts, rather than
			 * trying to set only those filter bits needed to match
			 * the range.  (At this time, the only use of address
			 * ranges is for IP multicast routing, for which the
			 * range is big enough to require all bits set.)
			 */
			hash[3] = hash[2] = hash[1] = hash[0] = 0xffff;
			ifp->if_flags |= IFF_ALLMULTI;
			goto chipit;
		}

		crc = ether_crc32_le(enm->enm_addrlo, ETHER_ADDR_LEN);

		/* Just want the 6 most significant bits. */
		crc >>= 26;

		/* Set the corresponding bit in the filter. */
		hash[crc >> 4] |= 1 << (crc & 0xf);

		ETHER_NEXT_MULTI(step, enm);
	}

	ifp->if_flags &= ~IFF_ALLMULTI;

chipit:
	bmac_write_reg(sc, HASH0, hash[0]);
	bmac_write_reg(sc, HASH1, hash[1]);
	bmac_write_reg(sc, HASH2, hash[2]);
	bmac_write_reg(sc, HASH3, hash[3]);
	x = bmac_read_reg(sc, RXCFG);
	x &= ~RxPromiscEnable;
	x |= RxHashFilterEnable;
	bmac_write_reg(sc, RXCFG, x);
}

int
bmac_mii_readreg(dev, phy, reg)
	struct device *dev;
	int phy, reg;
{
	return mii_bitbang_readreg(dev, &bmac_mbo, phy, reg);
}

void
bmac_mii_writereg(dev, phy, reg, val)
	struct device *dev;
	int phy, reg, val;
{
	mii_bitbang_writereg(dev, &bmac_mbo, phy, reg, val);
}

u_int32_t
bmac_mbo_read(dev)
	struct device *dev;
{
	struct bmac_softc *sc = (void *)dev;

	return bmac_read_reg(sc, MIFCSR);
}

void
bmac_mbo_write(dev, val)
	struct device *dev;
	u_int32_t val;
{
	struct bmac_softc *sc = (void *)dev;

	bmac_write_reg(sc, MIFCSR, val);
}

void
bmac_mii_statchg(dev)
	struct device *dev;
{
	struct bmac_softc *sc = (void *)dev;
	int x;

	/* Update duplex mode in TX configuration */
	x = bmac_read_reg(sc, TXCFG);
	if ((IFM_OPTIONS(sc->sc_mii.mii_media_active) & IFM_FDX) != 0)
		x |= TxFullDuplex;
	else
		x &= ~TxFullDuplex;
	bmac_write_reg(sc, TXCFG, x);

#ifdef BMAC_DEBUG
	printf("bmac_mii_statchg 0x%x\n",
	    IFM_OPTIONS(sc->sc_mii.mii_media_active));
#endif
}

void
bmac_mii_tick(v)
	void *v;
{
	struct bmac_softc *sc = v;
	int s;

	s = splnet();
	mii_tick(&sc->sc_mii);
	splx(s);

	callout_reset(&sc->sc_tick_ch, hz, bmac_mii_tick, sc);
}
