/*	$NetBSD: if_mcvar.h,v 1.6 2001/07/26 11:23:37 tsubai Exp $	*/

/*-
 * Copyright (c) 1997 David Huang <khym@bga.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <machine/dbdma.h>
#include <machine/pio.h>

#if 1
#define integrate
#define hide
#else
#define integrate	static /*__inline*/
#define hide		static
#endif

#define	MC_REGSPACING	16
#define	MC_REGSIZE	MACE_NREGS * MC_REGSPACING
#define	MACE_REG(x)	((x)*MC_REGSPACING)

#define	NIC_GET(sc, reg)	in8((sc)->sc_reg + MACE_REG(reg))
#define	NIC_PUT(sc, reg, val)	out8((sc)->sc_reg + MACE_REG(reg), val)

#ifndef	MC_RXDMABUFS
#define	MC_RXDMABUFS	4
#endif
#if (MC_RXDMABUFS < 2)
#error Must have at least two buffers for DMA!
#endif

struct mc_rxframe {
	u_int8_t	rx_rcvcnt;
	u_int8_t	rx_rcvsts;
	u_int8_t	rx_rntpc;
	u_int8_t	rx_rcvcc;
};

#define sc_if sc_ethercom.ec_if

struct mc_softc {
	struct device	sc_dev;		/* base device glue */
	struct ethercom	sc_ethercom;	/* Ethernet common part */
	struct ifmedia	sc_media;

	u_int8_t	sc_biucc;
	u_int8_t	sc_fifocc;
	u_int8_t	sc_plscc;
	u_int8_t	sc_enaddr[6];
	u_int8_t	sc_pad[3];
	int		sc_havecarrier; /* carrier status */
	void		(*sc_bus_init) __P((struct mc_softc *));
	void		(*sc_putpacket) __P((struct mc_softc *, u_int));

	u_char		*sc_reg;
	u_char		*sc_txbuf;
	struct mbuf	*sc_rxbuf[MC_RXDMABUFS];
	paddr_t		sc_txbuf_phys;
	int		sc_tail;

	int		sc_node;
	struct dbdma_reg *sc_txdma;
	struct dbdma_reg *sc_rxdma;
	struct dbdma_desc sc_txdmacmd[2];
	struct dbdma_desc sc_rxdmacmd[MC_RXDMABUFS + 1];
};

int	mcsetup __P((struct mc_softc *, u_int8_t *));
int	mcintr __P((void *));
void	mc_rint __P((struct mc_softc *, struct mbuf *, struct mc_rxframe *));
struct mbuf *mace_get __P((struct mc_softc *));
