/*	$Id: wdc_obio.c,v 1.25 2004/01/10 07:51:09 tsubai Exp $	*/

/*-
 * Copyright (c) 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Charles M. Hannum and by Onno van der Linden.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/device.h>
#include <sys/systm.h>

#include <uvm/uvm_extern.h>

#include <machine/autoconf.h>
#include <machine/bus.h>
#include <machine/dbdma.h>

#include <dev/ata/atareg.h>
#include <dev/ata/atavar.h>
#include <dev/ic/wdcvar.h>
#include <dev/ofw/openfirm.h>

#define WDC_DEFAULT_PIO_IRQ	13	/* XXX */
#define WDC_DEFAULT_DMA_IRQ	2	/* XXX */

#define WDC_OPTIONS_DMA 0x01
#define WDC_OPTIONS_PIO 0x02

#define AUXREG_OFFSET	0x160
#define CONFIG_REG	0x200		/* IDE access timing register */

struct wdc_obio_softc {
	struct wdc_softc sc_wdcdev;
	struct wdc_channel *wdc_chanptr;
	struct wdc_channel wdc_channel;
	struct ata_queue wdc_queue;
	struct dbdma_reg *sc_dmareg;
	struct dbdma_desc sc_dmacmd[20];
	u_int sc_timing_r[2];
	u_int sc_timing_w[2];
	void (*sc_calc_timing)(struct wdc_obio_softc *, int);
	void *sc_ih;
};

int wdc_obio_probe __P((struct device *, struct cfdata *, void *));
void wdc_obio_attach __P((struct device *, struct device *, void *));
int wdc_obio_detach __P((struct device *, int));
int wdc_obio_dma_init __P((void *, int, int, void *, size_t, int));
void wdc_obio_dma_start __P((void *, int, int));
int wdc_obio_dma_finish __P((void *, int, int, int));
void wdc_obio_set_modes __P((struct wdc_channel *));
static void calc_timing_ohare __P((struct wdc_obio_softc *, int));
static void calc_timing_ata3 __P((struct wdc_obio_softc *, int));
static void calc_timing_ata4 __P((struct wdc_obio_softc *, int));

struct cfattach wdc_obio_ca = {
	"wdc_obio", {}, sizeof(struct wdc_obio_softc),
	wdc_obio_probe, wdc_obio_attach, wdc_obio_detach, wdcactivate
};

int
wdc_obio_probe(parent, match, aux)
	struct device *parent;
	struct cfdata *match;
	void *aux;
{
	struct confargs *ca = aux;

	if (ca->ca_nreg < 8)
		return 0;

	/* XXX should not use name */
	if (strcmp(ca->ca_name, "ATA") == 0 ||
	    strcmp(ca->ca_name, "ata") == 0 ||
	    strcmp(ca->ca_name, "ata0") == 0 ||
	    strcmp(ca->ca_name, "ide") == 0)
		return 1;

	if (strcmp(ca->ca_compat, "heathrow-ata") == 0 ||
	    strcmp(ca->ca_compat, "keylargo-ata") == 0)
		return 1;

	return 0;
}

void
wdc_obio_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct wdc_obio_softc *sc = (void *)self;
	struct confargs *ca = aux;
	struct wdc_channel *chp = &sc->wdc_channel;
	int use_dma = 1;
	int i, intr;
	char name[80];

#if 0
	if (sc->sc_wdcdev.sc_dev.dv_cfdata->cf_flags & WDC_OPTIONS_DMA)
		use_dma = 1;
#endif
	if (sc->sc_wdcdev.sc_dev.dv_cfdata->cf_flags & WDC_OPTIONS_PIO)
		use_dma = 0;
	if (ca->ca_nreg < 16)
		use_dma = 0;

#ifdef DIAGNOSTIC
	if ((vaddr_t)sc->sc_dmacmd & 0x0f) {
		printf(": bad dbdma alignment\n");
		return;
	}
#endif

	/*
	 * XXX There is no AAPL,interrupts property on some compatibles.
	 * (ex. UMAX Apus)
	 */
	if (ca->ca_nintr == -1) {
		ca->ca_nintr = 4;
		ca->ca_intr[0] = WDC_DEFAULT_PIO_IRQ;
		printf(" irq property not found; using");
	}

	if (ca->ca_nintr >= 4) {
		intr = ca->ca_intr[0];
		printf(" irq %d\n", intr);
	} else {
		printf(": couldn't get irq property\n");
		return;
	}

	ca->ca_reg[0] += ca->ca_baseaddr;
	ca->ca_reg[2] += ca->ca_baseaddr;

	chp->cmd_iot = chp->ctl_iot =
	    macppc_make_bus_space_tag(ca->ca_reg[0], 0);

	if (bus_space_map(chp->cmd_iot, 0, NBPG, 0, &chp->cmd_baseioh)) {
		printf("%s: couldn't map registers\n", self->dv_xname);
		return;
	}

	/* XXX Note: bus_space_subregion() always succeeds. */
	bus_space_subregion(chp->cmd_iot, chp->cmd_baseioh, AUXREG_OFFSET, 1,
	     &chp->ctl_ioh);
	for (i = 0; i < WDC_NREG; i++)
		bus_space_subregion(chp->cmd_iot, chp->cmd_baseioh, i*16, 2,
		    &chp->cmd_iohs[i]);

	sc->sc_ih = intr_establish(intr, IST_LEVEL, IPL_BIO, wdcintr, chp);

	sc->sc_wdcdev.PIO_cap = 4;
	sc->sc_wdcdev.DMA_cap = 2;
	sc->sc_wdcdev.cap |= WDC_CAPABILITY_DATA16 | WDC_CAPABILITY_MODE;
	sc->wdc_chanptr = chp;
	sc->sc_wdcdev.channels = &sc->wdc_chanptr;
	sc->sc_wdcdev.nchannels = 1;
	sc->sc_wdcdev.dma_arg = sc;
	sc->sc_wdcdev.dma_init = wdc_obio_dma_init;
	sc->sc_wdcdev.dma_start = wdc_obio_dma_start;
	sc->sc_wdcdev.dma_finish = wdc_obio_dma_finish;
	sc->sc_wdcdev.set_modes = wdc_obio_set_modes;
	sc->sc_calc_timing = calc_timing_ohare;

	if (use_dma) {
		sc->sc_dmareg = mapiodev(ca->ca_reg[2], ca->ca_reg[3]);
		sc->sc_wdcdev.cap |= WDC_CAPABILITY_DMA;
	}

	/* Minimum cycle time is 150ns (DMA MODE 1) on OHare. */
	if (OF_finddevice("/bandit/ohare") != -1) {
		sc->sc_wdcdev.PIO_cap = 3;
		sc->sc_wdcdev.DMA_cap = 1;
	}

	if (strcmp(ca->ca_compat, "keylargo-ata") == 0)
		sc->sc_calc_timing = calc_timing_ata3;

	/* XXX use "model" */
	if (strcmp(ca->ca_name, "ata-4") == 0) {
		if (use_dma)
			sc->sc_wdcdev.cap |= WDC_CAPABILITY_UDMA;
		sc->sc_wdcdev.UDMA_cap = 4;
		sc->sc_calc_timing = calc_timing_ata4;
	}

	chp->ch_channel = 0;
	chp->ch_wdc = &sc->sc_wdcdev;
	chp->ch_queue = &sc->wdc_queue;

#define OHARE_FEATURE_REG	0xf3000038

	/* XXX Enable wdc1 by feature reg. */
	bzero(name, sizeof name);
	OF_package_to_path(ca->ca_node, name, sizeof name);
	if (strcmp(name, "/bandit@F2000000/ohare@10/ata@21000") == 0) {
		u_int x;

		x = in32rb(OHARE_FEATURE_REG);
		x |= 8;
		out32rb(OHARE_FEATURE_REG, x);
	}

	wdcattach(chp);
}

void
wdc_obio_set_modes(chp)
	struct wdc_channel *chp;
{
	struct wdc_obio_softc *sc = (void *)chp->ch_wdc;
	struct ata_drive_datas *drvp0 = &chp->ch_drive[0];
	struct ata_drive_datas *drvp1 = &chp->ch_drive[1];
	struct ata_drive_datas *drvp;
	int drive;

	if ((drvp0->drive_flags & DRIVE) && (drvp1->drive_flags & DRIVE)) {
		drvp0->PIO_mode = drvp1->PIO_mode =
		    min(drvp0->PIO_mode, drvp1->PIO_mode);
	}

	for (drive = 0; drive < 2; drive++) {
		drvp = &chp->ch_drive[drive];
		if (drvp->drive_flags & DRIVE) {
			(*sc->sc_calc_timing)(sc, drive);
			bus_space_write_4(chp->cmd_iot, chp->cmd_baseioh,
			    CONFIG_REG, sc->sc_timing_r[drive]);
		}
	}
}

/*
 * IDE transfer timings
 */
static const u_int pio_timing_ohare[] = {	/* 0x000007ff (1-5-5) */
	0x00000526,	/* Mode 0 */		/* XXX == pio_timing_ata3 */
	0x00000085,	/*      1 */
	0x00000025,	/*      2 */
	0x00000025,	/*      3 */
	0x00000025	/*      4 */
};
static const u_int dma_timing_ohare[] = {	/* 0x003ff800 (1-5-5) */
	0x00074000,	/* Mode 0 */
	0x00221000,	/*      1 */
	0x00211000	/*      2 */		/* XXX r/w? */
};

static const u_int pio_timing_ata3[] = {	/* 0x000007ff (1-5-5) */
	0x00000526,	/* Mode 0 */
	0x00000085,	/*      1 */
	0x00000025,	/*      2 */
	0x00000025,	/*      3 */
	0x00000025	/*      4 */
};
static const u_int dma_timing_ata3[] = {	/* 0x001ff800 (5-5) */
	0x00084000,	/* Mode 0 */
	0x00021800,	/*      1 */
	0x00011800	/*      2 */		/* XXX r/w? */
};

static const u_int pio_timing_ata4[] = {	/* 0x000003ff (5-5) */
	0x0000038c,	/* Mode 0 */
	0x0000020a,	/*      1 */
	0x00000127,	/*      2 */
	0x000000c6,	/*      3 */
	0x00000065 	/*      4 */
};
static const u_int dma_timing_ata4[] = {	/* 0x000ffc00 (5-5) */
	0x00084000,	/* Mode 0 */
	0x00029800,	/*      1 */
	0x00019400	/*      2 */
};
static const u_int udma_timing_ata4[] = {	/* 0x1ff00000 (4-4-1) */
	0x19100000,	/* Mode 0 */
	0x14d00000,	/*      1 */
	0x10900000,	/*      2 */
	0x0c700000,	/*      3 */
	0x0c500000	/*      4 */
};

/*
 * Timing calculation for OHare and Heathrow.
 */
void
calc_timing_ohare(sc, drive)
	struct wdc_obio_softc *sc;
	int drive;
{
	struct wdc_channel *chp = &sc->wdc_channel;
	struct ata_drive_datas *drvp = &chp->ch_drive[drive];
	int piomode = drvp->PIO_mode;
	int dmamode = drvp->DMA_mode;
	u_int pioconf, dmaconf;

	pioconf = pio_timing_ohare[piomode];

	if (drvp->drive_flags & DRIVE_DMA)
		dmaconf = dma_timing_ohare[dmamode];
	else
		dmaconf = 0;

	sc->sc_timing_r[drive] = pioconf | dmaconf;
	sc->sc_timing_w[drive] = pioconf | dmaconf;
}

/*
 * Timing calculation for KeyLargo ata-3.
 */
void
calc_timing_ata3(sc, drive)
	struct wdc_obio_softc *sc;
	int drive;
{
	struct wdc_channel *chp = &sc->wdc_channel;
	struct ata_drive_datas *drvp = &chp->ch_drive[drive];
	int piomode = drvp->PIO_mode;
	int dmamode = drvp->DMA_mode;
	u_int pioconf, dmaconf;

	pioconf = pio_timing_ata3[piomode];

	if (drvp->drive_flags & DRIVE_DMA)
		dmaconf = dma_timing_ata3[dmamode];
	else
		dmaconf = 0;

	sc->sc_timing_r[drive] = pioconf | dmaconf;
	sc->sc_timing_w[drive] = pioconf | dmaconf;
}

/*
 * Timing calculation for KeyLargo ata-4.
 */
void
calc_timing_ata4(sc, drive)
	struct wdc_obio_softc *sc;
	int drive;
{
	struct wdc_channel *chp = &sc->wdc_channel;
	struct ata_drive_datas *drvp = &chp->ch_drive[drive];
	int piomode = drvp->PIO_mode;
	int dmamode = drvp->DMA_mode;
	int udmamode = drvp->UDMA_mode;
	u_int pioconf, dmaconf;

	pioconf = pio_timing_ata4[piomode];
	dmaconf = 0;
	if (drvp->drive_flags & DRIVE_DMA)
		dmaconf = dma_timing_ata4[dmamode];
	if (drvp->drive_flags & DRIVE_UDMA)
		dmaconf = udma_timing_ata4[udmamode];

	sc->sc_timing_r[drive] = pioconf | dmaconf;
	sc->sc_timing_w[drive] = pioconf | dmaconf;

	if (drvp->drive_flags & DRIVE_UDMA)
		sc->sc_timing_r[drive] |= 0x00800000;
}

int
wdc_obio_detach(self, flags)
	struct device *self;
	int flags;
{
	struct wdc_obio_softc *sc = (void *)self;
	int error;

	if ((error = wdcdetach(self, flags)) != 0)
		return error;

	intr_disestablish(sc->sc_ih);

	/* Unmap our i/o space. */
	bus_space_unmap(chp->cmd_iot, chp->cmd_baseioh, NBPG);

	/* Unmap DMA registers. */
	/* XXX unmapiodev(sc->sc_dmareg); */

	return 0;
}

int
wdc_obio_dma_init(v, channel, drive, databuf, datalen, flags)
	void *v;
	int channel, drive, flags;
	void *databuf;
	size_t datalen;
{
	struct wdc_obio_softc *sc = v;
	struct dbdma_desc *cmdp = sc->sc_dmacmd;
	struct wdc_channel *chp = &sc->wdc_channel;
	vaddr_t va = (vaddr_t)databuf;
	int read = flags & WDC_DMA_READ;
	int cmd = read ? DBDMA_IN : DBDMA_OUT;
	u_int offset;

	bus_space_write_4(chp->cmd_iot, chp->cmd_baseioh, CONFIG_REG,
	    read ? sc->sc_timing_r[drive] : sc->sc_timing_w[drive]);
	bus_space_read_4(chp->cmd_iot, chp->cmd_baseioh, CONFIG_REG);

	offset = va & PGOFSET;

	/* if va is not page-aligned, setup the first page */
	if (offset != 0) {
		int rest = NBPG - offset;	/* the rest of the page */

		if (datalen > rest) {		/* if continues to next page */
			dbdma_cmd(cmdp++, cmd, vtophys(va), rest, 0);
			datalen -= rest;
			va += rest;
		}
	}

	/* now va is page-aligned */
	while (datalen > NBPG) {
		dbdma_cmd(cmdp++, cmd, vtophys(va), NBPG, 0);
		datalen -= NBPG;
		va += NBPG;
	}

	/* the last page (datalen <= NBPG here) */
	dbdma_cmd(cmdp++, cmd | DBDMA_LAST, vtophys(va), datalen, 0);
	dbdma_cmd(cmdp, DBDMA_STOP, 0, 0, 0);

	return 0;
}

void
wdc_obio_dma_start(v, channel, drive)
	void *v;
	int channel, drive;
{
	struct wdc_obio_softc *sc = v;

	dbdma_start(sc->sc_dmareg, sc->sc_dmacmd);
}

int
wdc_obio_dma_finish(v, channel, drive, force)
	void *v;
	int channel, drive, force;
{
	struct wdc_obio_softc *sc = v;

	dbdma_flush(sc->sc_dmareg);
	return 0;
}
