/*	$Id: dbdma.h,v 1.4 2002/10/31 14:49:07 tsubai Exp $	*/

/*-
 * Copyright (c) 2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _MACHINE_DBDMA_H_
#define _MACHINE_DBDMA_H_

/* DBDMA descriptor commands */
#define DBDMA_OUT	0x0000		/* Transfer more data from memory */
#define DBDMA_OUT_LAST	0x1000		/* Transfer last data from memory */
#define DBDMA_IN	0x2000		/* Transfer more data to memory */
#define DBDMA_IN_LAST	0x3000		/* Transfer last data to memory */
#define DBDMA_STQ	0x4000		/* Store immediate value */
#define DBDMA_LDQ	0x5000		/* Load immediate value */
#define DBDMA_NOP	0x6000		/* No operation */
#define DBDMA_STOP	0x7000		/* Suspend command processing */
#define DBDMA_LAST	0x1000

#define DBDMA_STREAM0	0x0000		/* Default stream */
#define DBDMA_STREAM1	0x0100
#define DBDMA_STREAM2	0x0200
#define DBDMA_STREAM3	0x0300
#define DBDMA_REGS	0x0500		/* Channel-state register space */
#define DBDMA_SYSTEM	0x0600		/* System memory space */

#define DBDMA_INTR_N	0x0000		/* Interrupt Never */
#define DBDMA_INTR_T	0x0010		/* Interrupt if True */
#define DBDMA_INTR_F	0x0020		/* Interrupt if False */
#define DBDMA_INTR_A	0x0030		/* Interrupt Always */
#define DBDMA_INTR	DBDMA_INTR_A

#define DBDMA_BR_N	0x0000		/* Branch Never */
#define DBDMA_BR_T	0x0004		/* Branch if True */
#define DBDMA_BR_F	0x0008		/* Branch if False */
#define DBDMA_BR_A	0x000c		/* Branch Always */
#define DBDMA_BR	DBDMA_BR_A

struct dbdma_desc {
	u_int16_t dd_count;
	u_int16_t dd_command;
	u_int32_t dd_address;
	u_int32_t dd_cmddep;
	volatile u_int16_t dd_resid;
	volatile u_int16_t dd_status;
} __attribute__((__aligned__(16)));

void dbdma_cmd(struct dbdma_desc *, u_int, paddr_t, u_int, u_int);

u_int16_t dbdma_ld16(volatile void *);
u_int32_t dbdma_ld32(volatile void *);
void dbdma_st16(volatile void *, u_int16_t);
void dbdma_st32(volatile void *, u_int32_t);

/*
 * Utility functions.
 */
extern inline u_int16_t
dbdma_ld16(addr)
	volatile void *addr;
{
	u_int16_t data;

	asm volatile ("lhbrx %0,0,%1" : "=r"(data) : "r"(addr));
	return data;
}

extern inline u_int32_t
dbdma_ld32(addr)
	volatile void *addr;
{
	u_int32_t data;

	asm volatile ("lwbrx %0,0,%1" : "=r"(data) : "r"(addr));
	return data;
}

extern inline void
dbdma_st16(addr, data)
	volatile void *addr;
	u_int16_t data;
{
	asm volatile ("sthbrx %0,0,%1" :: "r"(data), "r"(addr) : "memory");
}

extern inline void
dbdma_st32(addr, data)
	volatile void *addr;
	u_int32_t data;
{
	asm volatile ("stwbrx %0,0,%1" :: "r"(data), "r"(addr) : "memory");
}

/*
 * Build a dbdma descriptor.
 */
extern inline void
dbdma_cmd(dp, cmd, addr, len, cmddep)
	struct dbdma_desc *dp;
	u_int cmd, len, cmddep;
	paddr_t addr;
{
	dbdma_st16(&dp->dd_count, len);
	dbdma_st32(&dp->dd_address, addr);
	dbdma_st32(&dp->dd_cmddep, cmddep);
	dp->dd_resid = 0;
	dp->dd_status = 0;
	dbdma_st16(&dp->dd_command, cmd);
}

struct dbdma_reg {
	volatile u_int32_t dr_control;		/* Channel control */
	volatile u_int32_t dr_status;		/* Channel status */
	volatile u_int32_t dr_cmdptrhi;		/* Reserved */
	volatile u_int32_t dr_cmdptrlo;		/* DMA descriptor address */
	volatile u_int32_t dr_intrsel;		/* Interrupt condition */
	volatile u_int32_t dr_branchsel;	/* Branch condition */
	volatile u_int32_t dr_waitsel;		/* Wait condition */
};

/* dr_control and dr_status definitions */
#define DBDMA_RUN	0x8000		/* Start execution */
#define DBDMA_PAUSE	0x4000		/* Suspend execution */
#define DBDMA_FLUSH	0x2000		/* Update memory */
#define DBDMA_WAKE	0x1000		/* Restart */
#define DBDMA_DEAD	0x0800		/* Device error */
#define DBDMA_ACTIVE	0x0400		/* DMA is active */
#define DBDMA_BT	0x0100		/* Branch taken */
#define DBDMA_STATUS	0x00ff		/* Channel status mask */

/* DBDMA access functions */
void dbdma_start(struct dbdma_reg *, struct dbdma_desc *);
void dbdma_stop(struct dbdma_reg *);
void dbdma_pause(struct dbdma_reg *);
void dbdma_flush(struct dbdma_reg *);
void dbdma_continue(struct dbdma_reg *);
void dbdma_reset(struct dbdma_reg *);

#endif /* _MACHINE_DBDMA_H_ */
