/*	$Id: autoconf.c,v 1.19 2003/12/20 12:04:23 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/conf.h>
#include <sys/device.h>
#include <sys/reboot.h>
#include <sys/systm.h>

#include <machine/autoconf.h>
#include <machine/bus.h>
#include <machine/cpufunc.h>
#include <machine/pio.h>
#include <machine/stdarg.h>

#include <dev/ofw/openfirm.h>
#include <dev/pci/pcivar.h>
#include <dev/scsipi/scsi_all.h>
#include <dev/scsipi/scsipi_all.h>
#include <dev/scsipi/scsiconf.h>
#include <dev/ata/atavar.h>

void canonicalize_bootpath __P((void));
void ofw_stack __P((void));

static char cbootpath[256];
struct device *booted_device;	/* boot device */
int booted_partition;		/* ...and partition on that device */

u_int *heathrow_FCR = NULL;
char macppc_model[32];

/*
 * Determine device configuration for a machine.
 */
void
cpu_configure()
{
	int node;

	init_interrupt();
	calc_delay();
	canonicalize_bootpath();

	node = OF_finddevice("/");
	bzero(macppc_model, sizeof macppc_model);
	OF_getprop(node, "model", macppc_model, sizeof macppc_model);

	if (config_rootfound("mainbus", NULL) == NULL)
		panic("configure: mainbus not configured");

	spl0();

	/*
	 * Now allow hardware interrupts.
	 */
	mtmsr(mfmsr() | PSL_EE | PSL_RI);
}

void
canonicalize_bootpath()
{
	int node;
	char *p;
	char last[64], type[32];

	strcpy(cbootpath, bootpath);

	/*
	 * Replace alias.
	 */
	if (bootpath[0] != '/') {
		char alias[32], *src = bootpath, *dst = alias;

		while (strchr("/\\:[]@", *src) == NULL)
			*dst++ = *src++;
		*dst = 0;
		node = OF_finddevice("/aliases");
		if (node == -1)
			panic("no /aliases");
		bzero(cbootpath, sizeof cbootpath);
		if (OF_getprop(node, alias, cbootpath, sizeof cbootpath) != -1)
			strcat(cbootpath, src);
	}

	/*
	 * Strip kernel name.  bootpath contains "OF-path"/"kernel".
	 *
	 * for example:
	 *   /bandit@F2000000/gc@10/53c94@10000/sd@0,0/netbsd	(OF-1.x)
	 *   /pci/mac-io/ata-3@2000/disk@0:0/netbsd.new		(OF-3.x)
	 */
	while ((node = OF_finddevice(cbootpath)) == -1) {
		if ((p = strrchr(cbootpath, '/')) == NULL) {

			/* Couldn't canonicalize... use bootpath anyway. */
			strcpy(cbootpath, bootpath);
			return;
		}
		*p = 0;
	}

	/*
	 * cbootpath is a valid OF path.  Use package-to-path to
	 * canonicalize pathname.
	 */

	/* Back up the last component for later use. */
	if ((p = strrchr(cbootpath, '/')) != NULL)
		strcpy(last, p);
	else
		last[0] = 0;

	bzero(cbootpath, sizeof cbootpath);
	OF_package_to_path(node, cbootpath, sizeof cbootpath);

	/*
	 * "package-to-path" method seems to return no (or bogus) unit-address
	 * for disks.  (i.e. "/pci@.../.../disk" or
	 * "/bandit@.../.../sd@0,0" for ID != 0 disk).
	 *
	 * XXX Need rethink.
	 */
	bzero(type, sizeof type);
	OF_getprop(node, "device_type", type, sizeof type);
	if (strcmp(type, "block") == 0) {
		node = OF_parent(node);
		bzero(cbootpath, sizeof cbootpath);
		OF_package_to_path(node, cbootpath, sizeof cbootpath);
		strcat(cbootpath, last);
	}

	if ((p = strrchr(cbootpath, '/')) != NULL &&
	    (p = strrchr(p, ':')) != NULL) {
		*p++ = 0;
		/* booted_partition = *p - '0';		XXX correct? */
	}

	printf("bootpath: %s\n", cbootpath);
}

#define DEVICE_IS device_is

static int device_is(struct device *, char *);

int
device_is(dev, name)
	struct device *dev;
	char *name;
{
	int n = strlen(name);

	if (strncmp(dev->dv_xname, name, n) != 0)
		return 0;
	if (dev->dv_xname[n] < '0' || dev->dv_xname[n] > '9')
		return 0;

	return 1;
}

/*
 * device_register is called from config_attach as each device is
 * attached. We use it to find the NetBSD device corresponding to the
 * known OF boot device.
 */
void
device_register(dev, aux)
	struct device *dev;
	void *aux;
{
	static struct device *parent;
	static char *cp = cbootpath;
	u_int addr;
	char *p;

	if (booted_device)
		return;

	/* Skip over devices not represented in the OF tree. */
	if (DEVICE_IS(dev, "mainbus")) {
		parent = dev;
		return;
	}
	if (DEVICE_IS(dev, "atabus") || DEVICE_IS(dev, "atapibus") ||
	    DEVICE_IS(dev, "pci") || DEVICE_IS(dev, "scsibus"))
		return;

	if (DEVICE_IS(dev->dv_parent, "atabus") ||
	    DEVICE_IS(dev->dv_parent, "pci") ||
	    DEVICE_IS(dev->dv_parent, "scsibus")) {
		if (dev->dv_parent->dv_parent != parent)
			return;
	} else if (DEVICE_IS(dev->dv_parent, "atapibus")) {
		if (dev->dv_parent->dv_parent->dv_parent != parent)
			return;
	} else {
		if (dev->dv_parent != parent)
			return;
	}

	/* Get the address part of the current path component. */
	p = strchr(cp, '@');
	if (p)
		addr = strtoul(p + 1, &p, 16);
	else
		addr = 0;

	if (DEVICE_IS(dev->dv_parent, "mainbus")) {
		struct confargs *ca = aux;

		if (strcmp(ca->ca_name, "ofw") == 0)		/* XXX */
			return;
		if (addr != ca->ca_reg[0])
			return;
		goto matched;
	}
	if (DEVICE_IS(dev->dv_parent, "pci")) {
		struct pci_attach_args *pa = aux;

		if (addr == pa->pa_device)
			goto matched;
		return;
	}
	if (DEVICE_IS(dev->dv_parent, "obio")) {
		struct confargs *ca = aux;

		if (addr == ca->ca_reg[0])
			goto matched;
		return;
	}
	if (DEVICE_IS(dev->dv_parent, "scsibus") ||
	    DEVICE_IS(dev->dv_parent, "atapibus")) {
		struct scsipibus_attach_args *sa = aux;

		/* periph_target is target for scsi, drive # for atapi. */
		if (addr == sa->sa_periph->periph_target)
			goto matched;
		return;
	}
	if (DEVICE_IS(dev->dv_parent->dv_parent, "wdc")) {
		struct ata_device *adev = aux;

		if (addr == adev->adev_drv_data->drive)
			goto matched;
		return;
	}
	if (DEVICE_IS(dev->dv_parent->dv_parent, "acardide")) {
		struct ata_device *adev = aux;
		int channel = adev->adev_channel;
		int drive = adev->adev_drv_data->drive;

		/*
		 * ACARD PCI IDE card.
		 *  0: primary master  2: secondary master
		 *  1: primary slave   3: secondary slave
		 */
		if (addr == (channel << 1 | drive))
			goto matched;
		return;
	}
	if (DEVICE_IS(dev->dv_parent->dv_parent, "cmdide") ||
	    DEVICE_IS(dev->dv_parent->dv_parent, "pdcide")) {
		struct ata_device *adev = aux;
		int channel = adev->adev_channel;
		int drive = adev->adev_drv_data->drive;

		/*
		 * OF splits channel and drive into separate path components
		 * for on-board cmd646-ata, so check the addr part of the
		 * next component.
		 */
		if (addr != channel)
			return;
		p = strchr(p, '@');
		if (p == NULL)
			return;
		if (strtoul(p + 1, &p, 16) != drive)
			return;
		goto matched;
	}
	return;

matched:
	/*
	 * If we reach this point, then dev is a match for the current
	 * path component.
	 */
	if (p == NULL || *p == 0) {
		/* dev matches the last path component. */
		booted_device = dev;
		return;
	}

	parent = dev;
	cp = p;
}

/*
 * Setup root device.
 * Configure swap area.
 */
void
cpu_rootconf()
{
	printf("boot device: %s\n",
	    booted_device ? booted_device->dv_xname : "<unknown>");

	setroot(booted_device, booted_partition);
}

int
#ifdef __STDC__
OF_interpret(char *cmd, int nreturns, ...)
#else
OF_interpret(cmd, nreturns, va_alist)
	char *cmd;
	int nreturns;
	va_dcl
#endif
{
	va_list ap;
	int i;
	static struct {
		char *name;
		int nargs;
		int nreturns;
		char *cmd;
		int status;
		int results[8];
	} args = {
		"interpret",
		1,
		2,
	};

	ofw_stack();
	if (nreturns > 8)
		return -1;
	if ((i = strlen(cmd)) >= NBPG)
		return -1;
	ofbcopy(cmd, OF_buf, i + 1);
	args.cmd = OF_buf;
	args.nargs = 1;
	args.nreturns = nreturns + 1;
	if (openfirmware(&args) == -1)
		return -1;
	va_start(ap, nreturns);
	for (i = 0; i < nreturns; i++)
		*va_arg(ap, int *) = args.results[i];
	va_end(ap);
	return args.status;
}

/*
 * Compare OF "compatible" property.
 */
int
OF_compat(compat, name)
	char *compat, *name;
{
	while (*compat) {
		if (strcmp(compat, name) == 0)
			return 1;
		compat += strlen(compat) + 1;
	}
	return 0;
}

/*
 * Find OF-device corresponding to the PCI device.
 */
int
getnodebypci(pc, tag)
	pci_chipset_tag_t pc;
	pcitag_t tag;
{
	int bus, dev, func;
	u_int reg[5];
	int p, q;
	int l, b, d, f;

	pci_decompose_tag(pc, tag, &bus, &dev, &func);

	for (q = OF_peer(0); q; q = p) {
		l = OF_getprop(q, "assigned-addresses", reg, sizeof(reg));
		if (l > 4) {
			b = (reg[0] >> 16) & 0xff;
			d = (reg[0] >> 11) & 0x1f;
			f = (reg[0] >> 8) & 0x07;

			if (b == bus && d == dev && f == func)
				return q;
		}
		if ((p = OF_child(q)))
			continue;
		while (q) {
			if ((p = OF_peer(q)))
				break;
			q = OF_parent(q);
		}
	}
	return 0;
}

int
getnodebyname(start, target)
	int start;
	const char *target;
{
	int node, next;
	char name[64];

	if (start == 0)
		start = OF_peer(0);

	for (node = start; node; node = next) {
		bzero(name, sizeof name);
		OF_getprop(node, "name", name, sizeof name - 1);
		if (strcmp(name, target) == 0)
			break;

		if ((next = OF_child(node)) != 0)
			continue;
		while (node) {
			if ((next = OF_peer(node)) != 0)
				break;
			node = OF_parent(node);
		}
	}

	return node;
}
