/*	$Id: machdep.c,v 1.41 2004/01/01 14:41:37 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_altivec.h"
#include "opt_compat_netbsd.h"
#include "opt_ddb.h"
#include "opt_ipkdb.h"
#include "opt_multiprocessor.h"
#include "adb.h"
#include "ksyms.h"

#include <sys/param.h>
#include <sys/boot_flag.h>
#include <sys/buf.h>
#include <sys/kernel.h>
#include <sys/ksyms.h>
#include <sys/mount.h>
#include <sys/msgbuf.h>
#include <sys/reboot.h>
#include <sys/sa.h>
#include <sys/sysctl.h>
#include <sys/systm.h>
#include <sys/user.h>

#include <net/netisr.h>

#include <machine/db_machdep.h>
#include <ddb/db_extern.h>

#include <machine/autoconf.h>
#include <machine/bat.h>
#include <machine/powerpc.h>
#include <machine/spr.h>
#include <machine/trap.h>

#include <dev/cons.h>
#include <dev/ofw/openfirm.h>

#include <dev/wscons/wsksymvar.h>
#include <dev/wscons/wscons_callbacks.h>

#include <dev/usb/ukbdvar.h>

#include <macppc/dev/adbvar.h>

struct vm_map *exec_map = NULL;
struct vm_map *mb_map = NULL;
struct vm_map *phys_map = NULL;

/*
 * Global variables used here and there
 */
#ifndef MULTIPROCESSOR
struct pcb *curpcb;
struct pmap *curpm;
struct lwp *fpulwp;
struct lwp *veclwp;
#endif

extern struct user *proc0paddr;
extern int ofmsr;

struct bat battable[16];
char bootpath[256];
paddr_t msgbuf_paddr;
static int chosen;
struct pmap ofw_pmap;
int ofkbd_ihandle;

int msgbufmapped = 0;

#if NKSYMS || defined(DDB) || defined(LKM)
char *startsym, *endsym;
#endif

struct ofw_translations {
	vaddr_t va;
	int len;
	paddr_t pa;
	int mode;
};

void initppc(u_int, u_int, char *);
int ofkbd_cngetc(dev_t);
void cninit_kd(void);
int lcsplx(int);
int save_ofmap(struct ofw_translations *, int);
void restore_ofmap(struct ofw_translations *, int);
static void dumpsys(void);

void
initppc(startkernel, endkernel, args)
	u_int startkernel, endkernel;
	char *args;
{
	extern int trapcode[], alitrap[], dsitrap[], isitrap[];
	extern int decrint[], perftrap[], syscalltrap[];
	extern int tlbimiss[], tlbdlmiss[], tlbdsmiss[];
#ifdef DDB
	extern int ddblow[];
#endif
#ifdef IPKDB
	extern int ipkdblow[];
#endif
	int exc, scratch;
	struct mem_region *allmem, *availmem, *mp;
	struct ofw_translations *ofmap;
	int ofmaplen;
	u_int pvr;
	char bootargs[256];
#ifdef MULTIPROCESSOR
	struct cpu_info *ci = &cpu_info[0];
#else
	struct cpu_info *ci = &cpu_info_store;
#endif

	asm volatile ("mtsprg 0,%0" :: "r"(ci));

	asm volatile ("mfpvr %0" : "=r"(pvr));
	switch (pvr >> 16) {
	case MPC604:
	case MPC604ev:
	case MPC7400:
	case MPC7410:
	case MPC7450:
	case MPC7455:
	case MPC7457:
		asm volatile ("mtspr PIR,%0" :: "r"(0));
	}

	/*
	 * Initialize BAT registers to unmapped to not generate
	 * overlapping mappings below.
	 */
	asm volatile ("mtibatu 0,%0" :: "r"(0));
	asm volatile ("mtibatu 1,%0" :: "r"(0));
	asm volatile ("mtibatu 2,%0" :: "r"(0));
	asm volatile ("mtibatu 3,%0" :: "r"(0));
	asm volatile ("mtdbatu 0,%0" :: "r"(0));
	asm volatile ("mtdbatu 1,%0" :: "r"(0));
	asm volatile ("mtdbatu 2,%0" :: "r"(0));
	asm volatile ("mtdbatu 3,%0" :: "r"(0));
	if (pvr >> 16 == PPC750FX)
		ppc750fx_clearbat();

	/*
	 * Set up BAT0 to only map the lowest 256 MB area
	 */
	battable[0x0].batl = BATL(0x00000000, BATL_M);
	battable[0x0].batu = BATU(0x00000000, BATU_256M);

	/*
	 * Map PCI memory space.
	 */
	battable[0x8].batl = BATL(0x80000000, BATL_I | BATL_G);
	battable[0x8].batu = BATU(0x80000000, BATU_256M);

	battable[0x9].batl = BATL(0x90000000, BATL_I | BATL_G);
	battable[0x9].batu = BATU(0x90000000, BATU_256M);

	battable[0xa].batl = BATL(0xa0000000, BATL_I | BATL_G);
	battable[0xa].batu = BATU(0xa0000000, BATU_256M);

	battable[0xb].batl = BATL(0xb0000000, BATL_I | BATL_G);
	battable[0xb].batu = BATU(0xb0000000, BATU_256M);

	/*
	 * Map obio devices.
	 */
	battable[0xf].batl = BATL(0xf0000000, BATL_I | BATL_G);
	battable[0xf].batu = BATU(0xf0000000, BATU_256M);

	/*
	 * Now setup fixed bat registers
	 *
	 * Note that we still run in real mode, and the BAT
	 * registers were cleared above.
	 */
	/* BAT0 used for initial 256 MB segment */
	asm volatile ("mtibatl 0,%0; mtibatu 0,%1;"
		      "mtdbatl 0,%0; mtdbatu 0,%1;"
		      :: "r"(battable[0].batl), "r"(battable[0].batu));

	/*
	 * Set up battable to map all RAM regions.
	 * This is here because mem_regions() call needs bat0 set up.
	 */
	mem_regions(&allmem, &availmem);
	for (mp = allmem; mp->size; mp++) {
		paddr_t pa = mp->start & 0xf0000000;
		paddr_t end = mp->start + mp->size;

		do {
			u_int n = pa >> 28;

			battable[n].batl = BATL(pa, BATL_M);
			battable[n].batu = BATU(pa, BATU_256M);
			pa += 0x10000000;
		} while (pa < end);
	}

	/* BAT1 used for secondary 256 MB segment (if exists) */
	asm volatile ("mtibatl 1,%0; mtibatu 1,%1;"
		      "mtdbatl 1,%0; mtdbatu 1,%1;"
		      :: "r"(battable[1].batl), "r"(battable[1].batu));

	/*
	 * Use extended BAT size on 7455.
	 */
	switch (pvr >> 16) {

	case MPC7455:
	case MPC7457:
		battable[0].batl = BATL(0x00000000, BATL_M);
		battable[0].batu = BATU(0x00000000, BATU_1G);
		battable[1].batl = BATL(0x00000000, BATL_M);
		battable[1].batu = BATU(0x00000000, BATU_1G);
		battable[2].batl = BATL(0x00000000, BATL_M);
		battable[2].batu = BATU(0x00000000, BATU_1G);
		battable[3].batl = BATL(0x00000000, BATL_M);
		battable[3].batu = BATU(0x00000000, BATU_1G);

		mpc7455_xbsen();
		asm volatile ("mtibatl 0,%0; mtibatu 0,%1;"
			      "mtdbatl 0,%0; mtdbatu 0,%1;"
		    :: "r"(battable[0].batl), "r"(battable[0].batu));
		asm volatile ("mtibatu 1,%0" :: "r"(0));
		asm volatile ("mtdbatu 1,%0" :: "r"(0));
		break;
	}

	chosen = OF_finddevice("/chosen");

	ofmaplen = save_ofmap(NULL, 0);
	ofmap = alloca(ofmaplen);
	save_ofmap(ofmap, ofmaplen);

	lwp0.l_cpu = ci;
	lwp0.l_addr = proc0paddr;
	bzero(lwp0.l_addr, sizeof *lwp0.l_addr);

	curpcb = &proc0paddr->u_pcb;

	curpm = curpcb->pcb_pmreal = curpcb->pcb_pm = pmap_kernel();

#if 0	/* Needs some rethinking regarding real/virtual OFW */
	OF_set_callback(callback);
#endif

	/*
	 * Set up trap vectors
	 */
	for (exc = EXC_RSVD; exc <= EXC_LAST; exc += 0x100)
		bcopy(trapcode, (void *)exc, 0x100);

	bcopy(alitrap, (void *)EXC_ALI, 0x100);
	bcopy(dsitrap, (void *)EXC_DSI, 0x100);
	bcopy(isitrap, (void *)EXC_ISI, 0x100);
	bcopy(decrint, (void *)EXC_DECR, 0x100);
	bcopy(perftrap, (void *)EXC_PERF, 0x100); /* EXC_PERF and EXC_VEC */
	bcopy(tlbimiss, (void *)EXC_IMISS, 0x100);
	bcopy(tlbdlmiss, (void *)EXC_DLMISS, 0x100);
	bcopy(tlbdsmiss, (void *)EXC_DSMISS, 0x100);
	bcopy(syscalltrap, (void *)EXC_SC, 0x100);
#ifdef IPKDB
	bcopy(ipkdblow, (void *)EXC_PGM, 0x100);
	bcopy(ipkdblow, (void *)EXC_TRC, 0x100);
	bcopy(ipkdblow, (void *)EXC_BPT, 0x100);
#endif
#ifdef DDB
	bcopy(ddblow, (void *)EXC_PGM, 0x100);
	bcopy(ddblow, (void *)EXC_TRC, 0x100);
	bcopy(ddblow, (void *)EXC_BPT, 0x100);
#endif

	/*
	 * Cause a PGM trap if we branch to 0.
	 */
	bzero(0, 0x100);

	/*
	 * external interrupt handler install
	 */
	install_extint(ext_intr);

	__syncicache((void *)EXC_RSVD, EXC_LAST - EXC_RSVD + 0x100);

	/*
	 * Now enable translation (and machine checks/recoverable interrupts).
	 */
	asm volatile ("mfmsr %0; ori %0,%0,%1; mtmsr %0; isync"
		      : "=r"(scratch) : "K"(PSL_IR|PSL_DR|PSL_ME|PSL_RI));

	ofmsr &= ~PSL_IP;

	/*
	 * i386 port says, that this shouldn't be here,
	 * but I really think the console should be initialized
	 * as early as possible.
	 */
	consinit();

	/*
	 * Parse arg string.
	 */
#if NKSYMS || defined(DDB) || defined(LKM)
	bcopy(args + strlen(args) + 1, &startsym, sizeof startsym);
	bcopy(args + strlen(args) + 5, &endsym, sizeof endsym);
	if (startsym == NULL || endsym == NULL)
		startsym = endsym = NULL;
#endif

	strcpy(bootpath, args);
	args = strchr(bootpath, ' ');
	if (args)
		*args = 0;

	bzero(bootargs, sizeof bootargs);
	OF_getprop(chosen, "bootargs", bootargs, sizeof bootargs);
	if (bootargs[0] == '-')
		args = bootargs;
	else
		args = strchr(bootargs, ' ');
	if (args) {
		while (*args)
			BOOT_FLAG(*args++, boothowto);
	}

#if NKSYMS || defined(DDB) || defined(LKM)
	ksyms_init(endsym - startsym, startsym, endsym);
#endif
#ifdef DDB
	if (boothowto & RB_KDB)
		Debugger();
#endif
#ifdef IPKDB
	/*
	 * Now trap to IPKDB
	 */
	ipkdb_init();
	if (boothowto & RB_KDB)
		ipkdb_connect(0);
#endif

	/*
	 * Set the page size.
	 */
	uvm_setpagesize();

	/*
	 * Initialize pmap module.
	 */
	pmap_bootstrap(startkernel, endkernel);

	restore_ofmap(ofmap, ofmaplen);
}

int
save_ofmap(ofmap, maxlen)
	struct ofw_translations *ofmap;
	int maxlen;
{
	int mmui, mmu, len;

	OF_getprop(chosen, "mmu", &mmui, sizeof mmui);
	mmu = OF_instance_to_package(mmui);

	if (ofmap) {
		bzero(ofmap, maxlen);	/* to be safe */
		len = OF_getprop(mmu, "translations", ofmap, maxlen);
	} else
		len = OF_getproplen(mmu, "translations");

	return len;
}

void
restore_ofmap(ofmap, len)
	struct ofw_translations *ofmap;
	int len;
{
	int n = len / sizeof(struct ofw_translations);
	int i;

	pmap_pinit(&ofw_pmap);

	ofw_pmap.pm_sr[KERNEL_SR] = KERNEL_SEGMENT;

	for (i = 0; i < n; i++) {
		paddr_t pa = ofmap[i].pa;
		vaddr_t va = ofmap[i].va;
		int len = ofmap[i].len;

		if (va < 0xf0000000)			/* XXX */
			continue;

		while (len > 0) {
			pmap_enter(&ofw_pmap, va, pa, VM_PROT_ALL,
			    VM_PROT_ALL|PMAP_WIRED);
			pa += NBPG;
			va += NBPG;
			len -= NBPG;
		}
	}
	pmap_update(&ofw_pmap);
}

/*
 * This should probably be in autoconf!				XXX
 */
char cpu_model[80];
char machine[] = MACHINE;		/* from <machine/param.h> */
char machine_arch[] = MACHINE_ARCH;	/* from <machine/param.h> */

void
install_extint(handler)
	void (*handler) __P((void));
{
	extern int extint, extsize;
	extern u_long extint_call;
	u_long offset = (u_long)handler - (u_long)&extint_call;
	int omsr, msr;

#ifdef	DIAGNOSTIC
	if (offset > 0x1ffffff)
		panic("install_extint: too far away");
#endif
	asm volatile ("mfmsr %0; andi. %1,%0,%2; mtmsr %1"
		      : "=r"(omsr), "=r"(msr) : "K"((u_short)~PSL_EE));
	extint_call = (extint_call & 0xfc000003) | offset;
	bcopy(&extint, (void *)EXC_EXI, (size_t)&extsize);
	__syncicache((void *)&extint_call, sizeof extint_call);
	__syncicache((void *)EXC_EXI, (int)&extsize);
	asm volatile ("mtmsr %0" :: "r"(omsr));
}

/*
 * Machine dependent startup code.
 */
void
cpu_startup()
{
	caddr_t v;
	vaddr_t minaddr, maxaddr;
	char pbuf[9];

	initmsgbuf((caddr_t)msgbuf_paddr, round_page(MSGBUFSIZE));

	lwp0.l_addr = proc0paddr;
	v = (caddr_t)proc0paddr + USPACE;

	printf("%s", version);
	identifycpu(cpu_model);

	format_bytes(pbuf, sizeof(pbuf), ctob((u_int)physmem));
	printf("total memory = %s\n", pbuf);

	minaddr = 0;

	/*
	 * Allocate a submap for exec arguments.  This map effectively
	 * limits the number of processes exec'ing at any time.
	 */
	exec_map = uvm_km_suballoc(kernel_map, &minaddr, &maxaddr,
				 16*NCARGS, VM_MAP_PAGEABLE, FALSE, NULL);

	/*
	 * Allocate a submap for physio
	 */
	phys_map = uvm_km_suballoc(kernel_map, &minaddr, &maxaddr,
				 VM_PHYS_SIZE, 0, FALSE, NULL);

	/*
	 * No need to allocate an mbuf cluster submap.  Mbuf clusters
	 * are allocated via the pool allocator, and we use direct-mapped
	 * pool pages.
	 */

	format_bytes(pbuf, sizeof(pbuf), ptoa(uvmexp.free));
	printf("avail memory = %s\n", pbuf);
}

/*
 * consinit
 * Initialize system console.
 */
void
consinit()
{
	static int initted;

	if (initted)
		return;
	initted = 1;
	cninit();
}

/*
 * Machine dependent system variables.
 */
#if 0
static int
sysctl_machdep_cacheinfo(SYSCTLFN_ARGS)
{
	struct sysctlnode node = *rnode;
	static struct cache_info ci = {
		CACHELINESIZE, CACHELINESIZE,		/* XXX cache_size */
		CACHELINESIZE, CACHELINESIZE,
	};

	node.sysctl_data = &ci;
	node.sysctl_size = sizeof ci;
	return sysctl_lookup(SYSCTLFN_CALL(&node));
}
#endif

static int
sysctl_machdep_booted_kernel(SYSCTLFN_ARGS)
{
	struct sysctlnode node = *rnode;
	char *cp;

	cp = strrchr(bootpath, '/');
	if (cp == NULL)
		cp = bootpath;
	if (*cp == '/')
		cp++;
	node.sysctl_data = cp;
	node.sysctl_size = sizeof bootpath;
	return sysctl_lookup(SYSCTLFN_CALL(&node));
}

SYSCTL_SETUP(sysctl_machdep_setup, "sysctl machdep subtree setup")
{
	u_int pvr;

	sysctl_createv(SYSCTL_PERMANENT,
		       CTLTYPE_NODE, "machdep", NULL,
		       NULL, 0, NULL, 0,
		       CTL_MACHDEP, CTL_EOL);

	sysctl_createv(SYSCTL_PERMANENT|SYSCTL_IMMEDIATE,
		       CTLTYPE_INT, "cachelinesize", NULL,
		       NULL, CACHELINESIZE, NULL, 0,
		       CTL_MACHDEP, CPU_CACHELINE, CTL_EOL);
#if 0
	sysctl_createv(SYSCTL_PERMANENT,
		       CTLTYPE_STRUCT, "cacheinfo", NULL,
		       sysctl_machdep_cacheinfo, 0, NULL, 0,
		       CTL_MACHDEP, CPU_CACHEINFO, CTL_EOL);
#endif
	sysctl_createv(SYSCTL_PERMANENT,
		       CTLTYPE_STRING, "booted_kernel", NULL,
		       sysctl_machdep_booted_kernel, 0, NULL, 0,
		       CTL_MACHDEP, CPU_BOOTED_KERNEL, CTL_EOL);


	asm volatile ("mfpvr %0" : "=r"(pvr));
	switch (pvr >> 16) {
#ifdef ALTIVEC
	case MPC7400:
	case MPC7410:
	case MPC7450:
	case MPC7455:
	case MPC7457:
		sysctl_createv(SYSCTL_PERMANENT|SYSCTL_IMMEDIATE,
		       CTLTYPE_INT, "altivec", NULL,
		       NULL, 1, NULL, 0,
		       CTL_MACHDEP, CPU_ALTIVEC, CTL_EOL);
		break;
#endif
	default:
		sysctl_createv(SYSCTL_PERMANENT|SYSCTL_IMMEDIATE,
		       CTLTYPE_INT, "altivec", NULL,
		       NULL, 0, NULL, 0,
		       CTL_MACHDEP, CPU_ALTIVEC, CTL_EOL);
	}
}

/*
 * Crash dump handling.
 */

void
dumpsys()
{
	printf("dumpsys: TBD\n");
}

/*
 * Soft networking interrupts.
 */
void
softnet()
{
	int isr;

	isr = netisr;
	netisr = 0;

#define DONETISR(bit, fn) do {		\
	if (isr & (1 << bit))		\
		fn();			\
} while (0)

#include <net/netisr_dispatch.h>

#undef DONETISR

}

#include "zsc.h"
#include "com.h"

int zssoft __P((void *));
void comsoft __P((void));

/*
 * Soft tty interrupts.
 */
void
softserial()
{
#if NZSC > 0
	zssoft(NULL);
#endif
#if NCOM > 0
	comsoft();
#endif
}

#if 0
/*
 * Stray interrupts.
 */
void
strayintr(irq)
	int irq;
{
	log(LOG_ERR, "stray interrupt %d\n", irq);
}
#endif

/*
 * Halt or reboot the machine after syncing/dumping according to howto.
 */
void
cpu_reboot(howto, what)
	int howto;
	char *what;
{
	static int syncing;
	static char str[256];
	char *ap = str, *ap1 = ap;

	boothowto = howto;
	if (!cold && !(howto & RB_NOSYNC) && !syncing) {
		syncing = 1;
		vfs_shutdown();		/* sync */
		resettodr();		/* set wall clock */
	}

#ifdef MULTIPROCESSOR
	/* Halt other CPU.  XXX for now... */
	macppc_send_ipi(&cpu_info[1 - cpu_number()], MACPPC_IPI_HALT);
	delay(100000);	/* XXX */
#endif

	splhigh();

	if (!cold && (howto & RB_DUMP))
		dumpsys();

	doshutdownhooks();

	if ((howto & RB_POWERDOWN) == RB_POWERDOWN) {
#if NADB > 0
		delay(1000000);
		adb_poweroff();
		printf("WARNING: powerdown failed!\n");
#endif
	}

	if (howto & RB_HALT) {
		printf("halted\n\n");

		/* flush cache for msgbuf */
		__syncicache((void *)msgbuf_paddr, round_page(MSGBUFSIZE));

		ppc_exit();
	}

	printf("rebooting\n\n");
	if (what && *what) {
		if (strlen(what) > sizeof str - 5)
			printf("boot string too large, ignored\n");
		else {
			strcpy(str, what);
			ap1 = ap = str + strlen(str);
			*ap++ = ' ';
		}
	}
	*ap++ = '-';
	if (howto & RB_SINGLE)
		*ap++ = 's';
	if (howto & RB_KDB)
		*ap++ = 'd';
	*ap++ = 0;
	if (ap[-2] == '-')
		*ap1 = 0;

	/* flush cache for msgbuf */
	__syncicache((void *)msgbuf_paddr, round_page(MSGBUFSIZE));

#if NADB > 0
	adb_restart();	/* not return */
#endif
	ppc_exit();
}

#if 0
/*
 * OpenFirmware callback routine
 */
void
callback(p)
	void *p;
{
	panic("callback");	/* for now			XXX */
}
#endif

int
lcsplx(ipl)
	int ipl;
{
	return spllower(ipl); 	/* XXX */
}

/*
 * Allocate vm space and mapin the I/O address
 */
void *
mapiodev(pa, len)
	paddr_t pa;
	psize_t len;
{
	paddr_t faddr;
	vaddr_t taddr, va;
	int off;

	faddr = trunc_page(pa);
	off = pa - faddr;
	len = round_page(off + len);
	va = taddr = uvm_km_valloc(kernel_map, len);

	if (va == 0)
		return NULL;

	for (; len > 0; len -= NBPG) {
		pmap_kenter_pa(taddr, faddr, VM_PROT_READ | VM_PROT_WRITE);
		faddr += NBPG;
		taddr += NBPG;
	}
	pmap_update(pmap_kernel());
	return (void *)(va + off);
}

#include "akbd.h"
#include "ukbd.h"
#include "ofb.h"
#include "zstty.h"

void
cninit()
{
	struct consdev *cp;
	int stdout, node;
	char type[16];

	if (OF_getprop(chosen, "stdout", &stdout, sizeof(stdout))
	    != sizeof(stdout))
		goto nocons;

	node = OF_instance_to_package(stdout);
	bzero(type, sizeof(type));
	if (OF_getprop(node, "device_type", type, sizeof(type)) == -1)
		goto nocons;

#if NOFB > 0
	if (strcmp(type, "display") == 0) {
		cninit_kd();
		return;
	}
#endif /* NOFB > 0 */

#if NZSTTY > 0
	if (strcmp(type, "serial") == 0) {
		extern struct consdev consdev_zs;

		cp = &consdev_zs;
		(*cp->cn_probe)(cp);
		(*cp->cn_init)(cp);
		cn_tab = cp;

		return;
	}
#endif

nocons:
	return;
}

#if NOFB > 0
void
cninit_kd()
{
	int stdin, node;
	char name[16];

	/*
	 * Attach the console output now (so we can see debugging messages,
	 * if any).
	 */
	ofb_cnattach();

	/*
	 * We must determine which keyboard type we have.
	 */
	if (OF_getprop(chosen, "stdin", &stdin, sizeof(stdin))
	    != sizeof(stdin)) {
		printf("WARNING: no `stdin' property in /chosen\n");
		return;
	}

	node = OF_instance_to_package(stdin);
	bzero(name, sizeof(name));
	OF_getprop(node, "name", name, sizeof(name));
	if (strcmp(name, "keyboard") != 0) {
		printf("WARNING: stdin is not a keyboard: %s\n", name);
		return;
	}

#if NAKBD > 0
	bzero(name, sizeof(name));
	OF_getprop(OF_parent(node), "name", name, sizeof(name));
	if (strcmp(name, "adb") == 0) {
		printf("console keyboard type: ADB\n");
		akbd_cnattach();
		goto kbd_found;
	}
#endif

	/*
	 * We're not an ADB keyboard; must be USB.  Unfortunately,
	 * we have a few problems:
	 *
	 *	(1) The stupid Macintosh firmware uses a
	 *	    `psuedo-hid' (yes, they even spell it
	 *	    incorrectly!) which apparently merges
	 *	    all USB keyboard input into a single
	 *	    input stream.  Because of this, we can't
	 *	    actually determine which USB controller
	 *	    or keyboard is really the console keyboard!
	 *
	 *	(2) Even if we could, USB requires a lot of
	 *	    the kernel to be running in order for it
	 *	    to work.
	 *
	 * So, what we do is this:
	 *
	 *	(1) Tell the ukbd driver that it is the console.
	 *	    At autoconfiguration time, it will attach the
	 *	    first USB keyboard instance as the console
	 *	    keyboard.
	 *
	 *	(2) Until then, so that we have _something_, we
	 *	    use the OpenFirmware I/O facilities to read
	 *	    the keyboard.
	 */

	/*
	 * stdin is /psuedo-hid/keyboard.
	 */

#if NUKBD > 0
	/*
	 * Assume USB keyboard if "/.../usb/.../keyboard" exists.
	 */
	node = getnodebyname(0, "usb");
	while (node) {
		bzero(name, sizeof name);
		OF_getprop(node, "name", name, sizeof name);
		if (strcmp(name , "usb") == 0 &&
		    getnodebyname(node, "keyboard")) {
			printf("console keyboard type: USB\n");
			ukbd_cnattach();
			goto kbd_found;
		}
		node = OF_peer(node);
	}
#endif

#if NAKBD > 0
	/*
	 * Assume ADB keyboard if "/.../adb/keyboard" exists.
	 */
	node = getnodebyname(0, "adb");
	if (node && getnodebyname(node, "keyboard")) {
		printf("console keyboard type: ADB\n");
		akbd_cnattach();
		goto kbd_found;
	}
#endif

	/*
	 * No keyboard is found.  Just return.
	 */
	printf("no console keyboard\n");
	return;

#if NAKBD + NUKBD > 0
kbd_found:
	/*
	 * XXX This is a little gross, but we don't get to call
	 * XXX wskbd_cnattach() twice.
	 */
	ofkbd_ihandle = stdin;
	wsdisplay_set_cons_kbd(ofkbd_cngetc, NULL, NULL);
#endif
}
#endif

/*
 * Bootstrap console keyboard routines, using OpenFirmware I/O.
 */
int
ofkbd_cngetc(dev)
	dev_t dev;
{
	u_char c = '\0';
	int len;

	do {
		len = OF_read(ofkbd_ihandle, &c, 1);
	} while (len != 1);

	return c;
}

#ifdef MULTIPROCESSOR
#include <powerpc/fpu.h>

void
save_fpu_lwp(l)
	struct lwp *l;
{
	volatile struct cpu_info *fpcpu;
	int i;
	extern volatile int IPI[];	/* XXX */

	fpcpu = l->l_addr->u_pcb.pcb_fpcpu;
	if (fpcpu == curcpu()) {
		save_fpu(l);
		return;
	}

#if 0
	printf("save_fpu_lwp{%d} pid = %d, lid = %d, fpcpu->ci_cpuid = %d\n",
	    cpu_number(), l->l_proc->p_pid, l->l_lid, fpcpu->ci_cpuid);
#endif

	macppc_send_ipi(fpcpu, MACPPC_IPI_FLUSH_FPU);

	/* Wait for flush. */
#if 0
	while (fpcpu->ci_fpulwp);
#else
	for (i = 0; i < 0x3fffffff; i++) {
		if (fpcpu->ci_fpulwp == NULL)
			goto done;
	}
	printf("save_fpu_lwp{%d} timed out\n", cpu_number());
	for (i = 0; i < 0x3fffffff; i++) {
		if (fpcpu->ci_fpulwp == NULL)
			goto done;
	}
	printf("save_fpu_lwp{%d} pid = %d, lid = %d, fpcpu->ci_cpuid = %d\n",
	    cpu_number(), l->l_proc->p_pid, l->l_lid, fpcpu->ci_cpuid);
	printf("IPI[0] = 0x%x, IPI[1] = 0x%x\n", IPI[0], IPI[1]);
	printf("cpl 0x%x 0x%x\n", cpu_info[0].ci_cpl, cpu_info[1].ci_cpl);
	printf("ipending 0x%x 0x%x\n", cpu_info[0].ci_ipending, cpu_info[1].ci_ipending);
	panic("save_fpu_lwp");
done:;

#endif
}

#ifdef ALTIVEC
#include <powerpc/altivec.h>

void
save_vec_lwp(l)
	struct lwp *l;
{
	volatile struct cpu_info *veccpu;
	int i;
	extern volatile int IPI[];	/* XXX */

	veccpu = l->l_addr->u_pcb.pcb_veccpu;
	if (veccpu == curcpu()) {
		save_vec(l);
		return;
	}

	macppc_send_ipi(veccpu, MACPPC_IPI_FLUSH_VEC);

	/* Wait for flush. */
#if 0
	while (veccpu->ci_veclwp);
#else
	for (i = 0; i < 0x3fffffff; i++) {
		if (veccpu->ci_veclwp == NULL)
			goto done;
	}
	printf("save_vec_lwp{%d} timed out\n", cpu_number());
	for (i = 0; i < 0x3fffffff; i++) {
		if (veccpu->ci_veclwp == NULL)
			goto done;
	}
	printf("save_vec_lwp{%d} pid = %d, lid = %d, veccpu->ci_cpuid = %d\n",
	    cpu_number(), l->l_proc->p_pid, l->l_lid, veccpu->ci_cpuid);
	printf("IPI[0] = 0x%x, IPI[1] = 0x%x\n", IPI[0], IPI[1]);
	printf("cpl 0x%x 0x%x\n", cpu_info[0].ci_cpl, cpu_info[1].ci_cpl);
	printf("ipending 0x%x 0x%x\n", cpu_info[0].ci_ipending, cpu_info[1].ci_ipending);
	panic("save_vec_lwp");
done:;

#endif
}
#endif /* ALTIVEC */
#endif /* MULTIPROCESSOR */
