/*	$Id: hfslabel.c,v 1.2 2003/05/21 16:38:58 tsubai Exp $	*/

/*-
 * Copyright (c) 2001-2003 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/disklabel.h>
#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "extern.h"

static void write_apm(int);

u_int hfs_offset, hfs_size;
int hfs_part;

int
hfslabel(argc, argv)
	int argc;
	char **argv;
{
	int fd;
	int auto_search = 0;
	struct disklabel *label;
	char hfsdev[32], rawdev[32], disk[32];
	char labelbuf[512];

	if (argc - optind < 1)
		usage();

	if (argv[optind][0] == '/')
		strcpy(hfsdev, argv[optind]);
	else
		sprintf(hfsdev, "/dev/r%s", argv[optind]);

	strcpy(disk, hfsdev);
	fd = open(hfsdev, 0);
	if (fd == -1) {
		sprintf(hfsdev, "%s%c", disk, 'a' + RAW_PART);
		fd = open(hfsdev, 0);
		if (fd == -1)
			err(1, "open HFS");
		auto_search = 1;
		hfs_part = -1;
	} else {
		disk[strlen(disk) - 1] = 0;
		hfs_part = hfsdev[strlen(hfsdev) - 1] - 'a';
	}
	close(fd);

	sprintf(rawdev, "%s%c", disk, 'a' + RAW_PART);
	fd = open(rawdev, 2);
	if (fd == -1)
		err(1, "open raw");

	bzero(labelbuf, sizeof labelbuf);
	devread(fd, labelbuf, LABELSECTOR, sizeof labelbuf, "hfslabel");
	label = (void *)&labelbuf[LABELOFFSET];
	if (label->d_magic != DISKMAGIC) {
		printf("bad label magic\n");
		exit(1);
	}

	if (auto_search) {
		int part;

		for (part = 0; part < MAXPARTITIONS; part++) {
			if (label->d_partitions[part].p_fstype == FS_HFS) {
				hfs_part = part;
				break;
			}
		}
	}
	if (hfs_part == -1) {
		printf("no HFS partition\n");
		exit(1);
	}

	hfs_offset = label->d_partitions[hfs_part].p_offset;
	hfs_size = label->d_partitions[hfs_part].p_size;

	if (verbose) {
		printf("hfsdev: %s%c\n", disk, 'a' + hfs_part);
		printf("rawdev: %s\n", rawdev);
		printf("HFS offset: %d\n", hfs_offset);
		printf("HFS size  : %d\n", hfs_size);
	}

	if (!nowrite)
		write_apm(fd);

	exit(0);
}

void
write_apm(fd)
	int fd;
{
	struct drvr_map block0;
	struct part_map_entry block1, block2;

	bzero(&block0, 512);
	bzero(&block1, 512);
	bzero(&block2, 512);

	devread(fd, &block0, 0, 512, "write_apm");

	if (lseek(fd, 0, SEEK_SET) != 0)
		err(1, "write_apm: lseek");

	/* block 0 */
	block0.sbSig = DRIVER_MAP_MAGIC;
	block0.sbBlockSize = 512;
	block0.sbBlkCount = 0;		/* XXX */

	/* block 1: Apple Partition Map */
	block1.pmSig = PART_ENTRY_MAGIC;
	block1.pmMapBlkCnt = 2;
	block1.pmPyPartStart = 1;
	block1.pmPartBlkCnt = block1.pmDataCnt = 2;
	strcpy(block1.pmPartName, "Apple");
	strcpy(block1.pmPartType, "Apple_partition_map");
	block1.pmPartStatus = 0x37;

	/* block 2: NetBSD partition */
	block2.pmSig = PART_ENTRY_MAGIC;
	block2.pmMapBlkCnt = 2;
	block2.pmPyPartStart = hfs_offset;
	block2.pmPartBlkCnt = block2.pmDataCnt = hfs_size;
	strcpy(block2.pmPartName, "NetBSD");
	strcpy(block2.pmPartType, "Apple_HFS");
	block2.pmPartStatus = 0x3b;

	if (verbose)
		printf("writing HFS label...");
	write(fd, &block0, 512);
	write(fd, &block1, 512);
	write(fd, &block2, 512);
	if (verbose)
		printf("done\n");
}
