/*	$Id: i2c.c,v 1.4 2002/06/13 15:43:14 tsubai Exp $	*/

/*-
 * Copyright (c) 2001,2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>

#include <machine/mpc860dev.h>
#include <machine/sysconf.h>

#define DPRAM_I2C_BASE	((char *)0xff002850)	/* cpm_dpram[0x850] */

#define I2C_READ	1
#define I2C_WRITE	0

int
i2c_read(slave, offset, len, buf)
	int slave, offset, len;
	void *buf;
{
	volatile struct mpc860dev *immr = get_immr();
	volatile struct i2c_pram *i2c;
	volatile struct cpm_bdesc *txbd, *rxbd;

#ifdef DIAGNOSTIC
	if ((int)buf & 1)
		panic("i2c_read: odd receive buffer");
#endif
	flushcache(buf, len);

	rxbd = (void *)DPRAM_I2C_BASE;
	rxbd[0].bd_flags = 0;
	rxbd[1].bd_flags = 0;

	txbd = (void *)(DPRAM_I2C_BASE + 0x10);
	txbd[0].bd_flags = 0;
	txbd[0].bd_addr = DPRAM_I2C_BASE + 0x20;
	txbd[1].bd_flags = 0;
	txbd[1].bd_addr = DPRAM_I2C_BASE + 0x30;

	/* Setup port B pin 26 and 27 for I2C. */
	immr->cpm_pbpar |= 0x00000030;
	immr->cpm_pbdir |= 0x00000030;
	immr->cpm_pbodr |= 0x00000030;

	/* CLK = SYSCLK / PDIV / (2 * (DIV + 3)) */
	immr->cpm_i2mod = I2MOD_PDIV_32;
	immr->cpm_i2brg = 5;

	immr->cpm_i2com = I2COM_MASTER;
	immr->cpm_i2mod = I2MOD_PDIV_32 | I2MOD_EN;
	immr->cpm_i2add = 0;
	immr->cpm_i2cmr = 0;			/* No interrupts */
	immr->cpm_i2cer = 0xff;			/* Clear all events. */

	i2c = (void *)(immr->cpm_pram1 + 0x80);
	i2c->rbase = (int)rxbd & 0xffff;
	i2c->tbase = (int)txbd & 0xffff;
	i2c->rfcr = 0x10;
	i2c->tfcr = 0x10;
	i2c->mrblr = len;

	/* Issue an INIT TX AND RX command. */
	immr->cpm_cpcr = 0x0011;
	while (immr->cpm_cpcr & 1);		/* Wait for completion. */

	/* Write addr */
	txbd[0].bd_addr[0] = slave | I2C_WRITE;	/* Slave address | R/W. */
	txbd[0].bd_addr[1] = offset;
	txbd[0].bd_len = 2;
	txbd[0].bd_flags = TxBD_R | TxBD_L;
	txbd[1].bd_flags = TxBD_W;

	immr->cpm_i2com |= I2COM_STR;		/* Start transfer. */
	while (txbd[0].bd_flags & TxBD_R);

	/* Read */
	txbd[1].bd_addr[0] = slave | I2C_READ;	/* Slave address | R/W. */
	txbd[1].bd_len = len + 1;
	txbd[1].bd_flags = TxBD_R | TxBD_W | TxBD_L;

	rxbd[0].bd_flags = RxBD_E;
	rxbd[0].bd_addr = buf;
	rxbd[1].bd_flags = RxBD_W;

	immr->cpm_i2com |= I2COM_STR;		/* Start transfer. */
	while (rxbd[0].bd_flags & RxBD_E);

	/* XXX error check */

	return 0;
}
