/*	$Id: if_fec.c,v 1.24 2002/10/06 13:18:56 tsubai Exp $	*/

/*-
 * Copyright (c) 2001,2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_inet.h"
#include "bpfilter.h"

#include <sys/param.h>
#include <sys/device.h>
#include <sys/ioctl.h>
#include <sys/kernel.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/systm.h>

#include <uvm/uvm_extern.h>

#include <net/if.h>
#include <net/if_ether.h>
#include <net/if_media.h>

#include <machine/autoconf.h>
#include <machine/mpc860dev.h>
#include <machine/sysconf.h>

#if NBPFILTER > 0
#include <net/bpf.h>
#endif

#include <dev/mii/mii.h>
#include <dev/mii/miivar.h>

#define FEC_NRXBUF	8
#define FEC_NTXBUF	2

#ifdef FEC_DEBUG
#define DPRINTF printf
#else
#define DPRINTF while (0) printf
#endif

#define FEC_ECNTRL_PINMUX	0x00000004	/* FEC enable */
#define FEC_ECNTRL_ETHER_EN	0x00000002	/* Ethernet enable */
#define FEC_ECNTRL_RESET	0x00000001

#define FEC_RCNTRL_BC_REJ	0x00000010	/* Broadcast frame reject */
#define FEC_RCNTRL_PROM		0x00000008	/* Promiscuous mode */
#define FEC_RCNTRL_MII_MODE	0x00000004	/* MII mode */
#define FEC_RCNTRL_DRT		0x00000002	/* Disable recv on transmit */
#define FEC_RCNTRL_LOOP		0x00000001	/* Loopback */

#define FEC_XCNTRL_FDEN		0x00000004	/* Full duplex */
#define FEC_XCNTRL_HBC		0x00000002	/* Heartbeat control */
#define FEC_XCNTRL_GTS		0x00000001	/* Graceful trasmit stop */

#define FEC_IEVENT_HBERR	0x80000000
#define FEC_IEVENT_BABR		0x40000000
#define FEC_IEVENT_BABT		0x20000000
#define FEC_IEVENT_GRA		0x10000000
#define FEC_IEVENT_TFINT	0x08000000
#define FEC_IEVENT_TXB		0x04000000
#define FEC_IEVENT_RFINT	0x02000000
#define FEC_IEVENT_RXB		0x01000000
#define FEC_IEVENT_MII		0x00800000
#define FEC_IEVENT_EBERR	0x00400000
#define FEC_IEVENT_BITMASK "\20" \
    "\40HBERR\37BABR\36BABT\35GRA\34TFINT\33TXB\32RFINT\31RXB\30MII\27EBERR"

#define sc_if sc_ec.ec_if

struct fec_softc {
	struct device sc_dev;
	struct ethercom sc_ec;
	struct mii_data sc_mii;

	int sc_vec;
	char *sc_rxbuf;
	char *sc_txbuf;
	struct cpm_bdesc *sc_rxbd;
	struct cpm_bdesc *sc_next_rxbd;
	struct cpm_bdesc *sc_txbd;
	struct cpm_bdesc *sc_next_txbd;
};

extern char fec_enaddr[];

int fec_match(struct device *, struct cfdata *, void *);
void fec_attach(struct device *, struct device *, void *);

int fecintr(void *);
void fec_tint(struct fec_softc *);
void fec_rint(struct fec_softc *);

void fec_start(struct ifnet *);
int fec_ioctl(struct ifnet *, u_long, caddr_t);
void fec_watchdog(struct ifnet *);
int fec_init(struct ifnet *);
void fec_stop(struct ifnet *, int);
void fec_reset(struct fec_softc *);
void fec_setladrf(struct fec_softc *);

struct mbuf *fec_get(struct fec_softc *, caddr_t, int);
int fec_put(struct fec_softc *, caddr_t, struct mbuf *);

int fec_mii_readreg(struct device *, int, int);
void fec_mii_writereg(struct device *, int, int, int);
void fec_mii_statchg(struct device *);

int fec_mediachange(struct ifnet *);
void fec_mediastatus(struct ifnet *, struct ifmediareq *);


struct cfattach fec_ca = {
	"fec", {}, sizeof(struct fec_softc), fec_match, fec_attach
};

int
fec_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;

	if (strcmp(ca->ca_name, "fec") == 0)
		return 1;
	return 0;
}

void
fec_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct fec_softc *sc = (void *)self;
	struct cfdata *cf = self->dv_cfdata;
	int vec = cf->cf_vec;
	char *buf;
	struct ifnet *ifp = &sc->sc_if;
	struct mii_data *mii = &sc->sc_mii;
	struct cpm_bdesc *bd;
	int bufsize = 0;
	int i;

	bufsize += 0x800 * (FEC_NRXBUF + FEC_NTXBUF);
	bufsize += sizeof(struct cpm_bdesc) * (FEC_NRXBUF + FEC_NTXBUF);
	bufsize += 15;

	buf = malloc(bufsize, M_DEVBUF, M_NOWAIT | M_ZERO);
	if (buf == NULL) {
		printf(": cannot malloc\n");
		return;
	}
	buf = (void *)roundup((u_int)buf, 16);

	bd = (void *)(buf + 0x800 * (FEC_NRXBUF + FEC_NTXBUF));

	sc->sc_rxbd = sc->sc_next_rxbd = bd;
	sc->sc_rxbuf = buf;
	for (i = 0; i < FEC_NRXBUF; i++) {
		bd->bd_flags = RxBD_E;
		bd->bd_addr = (char *)vtophys((vaddr_t)buf);
		bd++;
		buf += 0x800;
	}
	(bd - 1)->bd_flags |= RxBD_W;

	sc->sc_txbd = sc->sc_next_txbd = bd;
	sc->sc_txbuf = buf;
	for (i = 0; i < FEC_NTXBUF; i++) {
		bd->bd_flags = 0;
		bd->bd_addr = (char *)vtophys((vaddr_t)buf);
		bd++;
		buf += 0x800;
	}
	(bd - 1)->bd_flags |= TxBD_W;

	sc->sc_vec = vec;

	printf(": vec %d", vec);
	printf(" address %s\n", ether_sprintf(fec_enaddr));

	bcopy(sc->sc_dev.dv_xname, ifp->if_xname, IFNAMSIZ);
	ifp->if_softc = sc;
	ifp->if_ioctl = fec_ioctl;
	ifp->if_start = fec_start;
	ifp->if_watchdog = fec_watchdog;
	ifp->if_init = fec_init;
	ifp->if_stop = fec_stop;
	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
	IFQ_SET_READY(&ifp->if_snd);

	fec_init(ifp);

	mii->mii_ifp = ifp;
	mii->mii_readreg = fec_mii_readreg;
	mii->mii_writereg = fec_mii_writereg;
	mii->mii_statchg = fec_mii_statchg;

	ifmedia_init(&mii->mii_media, 0, fec_mediachange, fec_mediastatus);
#ifdef openblocks
	/* XXX addr 0 is wildcard on tqphy */
	mii_attach(self, mii, 0xffffffff, 1, MII_OFFSET_ANY, 0);
#else
	mii_attach(self, mii, 0xffffffff, MII_PHY_ANY, MII_OFFSET_ANY, 0);
#endif

	/* Choose a default media. */
	if (LIST_FIRST(&mii->mii_phys))
		ifmedia_set(&mii->mii_media, IFM_ETHER|IFM_AUTO);
	else {
		ifmedia_add(&mii->mii_media, IFM_ETHER|IFM_NONE, 0, NULL);
		ifmedia_set(&mii->mii_media, IFM_ETHER|IFM_NONE);
	}
	if_attach(ifp);
	ether_ifattach(ifp, fec_enaddr);

	intr_establish(vec, IST_LEVEL, IPL_NET, fecintr, sc);
}

static inline void
eieio()
{
	asm volatile ("eieio; sync");
}

void
fec_reset(sc)
	struct fec_softc *sc;
{
	volatile struct mpc860dev *immr = get_immr();
	struct ifnet *ifp = &sc->sc_if;
	u_char *ea = fec_enaddr;
	struct cpm_bdesc *bd;
	int i;

	/* Disable ethernet. */
	immr->fec_ecntrl = FEC_ECNTRL_PINMUX;

	/* Reset ethernet. */
	immr->fec_ecntrl = FEC_ECNTRL_PINMUX | FEC_ECNTRL_RESET;
	delay(10);
	while (immr->fec_ecntrl & FEC_ECNTRL_RESET); /* Wait for reset done. */

	immr->fec_imask = FEC_IEVENT_TFINT | FEC_IEVENT_RFINT | FEC_IEVENT_MII;
	immr->fec_ievent = 0xffffffff;		/* Clear all events. */
	immr->fec_ivec = (sc->sc_vec << 28) & 0xe0000000;
	immr->fec_addr_high = ea[0] << 24 | ea[1] << 16 | ea[2] << 8 | ea[3];
	immr->fec_addr_low =  ea[4] << 24 | ea[5] << 16;
	immr->fec_hash_table_high = 0;
	immr->fec_hash_table_low = 0;
	immr->fec_r_buff_size = 0x5f0;
	immr->fec_r_des_start = (void *)vtophys((vaddr_t)sc->sc_rxbd);
	immr->fec_x_des_start = (void *)vtophys((vaddr_t)sc->sc_txbd);
	/* XXX half-duplex */
	immr->fec_r_cntrl = FEC_RCNTRL_MII_MODE | FEC_RCNTRL_DRT;
	if (ifp->if_flags & IFF_PROMISC)
		immr->fec_r_cntrl |= FEC_RCNTRL_PROM;
	immr->fec_x_cntrl = 0;
	immr->fec_fun_code = 0x7f000000;
	immr->fec_mii_speed = howmany(SYSCLK/2, 2500000) << 1;
	immr->cpm_pdpar = 0x1fff;
	immr->cpm_pddir = 0x1fff;
	immr->fec_ecntrl = FEC_ECNTRL_PINMUX;
	eieio();

	flushcache(sc->sc_rxbuf, FEC_NRXBUF * 0x800);
	bd = sc->sc_rxbd;
	for (i = 0; i < FEC_NRXBUF; i++) {
		bd->bd_flags &= RxBD_W;
		bd->bd_flags |= RxBD_E;
		bd++;
	}
	bd = sc->sc_txbd;
	for (i = 0; i < FEC_NTXBUF; i++) {
		bd->bd_flags &= TxBD_W;
		bd++;
	}
	sc->sc_next_rxbd = sc->sc_rxbd;
	sc->sc_next_txbd = sc->sc_txbd;
}

void
fec_setladrf(sc)
	struct fec_softc *sc;
{
	volatile struct mpc860dev *immr = get_immr();
	struct ifnet *ifp = &sc->sc_if;
	struct ether_multi *enm;
	struct ether_multistep step;
	struct ethercom *ec = &sc->sc_ec;
	u_int32_t crc, hash[2];

	/* Clear hash table */
	bzero(hash, sizeof hash);

	if (ifp->if_flags & IFF_PROMISC) {
		/* Turn on promiscuous mode; turn off the hash filter */
		ifp->if_flags |= IFF_ALLMULTI;
		goto chipit;
	}

	/*
	 * Set up multicast address filter by passing all multicast addresses
	 * through a crc generator, and then using the high order 8 bits as an
	 * index into the 256 bit logical address filter.  The high order bit
	 * selects the word, while the rest of the bits select the bit within
	 * the word.
	 */

	ETHER_FIRST_MULTI(step, ec, enm);
	while (enm) {
		if (bcmp(enm->enm_addrlo, enm->enm_addrhi, 6)) {
			/*
			 * We must listen to a range of multicast addresses.
			 * For now, just accept all multicasts, rather than
			 * trying to set only those filter bits needed to match
			 * the range.  (At this time, the only use of address
			 * ranges is for IP multicast routing, for which the
			 * range is big enough to require all bits set.)
			 */
			hash[0] = hash[1] = 0xffffffff;
			ifp->if_flags |= IFF_ALLMULTI;
			goto chipit;
		}

		crc = ether_crc32_le(enm->enm_addrlo, ETHER_ADDR_LEN);

		/* Just want the 6 most significant bits. */
		crc >>= 26;

		/* Set the corresponding bit in the filter. */
		hash[crc >> 5] |= 1 << (crc & 0x1f);

		ETHER_NEXT_MULTI(step, enm);
	}

	ifp->if_flags &= ~IFF_ALLMULTI;

chipit:
	/* Now load the hash table into the chip */
	immr->fec_hash_table_high = hash[1];
	immr->fec_hash_table_low = hash[0];
}

int
fec_init(ifp)
	struct ifnet *ifp;
{
	volatile struct mpc860dev *immr = get_immr();
	struct fec_softc *sc = ifp->if_softc;

	fec_reset(sc);
	fec_setladrf(sc);

	/* Enable ethernet */
	immr->fec_ecntrl = FEC_ECNTRL_PINMUX | FEC_ECNTRL_ETHER_EN;

	/* Enable receiver. */
	immr->fec_r_des_active = 0x100;
	eieio();

	ifp->if_flags |= IFF_RUNNING;
	ifp->if_flags &= ~IFF_OACTIVE;

	mii_pollstat(&sc->sc_mii);

	return 0;
}

int
fecintr(v)
	void *v;
{
	struct fec_softc *sc = v;
	volatile struct mpc860dev *immr = get_immr();
	int event = immr->fec_ievent;
	char bitmask[64];

	immr->fec_ievent = event;
	bitmask_snprintf(event, FEC_IEVENT_BITMASK, bitmask, sizeof bitmask);
	DPRINTF("fecintr event = %s\n", bitmask);
	if (event & FEC_IEVENT_RFINT)
		fec_rint(sc);
	if (event & FEC_IEVENT_TFINT)
		fec_tint(sc);
	if (event & FEC_IEVENT_MII)
		mii_tick(&sc->sc_mii);

	return 1;
}

void
fec_tint(sc)
	struct fec_softc *sc;
{
	struct ifnet *ifp = &sc->sc_if;

	ifp->if_flags &= ~IFF_OACTIVE;
	ifp->if_timer = 0;
	fec_start(ifp);
}

void
fec_start(ifp)
	struct ifnet *ifp;
{
	volatile struct mpc860dev *immr = get_immr();
	struct cpm_bdesc *txbd;
	struct fec_softc *sc = ifp->if_softc;
	struct mbuf *m;
	int len;

DPRINTF("fec_start\n");

	if ((ifp->if_flags & (IFF_RUNNING | IFF_OACTIVE)) != IFF_RUNNING)
		return;

	invalidatecache(sc->sc_txbd, FEC_NTXBUF * sizeof *txbd);
	txbd = sc->sc_next_txbd;
	for (;;) {
		int n = txbd - sc->sc_txbd;
		caddr_t va = sc->sc_txbuf + 0x800 * n;

		if (txbd->bd_flags & TxBD_R) {
			printf("txbuffer full\n");
			break;
		}
		IFQ_DEQUEUE(&ifp->if_snd, m);
		if (m == NULL)
			break;

DPRINTF("fec_start txbd = %p\n", txbd);
		/* 5 seconds to watch for failing to transmit */
		ifp->if_timer = 5;
		ifp->if_opackets++;

		ifp->if_flags |= IFF_OACTIVE;

		len = fec_put(sc, va, m);
#if NBPFILTER > 0
		/*
		 * If BPF is listening on this interface, let it see the
		 * packet before we commit it to the wire.
		 */
		if (ifp->if_bpf)
			bpf_tap(ifp->if_bpf, va, len);
#endif
		flushcache(va, len);
		m_freem(m);

		txbd->bd_flags &= TxBD_W;
		txbd->bd_flags |= TxBD_R|TxBD_L|TxBD_TC;
		txbd->bd_len = len;

		if (txbd->bd_flags & TxBD_W)
			txbd = sc->sc_txbd;
		else
			txbd++;


		/* if (ifp->if_flags & IFF_OACTIVE) */
			break;
	}
	sc->sc_next_txbd = txbd;
	eieio();
	flushcache(sc->sc_txbd, FEC_NTXBUF * sizeof *txbd);

	/* Start transmission. */
	immr->fec_x_des_active = 0x0100;
}


struct mbuf *
fec_get(sc, pkt, totlen)
	struct fec_softc *sc;
	caddr_t pkt;
	int totlen;
{
	struct mbuf *m;
	struct mbuf *top, **mp;
	int len;

	if (totlen > 2048) {
		printf("fec_get: len > 2048 (%d)\n", totlen);
		return NULL;
	}

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if (m == 0)
		return 0;
	m->m_flags |= M_HASFCS;
	m->m_pkthdr.rcvif = &sc->sc_if;
	m->m_pkthdr.len = totlen;
	len = MHLEN;
	top = 0;
	mp = &top;

	while (totlen > 0) {
		if (top) {
			MGET(m, M_DONTWAIT, MT_DATA);
			if (m == 0) {
				m_freem(top);
				return 0;
			}
			len = MLEN;
		}
		if (totlen >= MINCLSIZE) {
			MCLGET(m, M_DONTWAIT);
			if ((m->m_flags & M_EXT) == 0) {
				m_free(m);
				m_freem(top);
				return 0;
			}
			len = MCLBYTES;
		}
		m->m_len = len = min(totlen, len);
		bcopy(pkt, mtod(m, caddr_t), len);
		pkt += len;
		totlen -= len;
		*mp = m;
		mp = &m->m_next;
	}

	return top;
}

int
fec_put(sc, buff, m)
	struct fec_softc *sc;
	caddr_t buff;
	struct mbuf *m;
{
	int len, tlen = 0;

	for (; m; m = m->m_next) {
		len = m->m_len;
		if (len == 0)
			continue;
		bcopy(mtod(m, caddr_t), buff, len);
		buff += len;
		tlen += len;
	}
	if (tlen > 2048)
		panic("%s: fec_put packet overflow", sc->sc_dev.dv_xname);

	return tlen;
}

void
fec_rint(sc)
	struct fec_softc *sc;
{
	volatile struct mpc860dev *immr = get_immr();
	struct cpm_bdesc *rxbd;
	struct ifnet *ifp = &sc->sc_if;
	struct mbuf *m;
	int flags, i;

	invalidatecache(sc->sc_rxbd, FEC_NRXBUF * sizeof *rxbd);
	rxbd = sc->sc_next_rxbd;
	for (i = 0; i < FEC_NRXBUF; i++) {
		int n = rxbd - sc->sc_rxbd;
		caddr_t va = sc->sc_rxbuf + 0x800 * n;

		if (rxbd->bd_flags & RxBD_E)
			break;

		invalidatecache(va, rxbd->bd_len);
		m = fec_get(sc, va, rxbd->bd_len);
#if NBPFILTER > 0
		/*
		 * Check if there's a BPF listener on this interface. 
		 * If so, hand off the raw packet to BPF.
		 */
		if (ifp->if_bpf)
			bpf_tap(ifp->if_bpf, va, rxbd->bd_len);
#endif
		rxbd->bd_flags &= RxBD_W;
		rxbd->bd_flags |= RxBD_E;
		(*ifp->if_input)(ifp, m);

		if (rxbd->bd_flags & RxBD_W)
			rxbd = sc->sc_rxbd;
		else
			rxbd++;
	}
#ifdef DIAGNOSTIC_notyet
	if (i == 0)
		printf("%s: bogus receive interrupt\n", ifp->if_xname);
#endif
	sc->sc_next_rxbd = rxbd;
	eieio();
	flushcache(sc->sc_rxbd, FEC_NRXBUF * sizeof *rxbd);

	immr->fec_r_des_active = 0x100;		/* XXX needed? */
}

int
fec_mii_readreg(dev, phy, reg)
	struct device *dev;
	int phy, reg;
{
	volatile struct mpc860dev *immr = get_immr();

	immr->fec_ievent = 0x00800000;
	immr->fec_mii_data = 0x60020000 | phy << 23 | reg << 18;
	while ((immr->fec_ievent & 0x00800000) == 0);	/* XXX timeout */
	immr->fec_ievent = 0x00800000;

	return immr->fec_mii_data & 0xffff;
}


void
fec_mii_writereg(dev, phy, reg, val)
	struct device *dev;
	int phy, reg, val;
{
	volatile struct mpc860dev *immr = get_immr();

	immr->fec_ievent = 0x00800000;
	immr->fec_mii_data = 0x50020000 | phy << 23 | reg << 18 | val & 0xffff;
	while ((immr->fec_ievent & 0x00800000) == 0);	/* XXX timeout */
	immr->fec_ievent = 0x00800000;
}


void
fec_mii_statchg(dev)
	struct device *dev;
{
	volatile struct mpc860dev *immr = get_immr();
	struct fec_softc *sc = (void *)dev;
	struct ifnet *ifp = &sc->sc_if;

	/* Disable ethernet */
	immr->fec_ecntrl = FEC_ECNTRL_PINMUX;

	if (IFM_OPTIONS(sc->sc_mii.mii_media_active) & IFM_FDX) {
		DPRINTF("fec_mii_statchg FULL\n");
		immr->fec_r_cntrl = FEC_RCNTRL_MII_MODE;
		immr->fec_x_cntrl = FEC_XCNTRL_FDEN;
	} else {
		DPRINTF("fec_mii_statchg HALF\n");
		immr->fec_r_cntrl = FEC_RCNTRL_MII_MODE | FEC_RCNTRL_DRT;
		immr->fec_x_cntrl = 0;
	}
	if (ifp->if_flags & IFF_PROMISC)
		immr->fec_r_cntrl |= FEC_RCNTRL_PROM;

	/* Enable ethernet */
	immr->fec_ecntrl = FEC_ECNTRL_PINMUX|FEC_ECNTRL_ETHER_EN;
	immr->fec_r_des_active = 0x100;
}

int
fec_mediachange(ifp)
	struct ifnet *ifp;
{
	struct fec_softc *sc = ifp->if_softc;

	return mii_mediachg(&sc->sc_mii);
}

void
fec_mediastatus(ifp, ifmr)
	struct ifnet *ifp;
	struct ifmediareq *ifmr;
{
	struct fec_softc *sc = ifp->if_softc;

	mii_pollstat(&sc->sc_mii);

	ifmr->ifm_status = sc->sc_mii.mii_media_status;
	ifmr->ifm_active = sc->sc_mii.mii_media_active;
}

int
fec_ioctl(ifp, cmd, data)
	struct ifnet *ifp;
	u_long cmd;
	caddr_t data;
{
	struct fec_softc *sc = ifp->if_softc;
	struct ifreq *ifr = (struct ifreq *)data;
	int s, error;

	s = splnet();

	switch (cmd) {
	case SIOCSIFMEDIA:
	case SIOCGIFMEDIA:
		error = ifmedia_ioctl(ifp, ifr, &sc->sc_mii.mii_media, cmd);
		break;

	default:
		error = ether_ioctl(ifp, cmd, data);
		if (error == ENETRESET) {
			if (ifp->if_flags & IFF_UP) {
				/*
				 * Multicast list has changed.  Set the
				 * hardware filter accordingly.
				 */
				fec_init(ifp);
			}
			error = 0;
		}
		break;
	}

	/* Try to get more packets going. */
	if (ifp->if_flags & IFF_UP)
		fec_start(ifp);

	splx(s);
	return error;
}


void
fec_watchdog(ifp)
	struct ifnet *ifp;
{
	struct fec_softc *sc = ifp->if_softc;

	printf("%s: device timeout\n", ifp->if_xname);
	ifp->if_oerrors++;

	fec_init(ifp);
}


void
fec_stop(ifp, disable)
	struct ifnet *ifp;
	int disable;
{
printf("fec_stop disable = %d\n", disable);
}

