/*	$Id: if_sce.c,v 1.17 2002/10/06 13:18:56 tsubai Exp $	*/

/*-
 * Copyright (c) 2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_inet.h"
#include "bpfilter.h"

#include <sys/param.h>
#include <sys/device.h>
#include <sys/ioctl.h>
#include <sys/kernel.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/systm.h>

#include <uvm/uvm_extern.h>

#include <net/if.h>
#include <net/if_ether.h>

#include <machine/autoconf.h>
#include <machine/mpc860dev.h>

#if NBPFILTER > 0
#include <net/bpf.h>
#endif

#ifdef SCE_DEBUG
#define DPRINTF printf
#else
#define DPRINTF while (0) printf
#endif

#define SCE_NRXBUF	8
#define SCE_NTXBUF	2

struct sce_softc {
	struct device sc_dev;
	struct ethercom sc_ec;

	char *sc_rxbuf;
	char *sc_txbuf;
	struct cpm_bdesc *sc_rxbd;
	struct cpm_bdesc *sc_next_rxbd;
	struct cpm_bdesc *sc_txbd;
	struct cpm_bdesc *sc_next_txbd;
};

extern char sce_enaddr[];

int sce_match(struct device *, struct cfdata *, void *);
void sce_attach(struct device *, struct device *, void *);

int sceintr(void *);
void sce_tint(struct sce_softc *);
void sce_rint(struct sce_softc *);
 
void sce_start(struct ifnet *);
int sce_ioctl(struct ifnet *, u_long, caddr_t);
void sce_watchdog(struct ifnet *);
int sce_init(struct ifnet *);
void sce_setladrf(struct sce_softc *);
void sce_stop(struct ifnet *, int);

struct mbuf *sce_get(struct sce_softc *, caddr_t, int);
int sce_put(struct sce_softc *, caddr_t, struct mbuf *);

struct cfattach sce_ca = {
	"sce", {}, sizeof(struct sce_softc), sce_match, sce_attach
};

int
sce_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;

	if (strcmp(ca->ca_name, "sce") == 0)
		return 1;
	return 0;
}

void
sce_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct sce_softc *sc = (void *)self;
	struct cfdata *cf = self->dv_cfdata;
	int vec = cf->cf_vec;
	struct ifnet *ifp = &sc->sc_ec.ec_if;
	caddr_t buf;

	buf = malloc(0x800 * (SCE_NRXBUF + SCE_NTXBUF), M_DEVBUF, M_NOWAIT);
	if (buf == NULL) {
		printf(": cannot malloc\n");
		return;
	}
	sc->sc_rxbuf = buf;
	buf += 0x800 * SCE_NRXBUF;
	sc->sc_txbuf = buf;

	printf(": vec %d", vec);
	printf(" address %s\n", ether_sprintf(sce_enaddr));

	bcopy(sc->sc_dev.dv_xname, ifp->if_xname, IFNAMSIZ);
	ifp->if_softc = sc;
	ifp->if_ioctl = sce_ioctl;
	ifp->if_start = sce_start;
	ifp->if_watchdog = sce_watchdog;
	ifp->if_init = sce_init;
	ifp->if_stop = sce_stop;
	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
	IFQ_SET_READY(&ifp->if_snd);

	sce_init(ifp);

	if_attach(ifp);
	ether_ifattach(ifp, sce_enaddr);

	intr_establish(vec, IST_LEVEL, IPL_NET, sceintr, sc);

}

static inline void
eieio()
{
	asm volatile ("eieio; sync");
}

int
sce_init(ifp)
	struct ifnet *ifp;
{
	volatile struct mpc860dev *immr = get_immr();
	struct sce_softc *sc = ifp->if_softc;
	struct scc_enet *sce = (void *)immr->cpm_pram2;
	struct cpm_bdesc *rxbd, *txbd;
	vaddr_t buf;
	int i;

	/* Enable TXD2 and RXD2. */
	immr->cpm_papar |= 0x000c;
	immr->cpm_padir &= ~0x000c;
	immr->cpm_paodr &= ~0x0008;

	/* Enable CTS2(CLSN) and CD2(RENA). */
	immr->cpm_pcpar &= ~0x00c0;
	immr->cpm_pcdir &= ~0x00c0;
	immr->cpm_pcso |= 0x00c0;

	/* Enable CLK1 and CLK2. */
	immr->cpm_papar |= 0x0300;
	immr->cpm_padir &= ~0x0300;

	/* Disable TENA(RTS). */
	immr->cpm_pbpar &= ~0x2000;
	immr->cpm_pbdir &= ~0x2000;


#define SICR_SCC1 0x000000ff
#define SICR_SCC2 0x0000ff00
#define SICR_SCC3 0x00ff0000
#define SICR_SCC4 0xff000000

#define SICR_GR     0x80
#define SICR_SC     0x40
#define SICR_RxBRG1 0x00
#define SICR_RxBRG2 0x08
#define SICR_RxBRG3 0x10
#define SICR_RxBRG4 0x18
#define SICR_RxCLK1 0x20
#define SICR_RxCLK2 0x28
#define SICR_RxCLK3 0x30
#define SICR_RxCLK4 0x38
#define SICR_TxBRG1 0x00
#define SICR_TxBRG2 0x01
#define SICR_TxBRG3 0x02
#define SICR_TxBRG4 0x03
#define SICR_TxCLK1 0x04
#define SICR_TxCLK2 0x05
#define SICR_TxCLK3 0x06
#define SICR_TxCLK4 0x07


	/* Connect CLK1 and CLK2 to SCC2. */
	immr->cpm_sicr &= ~SICR_SCC2;
	immr->cpm_sicr |= (SICR_RxCLK2 | SICR_TxCLK1) << 8;

	/* Connect NMSI and SCC2. */
	immr->cpm_sicr &= ~(SICR_SC << 8);

#if 0 /* Not here */
	/* Initialize SDMA. */
	immr->siu_sdcr = 0x0001;
#endif

	/* Initialize SCC2 parameter ram. */

#define DPRAM_SCC2_BASE	((char *)0xff002900)	/* cpm_dpram[0x900] */

	rxbd = (void *)DPRAM_SCC2_BASE;
	txbd = rxbd + SCE_NRXBUF;

	sc->sc_rxbd = sc->sc_next_rxbd = rxbd;
	sc->sc_txbd = sc->sc_next_txbd = txbd;

	sce->rbase = (int)rxbd;
	sce->tbase = (int)txbd;

#define RxBD_I   0x1000

	buf = (vaddr_t)sc->sc_rxbuf;
	for (i = 0; i < SCE_NRXBUF; i++) {
		rxbd->bd_flags = RxBD_E | RxBD_I;
		rxbd->bd_addr = (char *)vtophys(buf);
		buf += 0x800;
		rxbd++;
	}
	(rxbd - 1)->bd_flags |= RxBD_W;

#define TxBD_PAD 0x4000
#define TxBD_I   0x1000

	buf = (vaddr_t)sc->sc_txbuf;
	for (i = 0; i < SCE_NTXBUF; i++) {
		txbd->bd_flags = TxBD_PAD | TxBD_I | TxBD_L | TxBD_TC;
		txbd->bd_len = 0;
		txbd->bd_addr = (char *)vtophys(buf);
		buf += 0x800;
		txbd++;
	}
	(txbd - 1)->bd_flags |= TxBD_W;


#define CPCR_RST   0x8000	/* CP reset */
#define CPCR_SCC1  0x0000
#define CPCR_I2C   0x0010
#define CPCR_IDMA  CPCR_I2C
#define CPCR_SCC2  0x0040
#define CPCR_SPI   0x0050
#define CPCR_IDMA2 CPCR_SPI
#define CPCR_RT    CPCR_SPI
#define CPCR_SCC3  0x0080
#define CPCR_SMC1  0x0090
#define CPCR_DSP1  CPCR_SMC1
#define CPCR_SCC4  0x00c0
#define CPCR_SMC2  0x00d0
#define CPCR_DSP2  CPCR_SMC2
#define CPCR_PIP   CPCR_SMC2
#define CPCR_FLG   0x0001	/* Command in progress flag */

#define CPCR_INIT_RX_AND_TX_PARAMS	0x0000
#define CPCR_INIT_RX_PARAMS		0x0100
#define CPCR_SET_GROUP_ADDRESS		0x0800
/* ... */


	/* Issue an INIT-RX-AND-TX-PARAMETERS command on SCC2. */
	immr->cpm_cpcr = CPCR_SCC2|CPCR_INIT_RX_AND_TX_PARAMS|CPCR_FLG;
	while (immr->cpm_cpcr & CPCR_FLG);

	/* Normal operation. */
	sce->rfcr = 0x10;
	sce->tfcr = 0x10;	/* XXX  @@ Ϥ줾 0x18 Ƥ */

	/* Max receive buffer length. */
	sce->mrblr = 1520;

	/* 32bit CCITT-CRC. */
	sce->c_pres = 0xffffffff;
	sce->c_mask = 0xdebb20e3;

	/* Initialize error counters. */
	sce->crcec = 0;
	sce->alec = 0;

	/* Padding. */
	sce->pads = 0x8888;

	/* Retry limit. */
	sce->ret_lim = 0x000f;

	/* Max frame length. */
	sce->mflr = 1518;

	/* Min frame length. */
	sce->minflr = 64;

	/* Max DMA count. */
	sce->maxd1 = 1520;
	sce->maxd2 = 1520;

	/* Group addresses are not used. */
	sce->gaddr1 = sce->gaddr2 = sce->gaddr3 = sce->gaddr4 = 0;

	/* My MAC address. */
	sce->paddr1_h = sce_enaddr[5] << 8 | sce_enaddr[4];
	sce->paddr1_m = sce_enaddr[3] << 8 | sce_enaddr[2];
	sce->paddr1_l = sce_enaddr[1] << 8 | sce_enaddr[0];

	sce->p_per = 0;

	/* Initialize hash table. */
	sce->iaddr1 = sce->iaddr2 = sce->iaddr3 = sce->iaddr4 = 0;

	sce->taddr_h = 0;
	sce->taddr_m = 0;
	sce->taddr_l = 0;

/* SCC Ethernet event masks. */
#define SCCE_GRA 0x0080  /* Graceful stop complete */
#define SCCE_TXE 0x0010  /* Transmit error */
#define SCCE_RXF 0x0008  /* Rx frame */
#define SCCE_BSY 0x0004  /* Busy (rx buffer full) */
#define SCCE_TXB 0x0002  /* Tx buffer */
#define SCCE_RXB 0x0001  /* Rx buffer */

	/* Clear all events. */
	immr->cpm_scce2 = 0xffff;

	/* Enable TXE, RXF, and TXB interrupts. */
	immr->cpm_sccm2 = SCCE_TXE | SCCE_RXF | SCCE_TXB;

	/* Enable SCC2 interrupt. */
	immr->cpm_cimr |= 0x20000000;

	immr->cpm_cicr &= ~0x000c0000;
	immr->cpm_cicr |= 0x00048080;

	/* Normal operation. */
	immr->cpm_gsmr_h2 = 0;

	immr->cpm_gsmr_l2 = 0x1088000c;

	immr->cpm_dsr2 = 0xd555;

/* PSMR values of SCC Ethernet */
#define PSMR_HBC	0x8000	/* Heartbeat cheching */
#define PSMR_FC		0x4000	/* Force collision */
#define PSMR_RSH	0x2000	/* Receive short frames */
#define PSMR_IAM	0x1000	/* Individual address mode */
#define PSMR_CRC	0x0800	/* CRC selection ([4-5]=0b10) */
#define PSMR_PRO	0x0200	/* Promiscuous */
#define PSMR_BRO	0x0100	/* Broadcast address */
#define PSMR_SBT	0x0080	/* Stop backoff timer */
#define PSMR_LPB	0x0040	/* Loopback operation */
#define PSMR_SIP	0x0020	/* Sample input pins */
#define PSMR_LCW	0x0010	/* Late collision window */
#define PSMR_NIB	0x000e	/* Number of ignored bits */
#define  PSMR_NIB13	0x0000
#define  PSMR_NIB14	0x0002
#define  PSMR_NIB22	0x000a
#define  PSMR_NIB23	0x000c
#define  PSMR_NIB24	0x000e
#define PSMR_FDE	0x0001	/* Full duplex mode */

	/* Promisc, hash, duplex, etc. */
	immr->cpm_psmr2 = PSMR_CRC | PSMR_NIB22;
	if (ifp->if_flags & IFF_PROMISC)
		immr->cpm_psmr2 |= PSMR_PRO;

	sce_setladrf(sc);

	/* Enable TENA(RTS). */
	immr->cpm_pbpar |= 0x2000;
	immr->cpm_pbdir |= 0x2000;

#define SCE_LOOPBACK 0x0800
#define SCE_LEDC     0x0080
#define SCE_LEDL     0x0040

	immr->cpm_papar &= ~(SCE_LEDC | SCE_LEDL);
	immr->cpm_padir &= ~SCE_LEDC;
	immr->cpm_padir &= ~SCE_LEDL;

	immr->cpm_papar &= ~SCE_LOOPBACK;
	immr->cpm_padir |= SCE_LOOPBACK;
	immr->cpm_padat &= ~SCE_LOOPBACK;

	/* Enable ENT and ENR. */
	immr->cpm_gsmr_l2 = 0x1088003c;

	ifp->if_flags |= IFF_RUNNING;
	ifp->if_flags &= ~IFF_OACTIVE;

	return 0;
}

void
sce_setladrf(sc)
	struct sce_softc *sc;
{
	volatile struct mpc860dev *immr = get_immr();
	struct scc_enet *sce = (void *)immr->cpm_pram2;
	struct ifnet *ifp = &sc->sc_ec.ec_if;
	struct ether_multi *enm;
	struct ether_multistep step;
	struct ethercom *ec = &sc->sc_ec;

	if (ifp->if_flags & IFF_PROMISC) {
		/* Turn on promiscuous mode; turn off the hash filter */
		ifp->if_flags |= IFF_ALLMULTI;
		goto allmulti;
	}

	/*
	 * Set up multicast address filter by passing all multicast addresses
	 * through a crc generator, and then using the high order 8 bits as an
	 * index into the 256 bit logical address filter.  The high order bit
	 * selects the word, while the rest of the bits select the bit within
	 * the word.
	 */

	ETHER_FIRST_MULTI(step, ec, enm);
	while (enm) {
		if (bcmp(enm->enm_addrlo, enm->enm_addrhi, 6)) {
			/*
			 * We must listen to a range of multicast addresses.
			 * For now, just accept all multicasts, rather than
			 * trying to set only those filter bits needed to match
			 * the range.  (At this time, the only use of address
			 * ranges is for IP multicast routing, for which the
			 * range is big enough to require all bits set.)
			 */
			ifp->if_flags |= IFF_ALLMULTI;
			goto allmulti;
		}

		sce->taddr_h = enm->enm_addrlo[5] << 8 | enm->enm_addrlo[4];
		sce->taddr_m = enm->enm_addrlo[3] << 8 | enm->enm_addrlo[2];
		sce->taddr_l = enm->enm_addrlo[1] << 8 | enm->enm_addrlo[0];

		immr->cpm_cpcr = CPCR_SCC2|CPCR_SET_GROUP_ADDRESS|CPCR_FLG;
		while (immr->cpm_cpcr & CPCR_FLG);

		ETHER_NEXT_MULTI(step, enm);
	}

	ifp->if_flags &= ~IFF_ALLMULTI;
	return;

allmulti:
	sce->gaddr1 = 0xffff;
	sce->gaddr2 = 0xffff;
	sce->gaddr3 = 0xffff;
	sce->gaddr4 = 0xffff;
}

/* CPM interrupt vector number.  XXX not here */
#define CPM_IVEC_PC15	0x1f
#define CPM_IVEC_SCC1	0x1e
#define CPM_IVEC_SCC2	0x1d
/* ... */


int
sceintr(v)
	void *v;
{
	volatile struct mpc860dev *immr = get_immr();
	struct sce_softc *sc = v;
	int event;

	/* Check if this interrupt is mine. */
	if (immr->cpm_civr >> 11 != CPM_IVEC_SCC2)
		return 0;

	/* Acknowledge interrupt. */
	immr->cpm_civr |= 1;

	event = immr->cpm_scce2;
	DPRINTF("event = 0x%x\n", event);

	if (event & SCCE_TXE)
		printf("sceintr: tx error\n");
	if (event & SCCE_RXF)
		sce_rint(sc);
	if (event & SCCE_TXB)
		sce_tint(sc);

	/* Clear SCC Ethernet events. */
	immr->cpm_scce2 = event;

	/* Clear SCC2 interrupt condition. */
	immr->cpm_cisr |= 1 << CPM_IVEC_SCC2;

	return 1;
}

void
sce_tint(sc)
	struct sce_softc *sc;
{
	struct ifnet *ifp = &sc->sc_ec.ec_if;

	ifp->if_flags &= ~IFF_OACTIVE;
	ifp->if_timer = 0;
	sce_start(ifp);
}

void
sce_start(ifp)
	struct ifnet *ifp;
{
	struct cpm_bdesc *txbd;
	struct sce_softc *sc = ifp->if_softc;
	struct mbuf *m;
	int len;

	DPRINTF("sce_start\n");

	if ((ifp->if_flags & (IFF_RUNNING | IFF_OACTIVE)) != IFF_RUNNING)
		return;

	invalidatecache(sc->sc_txbd, SCE_NTXBUF * sizeof *txbd);
	txbd = sc->sc_next_txbd;
	for (;;) {
		int n = txbd - sc->sc_txbd;
		caddr_t va = sc->sc_txbuf + 0x800 * n;

		if (txbd->bd_flags & TxBD_R) {
			printf("txbuffer full\n");
			break;
		}
		IFQ_DEQUEUE(&ifp->if_snd, m);
		if (m == NULL)
			break;

		DPRINTF("sce_start txbd = %p\n", txbd);

		/* 5 seconds to watch for failing to transmit */
		ifp->if_timer = 5;
		ifp->if_opackets++;

		ifp->if_flags |= IFF_OACTIVE;

		len = sce_put(sc, va, m);
#if NBPFILTER > 0
		/*
		 * If BPF is listening on this interface, let it see the
		 * packet before we commit it to the wire.
		 */
		if (ifp->if_bpf)
			bpf_tap(ifp->if_bpf, va, len);
#endif
		flushcache(va, len);
		m_freem(m);

		txbd->bd_flags &= TxBD_W;
		txbd->bd_flags |= TxBD_R | TxBD_PAD | TxBD_I | TxBD_L | TxBD_TC;
		txbd->bd_len = len;

		if (txbd->bd_flags & TxBD_W)
			txbd = sc->sc_txbd;
		else
			txbd++;

		/* if (ifp->if_flags & IFF_OACTIVE) */
			break;
	}
	sc->sc_next_txbd = txbd;
	eieio();
	flushcache(sc->sc_txbd, SCE_NTXBUF * sizeof *txbd);

	/* Start transmission. */
	/* immr->sce_x_des_active = 0x0100; */
}


struct mbuf *
sce_get(sc, pkt, totlen)
	struct sce_softc *sc;
	caddr_t pkt;
	int totlen;
{
	struct mbuf *m;
	struct mbuf *top, **mp;
	int len;

	if (totlen > 2048) {
		printf("sce_get: len > 2048 (%d)\n", totlen);
		return NULL;
	}

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if (m == 0)
		return 0;
	m->m_flags |= M_HASFCS;
	m->m_pkthdr.rcvif = &sc->sc_ec.ec_if;
	m->m_pkthdr.len = totlen;
	len = MHLEN;
	top = 0;
	mp = &top;

	while (totlen > 0) {
		if (top) {
			MGET(m, M_DONTWAIT, MT_DATA);
			if (m == 0) {
				m_freem(top);
				return 0;
			}
			len = MLEN;
		}
		if (totlen >= MINCLSIZE) {
			MCLGET(m, M_DONTWAIT);
			if ((m->m_flags & M_EXT) == 0) {
				m_free(m);
				m_freem(top);
				return 0;
			}
			len = MCLBYTES;
		}
		m->m_len = len = min(totlen, len);
		bcopy(pkt, mtod(m, caddr_t), len);
		pkt += len;
		totlen -= len;
		*mp = m;
		mp = &m->m_next;
	}

	return top;
}

int
sce_put(sc, buff, m)
	struct sce_softc *sc;
	caddr_t buff;
	struct mbuf *m;
{
	int len, tlen = 0;

	for (; m; m = m->m_next) {
		len = m->m_len;
		if (len == 0)
			continue;
		bcopy(mtod(m, caddr_t), buff, len);
		buff += len;
		tlen += len;
	}
	if (tlen > 2048)
		panic("%s: sce_put packet overflow", sc->sc_dev.dv_xname);

	return tlen;
}

void
sce_rint(sc)
	struct sce_softc *sc;
{
	struct cpm_bdesc *rxbd;
	struct ifnet *ifp = &sc->sc_ec.ec_if;
	struct mbuf *m;
	int i;

	invalidatecache(sc->sc_rxbd, SCE_NRXBUF * sizeof *rxbd);
	rxbd = sc->sc_next_rxbd;
	for (i = 0; i < SCE_NRXBUF; i++) {
		int n = rxbd - sc->sc_rxbd;
		caddr_t va = sc->sc_rxbuf + 0x800 * n;

		if (rxbd->bd_flags & RxBD_E)
			break;

		invalidatecache(va, rxbd->bd_len);
		m = sce_get(sc, va, rxbd->bd_len);
#if NBPFILTER > 0
		/*
		 * Check if there's a BPF listener on this interface. 
		 * If so, hand off the raw packet to BPF.
		 */
		if (ifp->if_bpf)
			bpf_tap(ifp->if_bpf, va, rxbd->bd_len);
#endif
		rxbd->bd_flags &= RxBD_W;
		rxbd->bd_flags |= RxBD_E | RxBD_I;
		(*ifp->if_input)(ifp, m);

		if (rxbd->bd_flags & RxBD_W)
			rxbd = sc->sc_rxbd;
		else
			rxbd++;
	}
#ifdef DIAGNOSTIC_notyet
	if (i == 0)
		printf("%s: bogus receive interrupt\n", ifp->if_xname);
#endif
	sc->sc_next_rxbd = rxbd;
	eieio();
	flushcache(sc->sc_rxbd, SCE_NRXBUF * sizeof *rxbd);
}

int
sce_ioctl(ifp, cmd, data)
	struct ifnet *ifp;
	u_long cmd;
	caddr_t data;
{
	int s, error;

	s = splnet();

	switch (cmd) {
	default:
		error = ether_ioctl(ifp, cmd, data);
		if (error == ENETRESET) {
			if (ifp->if_flags & IFF_UP) {
				/*
				 * Multicast list has changed.  Set the
				 * hardware filter accordingly.
				 */
				sce_init(ifp);
			}
			error = 0;
		}
		break;
	}

	/* Try to get more packets going. */
	if (ifp->if_flags & IFF_UP)
		sce_start(ifp);

	splx(s);
	return error;
}

void
sce_watchdog(ifp)
	struct ifnet *ifp;
{
	printf("%s: device timeout\n", ifp->if_xname);
	ifp->if_oerrors++;

	sce_init(ifp);
}

void
sce_stop(ifp, disable)
	struct ifnet *ifp;
	int disable;
{
	printf("sce_stop disable = %d\n", disable);
}

/*
 * 
 *

TODR[TOD] 򥻥åȤȤϤޤ롣
ۤäȤ 128ꥢ륯åȤΥݡ󥰤롣
顼 BD Τ

GPIO
PA4: EN_LBK  (O)
PA8: EN_LEDC (I)
PA9: EN_LEDL (I)
*/
