/*	$Id: scc.c,v 1.12 2003/03/22 10:17:35 tsubai Exp $	*/

/*-
 * Copyright (c) 2001,2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/callout.h>
#include <sys/device.h>
#include <sys/ioctl.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/tty.h>

#include <sys/fcntl.h>

#include <uvm/uvm_extern.h>
#include <dev/cons.h>

#include <machine/autoconf.h>
#include <machine/mpc860dev.h>

#define SCCUNIT(x)	(minor(x) & 0x7ffff)
#define SCCDIALOUT(x)	(minor(x) & 0x80000)

#define SCC_HW_CONSOLE	0x0001


/* Macros to clear/set/test flags. */
#define SET(t, f)	(t) |= (f)
#define CLR(t, f)	(t) &= ~(f)
#define ISSET(t, f)	((t) & (f))

struct scc_softc {
	struct device sc_dev;

	struct tty *sc_tty;
	int sc_hwflags;
	int sc_swflags;

	struct cpm_bdesc *sc_rxbd;
	struct cpm_bdesc *sc_txbd;
};

int scc_match(struct device *, struct cfdata *, void *);
void scc_attach(struct device *, struct device *, void *);

void sccstart(struct tty *);
int sccparam(struct tty *, struct termios *);

int sccintr(void *);

struct cfattach scc_ca = {
	"scc", {}, sizeof(struct scc_softc), scc_match, scc_attach
};

dev_type_open(sccopen);
dev_type_close(sccclose);
dev_type_read(sccread);
dev_type_write(sccwrite);
dev_type_ioctl(sccioctl);
dev_type_stop(sccstop);
dev_type_tty(scctty);
dev_type_poll(sccpoll);

const struct cdevsw scc_cdevsw = {
	sccopen, sccclose, sccread, sccwrite, sccioctl,
	sccstop, scctty, sccpoll, nommap, nokqfilter, D_TTY
};

int sccconsspeed = 9600;
int sccconscflag = TTYDEF_CFLAG|CLOCAL;

extern struct cfdriver scc_cd;

int
scc_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;

	if (strcmp(ca->ca_name, "scc") == 0)
		return 1;

	return 0;
}

void
scc_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct scc_softc *sc = (void *)self;
	struct cfdata *cf = self->dv_cfdata;
	int vec = cf->cf_vec;
	struct tty *tp;
	struct cpm_bdesc *txbd, *rxbd;

	printf(": vec %d", vec);

	rxbd = (void *)0xff002820;	/* cpm_dpram[0x820] */
	txbd = (void *)0xff002830;	/* cpm_dpram[0x830] */
	sc->sc_rxbd = rxbd;
	sc->sc_txbd = txbd;

	tp = ttymalloc();
	tp->t_oproc = sccstart;
	tp->t_param = sccparam;
	sc->sc_tty = tp;

	tty_attach(tp);

	if (1) {
		int maj;

		maj = cdevsw_lookup_major(&scc_cdevsw);
		if (maj == -1)
			panic("no scc in cdevsw?");
		cn_tab->cn_dev = makedev(maj, sc->sc_dev.dv_unit);
		sc->sc_hwflags |= SCC_HW_CONSOLE;
		printf(" (console)");
	}

	printf("\n");
	intr_establish(vec, IST_LEVEL, IPL_TTY, sccintr, sc);

	rxbd[0].bd_flags = 0x8000;
	rxbd[1].bd_flags = 0xa000;
}

static inline void
eieio()
{
	asm volatile ("eieio; sync");
}

void
scc_uart_init()
{
	volatile struct mpc860dev *immr = get_immr();
	struct cpm_bdesc *txbd, *rxbd;
	struct scc_uart *scc = (void *)immr->cpm_pram1;
	int i;

	immr->cpm_papar |=  0x0003;	/* PA14 = TXD1, PA15 = RXD1 */
	immr->cpm_padir &= ~0x0003;	/* Select func 0 on pin 14 and 15. */
	immr->cpm_paodr &= ~0x0002;	/* TXD1 is NOT open-drain */

	immr->cpm_pbpar |=  0x1000;	/* PB19 = RTS1 */
	immr->cpm_pbdir |=  0x1000;	/* Select function 1 on pin 19. */
	immr->cpm_pbodr &= ~0x1000;

	immr->cpm_pcpar &= ~0x0030;	/* PC10 = CD1, PC11 = CTS1 */
	immr->cpm_pcdir &= ~0x0030;
	immr->cpm_pcso  |=  0x0030;

	immr->cpm_brgc2 = 0x1028a;	/* Enable BRG2, div = 325 (9600bps) */
					/* 50*1000000/16/(325+1) = 9586 */
	immr->cpm_sicr &= ~0xff;	/* Connect BRG1 to SCC1 */
	immr->cpm_sicr |=  0x09;	/* Connect BRG2 to SCC1 */

/* PA11 = DTR1 (out) */
//immr->cpm_papar &= ~0x0010;
//immr->cpm_padat |=  0x0010;
//immr->cpm_padat &= ~0x0010;

	/* siu_sdcr = 1; */

	rxbd = (void *)0xff002820;
	txbd = (void *)0xff002830;
	scc->rbase = (int)rxbd;
	scc->tbase = (int)txbd;

	rxbd->bd_flags = 0;
	rxbd->bd_addr = (char *)0xff002840;
	rxbd++;
	rxbd->bd_flags = 0;
	rxbd->bd_addr = (char *)0xff002842;

	txbd->bd_flags = 0;
	txbd->bd_addr = (char *)0xff002844;
	txbd++;
	txbd->bd_flags = 0;
	txbd->bd_addr = (char *)0xff002846;
	eieio();

	immr->cpm_cpcr = 0x0001;	/* Initialize SCC1 */

	scc->rfcr = 0x10;
	scc->tfcr = 0x10;
	scc->mrblr = 1;		/* maximum receive length */
	scc->max_idl = 0;
	scc->brkcr = 1;
	scc->parec = 0;
	scc->frmec = 0;
	scc->brkec = 0;
	scc->uaddr1 = 0;
	scc->uaddr2 = 0;
	scc->toseq = 0;

	for (i = 0; i < 8; i++)
		scc->character[i] = 0x8000;

	scc->rccm = 0xc0ff;

	rxbd = (void *)0xff002820;
	txbd = (void *)0xff002830;

	rxbd->bd_flags = 0x8000;
	rxbd++;
	rxbd->bd_flags = 0xa000;

	txbd->bd_flags = 0x2000;

	immr->cpm_scce1 = 0xffff;	/* Clear any previous events. */
	immr->cpm_sccm1 = 0xffff;	/* Unmask all events. */

	immr->cpm_cimr = 0;		/* XXX */
	immr->cpm_cimr |= 0x40000000;	/* Enable SCC1 interrupt. */
	immr->cpm_cicr = 0x00008080;	/* Enable CPM interrupts. */

	immr->cpm_gsmr_h1 = 0x00000020;
	immr->cpm_gsmr_l1 = 0x00028004;

	immr->cpm_psmr1 = 0xb000;
	eieio();

	immr->cpm_gsmr_l1 = 0x00028034;	/* Enable Tx and Rx */
}

void
scccnprobe(cp)
	struct consdev *cp;
{
	cp->cn_dev = makedev(0, 0);
	cp->cn_pri = CN_REMOTE;
}

void
scccninit(cp)
	struct consdev *cp;
{
	scc_uart_init();
}

int
scccngetc(dev)
	dev_t dev;
{
	volatile struct cpm_bdesc *rxbd = (void *)0xff002820;	/* XXX */
	int c;

	for (;;) {
		if ((rxbd[0].bd_flags & 0x8000) == 0) {
			c = *rxbd[0].bd_addr;
			rxbd[0].bd_flags = 0x8000;
			break;
		}
		if ((rxbd[1].bd_flags & 0x8000) == 0) {
			c = *rxbd[1].bd_addr;
			rxbd[1].bd_flags = 0xa000;
			break;
		}
		sswt();
	}

	return c;
}

void
scccnputc(dev, c)
	dev_t dev;
	int c;
{
	volatile struct cpm_bdesc *txbd = (void *)0xff002830;	/* XXX */

	while (txbd->bd_flags & 0x8000)
		sswt();
	eieio();

	*(txbd->bd_addr) = c;
	txbd->bd_len = 1;

	eieio();
	txbd->bd_flags |= 0x8000;
}

struct consdev consdev_scc = {
	scccnprobe,
	scccninit,
	scccngetc,
	scccnputc,
	nullcnpollc,
	NULL
};

int
sccopen(dev, flag, mode, p)
	dev_t dev;
	int flag, mode;
	struct proc *p;
{
	struct scc_softc *sc = device_lookup(&scc_cd, SCCUNIT(dev));
	struct tty *tp;
	int s, error;

	if (sc == NULL)
		return ENXIO;

	if (ISSET(sc->sc_dev.dv_flags, DVF_ACTIVE) == 0)
		return ENXIO;

	tp = sc->sc_tty;

	if (ISSET(tp->t_state, TS_ISOPEN) &&
	    ISSET(tp->t_state, TS_XCLUDE) &&
	    p->p_ucred->cr_uid != 0)
		return EBUSY;

	s = spltty();

	/*
	 * Do the following if this is a first open.
	 */
	if (!ISSET(tp->t_state, TS_ISOPEN) && tp->t_wopen == 0) {
		struct termios t;

		tp->t_dev = dev;

		/*
		 * Initialize the termios status to the defaults.  Add in the
		 * sticky bits from TIOCSFLAGS.
		 */
		t.c_ispeed = 0;
		if (ISSET(sc->sc_hwflags, SCC_HW_CONSOLE)) {
			t.c_ospeed = sccconsspeed;
			t.c_cflag = sccconscflag;
		} else {
			t.c_ospeed = TTYDEF_SPEED;
			t.c_cflag = TTYDEF_CFLAG;
		}
		if (ISSET(sc->sc_swflags, TIOCFLAG_CLOCAL))
			SET(t.c_cflag, CLOCAL);
		if (ISSET(sc->sc_swflags, TIOCFLAG_CRTSCTS))
			SET(t.c_cflag, CRTSCTS);
		if (ISSET(sc->sc_swflags, TIOCFLAG_MDMBUF))
			SET(t.c_cflag, MDMBUF);

		/* Make sure comparam() will do something. */
		tp->t_ospeed = 0;
		sccparam(tp, &t);
		tp->t_iflag = TTYDEF_IFLAG;
		tp->t_oflag = TTYDEF_OFLAG;
		tp->t_lflag = TTYDEF_LFLAG;
		ttychars(tp);
		ttsetwater(tp);

#if notyet
		s2 = splserial();

		/*
		 * Turn on DTR.  We must always do this, even if carrier is not
		 * present, because otherwise we'd have to use TIOCSDTR
		 * immediately after setting CLOCAL, which applications do not
		 * expect.  We always assert DTR while the device is open
		 * unless explicitly requested to deassert it.
		 */
		scc_modem(sc, 1);

		splx(s2);
#endif
	}
	
	splx(s);

	error = ttyopen(tp, SCCDIALOUT(dev), ISSET(flag, O_NONBLOCK));
	if (error)
		goto bad;

	error = (*tp->t_linesw->l_open)(dev, tp);
	if (error)
		goto bad;

	return 0;

bad:
	if (!ISSET(tp->t_state, TS_ISOPEN) && tp->t_wopen == 0) {
		/*
		 * We failed to open the device, and nobody else had it opened.
		 * Clean up the state as appropriate.
		 */
//		scc_shutdown(sc);
	}

	return error;
}

int
sccclose(dev, flag, mode, p)
	dev_t dev;
	int flag, mode;
	struct proc *p;
{
	struct scc_softc *sc = device_lookup(&scc_cd, SCCUNIT(dev));
	struct tty *tp = sc->sc_tty;

	(*tp->t_linesw->l_close)(tp, flag);
	ttyclose(tp);

	return 0;
}

int
sccread(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	struct scc_softc *sc = device_lookup(&scc_cd, SCCUNIT(dev));
	struct tty *tp = sc->sc_tty;

	return (*tp->t_linesw->l_read)(tp, uio, flag);
}

int
sccwrite(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	struct scc_softc *sc = device_lookup(&scc_cd, SCCUNIT(dev));
	struct tty *tp = sc->sc_tty;

	return (*tp->t_linesw->l_write)(tp, uio, flag);
}

int
sccpoll(dev, events, p)
	dev_t dev;
	int events;
	struct proc *p;
{
	struct scc_softc *sc = device_lookup(&scc_cd, SCCUNIT(dev));
	struct tty *tp = sc->sc_tty;

	return (*tp->t_linesw->l_poll)(tp, events, p);
}

struct tty *
scctty(dev)
	dev_t dev;
{
	struct scc_softc *sc = device_lookup(&scc_cd, SCCUNIT(dev));
	struct tty *tp = sc->sc_tty;

	return tp;
}

int
sccioctl(dev, cmd, data, flag, p)
	dev_t dev;
	u_long cmd;
	caddr_t data;
	int flag;
	struct proc *p;
{
	struct scc_softc *sc = device_lookup(&scc_cd, SCCUNIT(dev));
	struct tty *tp = sc->sc_tty;
	int error;

	error = (*tp->t_linesw->l_ioctl)(tp, cmd, data, flag, p);
	if (error >= 0)
		return error;

	error = ttioctl(tp, cmd, data, flag, p);
	if (error >= 0)
		return error;

	return 0;
}

void
sccstop(tp, flag)
	struct tty *tp;
	int flag;
{
}



void
sccstart(tp)
	struct tty *tp;
{
	struct scc_softc *sc = device_lookup(&scc_cd, SCCUNIT(tp->t_dev));
	char *tba;
	int tbc, len;

	tba = tp->t_outq.c_cf;
	tbc = ndqb(&tp->t_outq, 0);

	for (len = tbc; len > 0; len--)
		scccnputc(NULL, *tba++);


	ndflush(&tp->t_outq, tbc);
//	(*tp->t_linesw->l_start)(tp);
}

int
sccparam(tp, t)
	struct tty *tp;
	struct termios *t;
{
	struct scc_softc *sc = device_lookup(&scc_cd, SCCUNIT(tp->t_dev));
	volatile struct mpc860dev *immr = get_immr();

	if (ISSET(sc->sc_swflags, TIOCFLAG_SOFTCAR) ||
	    ISSET(sc->sc_hwflags, SCC_HW_CONSOLE)) {
		    SET(t->c_cflag, CLOCAL);
		    CLR(t->c_cflag, HUPCL);
	}

	tp->t_ispeed = 0;
	tp->t_ospeed = t->c_ospeed;
	tp->t_cflag = t->c_cflag;

	if (ISSET(t->c_cflag, CLOCAL))
		immr->cpm_pcso &= ~0x0020;	/* Ignore CD1 */
	else
		immr->cpm_pcso |= 0x0020;	/* PC10 = CD1 */

	(*tp->t_linesw->l_modem)(tp, 1);

	return 0;
}

#define CPM_IVEC_PC15	0x1f
#define CPM_IVEC_SCC1	0x1e
#define CPM_IVEC_SCC2	0x1d
/* ... */

int
sccintr(v)
	void *v;
{
	struct scc_softc *sc = v;
	struct tty *tp = sc->sc_tty;
	int (*rint)(int, struct tty *) = tp->t_linesw->l_rint;
	volatile struct mpc860dev *immr = get_immr();
	volatile struct cpm_bdesc *txbd, *rxbd;
	int vec;
	int event, c;

	/* Check if this interrupt is mine. */
	vec = immr->cpm_civr >> 11;
	if (vec == 0) {			/* XXX FOR NOW */
		immr->cpm_civr |= 1;
		goto out;
	}
	if (vec != CPM_IVEC_SCC1)
		return 0;

	/* Acknowledge interrupt. */
	immr->cpm_civr |= 1;
	eieio();

	rxbd = (void *)0xff002820;
	txbd = (void *)0xff002830;

	event = immr->cpm_scce1;	/* SCC event */

/* printf("sccintr vec = %d, event = 0x%x\n", vec, event); */


#define SCCE_BUSY	0x0100

	if (event & SCCE_BUSY) {	/* XXX ? */
		if ((rxbd[0].bd_flags & 0x8000) == 0) {
			c = *rxbd[0].bd_addr;
			rxbd[0].bd_flags |= 0x8000;
			rint(c, tp);
		}
		if ((rxbd[1].bd_flags & 0x8000) == 0) {
			c = *rxbd[1].bd_addr;
			rxbd[1].bd_flags |= 0xa000;
			rint(c, tp);
		}
		goto out;
	}

out:
	immr->cpm_scce1 = event;

	/* Clear SCC1 interrupt condition. */
	immr->cpm_cisr |= 1 << CPM_IVEC_SCC1;

	return 1;
}

