/*	$Id: wdc_pbus.c,v 1.18 2004/01/10 08:15:12 tsubai Exp $	*/

/*-
 * Copyright (c) 2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/device.h>
#include <sys/malloc.h>
#include <sys/systm.h>

#include <machine/autoconf.h>
#include <machine/bus.h>
#include <machine/mpc860dev.h>
#include <machine/pio.h>

#include <uvm/uvm_extern.h>
#include <dev/ata/atavar.h>
#include <dev/ic/wdcvar.h>

struct wdc_pbus_softc {
	struct wdc_softc sc_wdcdev;
	struct wdc_channel *wdc_chanptr;
	struct wdc_channel wdc_channel;
	struct ata_queue wdc_queue;
};

int wdc_pbus_match(struct device *, struct cfdata *, void *);
void wdc_pbus_attach(struct device *, struct device *, void *);

struct cfattach wdc_pbus_ca = {
	"wdc_pbus", {}, sizeof(struct wdc_pbus_softc),
	wdc_pbus_match, wdc_pbus_attach
};

int
wdc_pbus_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;

	if (strcmp(ca->ca_name, "wdc") == 0)
		return 1;
	return 0;
}

void
wdc_pbus_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct wdc_pbus_softc *sc = (void *)self;
	struct cfdata *cf = self->dv_cfdata;
	struct wdc_channel *chp = &sc->wdc_channel;
	volatile struct mpc860dev *immr = get_immr();
	int vec, i;

	vec = cf->cf_vec;
	printf(": vec %d\n", vec);

	/*
	 * 0x80000010 data register
	 * 0x8020000c alt status register
	 */
	chp->cmd_iot = chp->ctl_iot = _make_bus_space_tag(0x80000000, 0);

	if (bus_space_map(chp->cmd_iot, 0x10, 16, 0, &chp->cmd_baseioh) ||
	    bus_space_map(chp->ctl_iot, 0x20000c, 1, 0, &chp->ctl_ioh)) {
		printf("%s: couldn't map registers\n", self->dv_xname);
		return;
	}
	/* XXX Note: bus_space_subregion() always succeeds. */
	for (i = 0; i < WDC_NREG; i++)
		bus_space_subregion(chp->cmd_iot, chp->cmd_baseioh,
		    i * 2, 2, &chp->cmd_iohs[i]);

	/* Configure IDE reset pin. */
	immr->cpm_pcpar &= ~0x0008;
	immr->cpm_pcdir |= 0x0008;
	immr->cpm_pcso &= ~0x0008;

	/* Reset IDE. */
	immr->cpm_pcdat &= ~0x0008;	/* assert */
	delay(100);
	immr->cpm_pcdat |= 0x0008;	/* negate */
	delay(10000);

	/* Wait for reset done. */
	while (bus_space_read_1(chp->cmd_iot, chp->cmd_iohs[7], 0) & 0x80);

	intr_establish(vec, IST_LEVEL, IPL_BIO, wdcintr, chp);

	sc->sc_wdcdev.PIO_cap = 4;
	sc->sc_wdcdev.cap |= WDC_CAPABILITY_DATA16;
	sc->wdc_chanptr = chp;
	sc->sc_wdcdev.channels = &sc->wdc_chanptr;
	sc->sc_wdcdev.nchannels = 1;

	chp->ch_channel = 0;
	chp->ch_wdc = &sc->sc_wdcdev;
	chp->ch_queue = &sc->wdc_queue;

	wdcattach(chp);
}
