/*	$Id: bus.h,v 1.3 2003/12/18 08:17:01 tsubai Exp $	*/

/*-
 * Copyright (c) 2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _MACHINE_BUS_H_
#define _MACHINE_BUS_H_

#define __BUS_SPACE_HAS_STREAM_METHODS

typedef u_long bus_addr_t;
typedef u_long bus_size_t;

typedef u_int bus_space_tag_t;
typedef u_long bus_space_handle_t;

/* bus_space_map flags */
#define BUS_SPACE_MAP_CACHEABLE		0x01
#define BUS_SPACE_MAP_LINEAR		0x02
#define BUS_SPACE_MAP_PREFETCHABLE	0x04

/* bus_space_barrier flags */
#define BUS_SPACE_BARRIER_READ	0x01		/* force read barrier */
#define BUS_SPACE_BARRIER_WRITE	0x02		/* force write barrier */

#define BUS_SPACE_ALIGNED_POINTER(p, t) ALIGNED_POINTER(p, t)

#define bus_space_read_stream_1(t, h, off) \
	bus_space_read_1(t, h, off)

#define bus_space_read_multi_stream_1(t, h, off, addr, count) \
	bus_space_read_multi_1(t, h, off, addr, count)

#define bus_space_read_region_stream_1(t, h, off, addr, count) \
	bus_space_read_region_1(t, h, off, addr, count)

#define bus_space_write_stream_1(t, h, off, val) \
	bus_space_write_1(t, h, off, val)

#define bus_space_write_multi_stream_1(t, h, off, addr, count) \
	bus_space_write_multi_1(t, h, off, addr, count)

#define bus_space_write_region_stream_1(t, h, off, addr, count) \
	bus_space_write_region_1(t, h, off, addr, count)

bus_space_tag_t _make_bus_space_tag(paddr_t, int);

int bus_space_map(bus_space_tag_t, bus_addr_t, bus_size_t, int,
    bus_space_handle_t *);
void bus_space_unmap(bus_space_tag_t, bus_space_handle_t, bus_size_t);
int bus_space_subregion(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    bus_size_t, bus_space_handle_t *);
paddr_t bus_space_mmap(bus_space_tag_t, bus_addr_t, off_t, int, int);
u_int8_t bus_space_read_1(bus_space_tag_t, bus_space_handle_t, bus_size_t);
u_int16_t bus_space_read_2(bus_space_tag_t, bus_space_handle_t, bus_size_t);
u_int32_t bus_space_read_4(bus_space_tag_t, bus_space_handle_t, bus_size_t);
void bus_space_write_1(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    u_int8_t);
void bus_space_write_2(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    u_int16_t);
void bus_space_write_4(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    u_int32_t);
void bus_space_barrier(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    bus_size_t, int);
void bus_space_read_region_1(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    u_int8_t *, bus_size_t);
void bus_space_read_region_2(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    u_int16_t *, bus_size_t);
void bus_space_read_region_4(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    u_int32_t *, bus_size_t);
void bus_space_write_region_1(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    const u_int8_t *, bus_size_t);
void bus_space_write_region_2(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    const u_int16_t *, bus_size_t);
void bus_space_write_region_4(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    const u_int32_t *, bus_size_t);
void bus_space_set_region_1(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    u_int8_t, bus_size_t);
void bus_space_set_region_2(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    u_int16_t, bus_size_t);
void bus_space_set_region_4(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    u_int32_t, bus_size_t);
void bus_space_read_multi_1(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    u_int8_t *, bus_size_t);
void bus_space_read_multi_2(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    u_int16_t *, bus_size_t);
void bus_space_read_multi_4(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    u_int32_t *, bus_size_t);
void bus_space_write_multi_1(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    const u_int8_t *, bus_size_t);
void bus_space_write_multi_2(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    const u_int16_t *, bus_size_t);
void bus_space_write_multi_4(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    const u_int32_t *, bus_size_t);
u_int8_t bus_space_read_stream_1(bus_space_tag_t, bus_space_handle_t,
    bus_size_t);
u_int16_t bus_space_read_stream_2(bus_space_tag_t, bus_space_handle_t,
    bus_size_t);
u_int32_t bus_space_read_stream_4(bus_space_tag_t, bus_space_handle_t,
    bus_size_t);
void bus_space_read_multi_stream_1(bus_space_tag_t, bus_space_handle_t,
    bus_size_t, u_int8_t *, bus_size_t);
void bus_space_read_multi_stream_2(bus_space_tag_t, bus_space_handle_t,
    bus_size_t, u_int16_t *, bus_size_t);
void bus_space_read_multi_stream_4(bus_space_tag_t, bus_space_handle_t,
    bus_size_t, u_int32_t *, bus_size_t);
void bus_space_read_region_stream_1(bus_space_tag_t, bus_space_handle_t,
    bus_size_t, u_int8_t *, bus_size_t);
void bus_space_read_region_stream_2(bus_space_tag_t, bus_space_handle_t,
    bus_size_t, u_int16_t *, bus_size_t);
void bus_space_read_region_stream_4(bus_space_tag_t, bus_space_handle_t,
    bus_size_t, u_int32_t *, bus_size_t);
void bus_space_write_stream_1(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    u_int8_t);
void bus_space_write_stream_2(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    u_int16_t);
void bus_space_write_stream_4(bus_space_tag_t, bus_space_handle_t, bus_size_t,
    u_int32_t);
void bus_space_write_multi_stream_1(bus_space_tag_t, bus_space_handle_t,
    bus_size_t, const u_int8_t *, bus_size_t);
void bus_space_write_multi_stream_2(bus_space_tag_t, bus_space_handle_t,
    bus_size_t, const u_int16_t *, bus_size_t);
void bus_space_write_multi_stream_4(bus_space_tag_t, bus_space_handle_t,
    bus_size_t, const u_int32_t *, bus_size_t);
void bus_space_write_region_stream_1(bus_space_tag_t, bus_space_handle_t,
    bus_size_t, const u_int8_t *, bus_size_t);
void bus_space_write_region_stream_2(bus_space_tag_t, bus_space_handle_t,
    bus_size_t, const u_int16_t *, bus_size_t);
void bus_space_write_region_stream_4(bus_space_tag_t, bus_space_handle_t,
    bus_size_t, const u_int32_t *, bus_size_t);

#endif /* _MACHINE_BUS_H_ */
