/*	$Id: pte.h,v 1.4 2002/06/04 18:18:20 tsubai Exp $	*/

/*-
 * Copyright (c) 2001 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define CTR_GPM		0x80000000	/* Group protection mode */
#define CTR_PPM		0x40000000	/* Page protection mode */
#define CTR_CIDEF	0x20000000	/* CI default when MMU is disabled */
#define CTR_WTDEF	0x10000000	/* WT default when MMU is disabled */
#define CTR_RSV4	0x08000000	/* Reserve four TLB entries */
#define CTR_TWAM	0x04000000	/* Tablewalk assist mode (D) */
#define CTR_PPCS	0x02000000	/* Priv/user state compare mode */
#define CTR_TLB_INDX	0x00001f00	/* TLB index to be loaded */

#define EPN_EPN		0xfffff000	/* Effective page number */
#define EPN_EV		0x00000200	/* TLB entry valid */
#define EPN_ASID	0x0000000f	/* ASID */

#define TWC_L2TB	0xfffff000	/* Tablewalk level 2 table base (D) */
#define TWC_L2APG	0x000001e0	/* Access protection group */
#define TWC_G		0x00000010	/* Guarded */
#define TWC_PS		0x0000000c	/* Level 1 page size */
#define TWC_PS_SMALL		0x00000000	/* 4K or 16K */
#define TWC_PS_512K		0x00000004
#define TWC_PS_8M		0x0000000c
#define TWC_WT		0x00000002	/* Writethrough (D) */
#define TWC_V		0x00000001	/* Valid */

#define RPN_RPN		0xfffff000	/* Real page number */
#define RPN_PP		0x00000ff0
#define RPN_PP_ACC	0x00000e00	/* Access modes */
#define RPN_M		0x00000100	/* Modified bit (D) */
#define RPN_PP_SV	0x000000f0	/* Subpage validity */
#define RPN_SPS		0x00000008	/* Small page size */
#define RPN_SH		0x00000004	/* Shared page */
#define RPN_CI		0x00000002	/* Cache inhibit */
#define RPN_V		0x00000001	/* Vaild */

#define RPN_4K		RPN_PP_SV
#define RPN_16K		(RPN_PP_SV | RPN_SPS)
#define RPN_512K	(RPN_PP_SV | RPN_SPS)
#define RPN_8M		(RPN_PP_SV | RPN_SPS)

#define PG_V		0x001
#define PG_CI		0x002
#define PG_SH		0x004
#define PG_M		0x100
#define PG_4K		0x0f0
#define PG_16K		0x0f8
#define PG_512K		0x0f8
#define PG_8M		0x0f8
				/* kern	user */
#define PG_KR		0xc00	/* RO	RO   */
#define PG_KW		0x000	/* RW	--   */
#define PG_UR		0x400	/* RW	RO   */
#define PG_UW		0x800	/* RW	RW   */

#define PG_FRAME	0xfffff000

#ifndef _LOCORE
typedef u_int pt_entry_t;
typedef u_int pd_entry_t;
#endif

/* for now */
#define DSISR_PROT	0x80000000	/* XXX ? */
#define DSISR_MISS	0x40000000
#define DSISR_STORE	0x02000000
#define DSISR_ALL	(DSISR_PROT|DSISR_MISS|DSISR_STORE)

#define PDSHIFT		22		/* offset of PD index in VA */
#define NBPD		(1 << PDSHIFT)	/* # bytes mapped by PD (4MB) */
#define PDOFSET		(NBPD-1)	/* mask for non-PD part of VA */
#define PTES_PER_PTP	(NBPD / NBPG)	/* # of PTEs in a PTP */
#define PD_MASK		0xffc00000	/* page directory address bits */
#define PT_MASK		0x003ff000	/* page table address bits */

/* XXX compat i386 */
#define PG_u		PG_UW
#define PG_G		PG_SH
#define PG_U		0
#define PG_RO		0
#define PG_RW		PG_AVAIL3
#define PG_N		PG_CI
#define PG_PROT		(PG_RW|PG_M|0xc00)

#define PG_AVAIL1	0x010
#define PG_AVAIL2	0x020
#define PG_AVAIL3	0x040
#define PG_AVAIL4	0x080		/* ref */
