/*	$Id: autoconf.c,v 1.7 2002/10/06 16:27:27 tsubai Exp $	*/

/*-
 * Copyright (c) 2001,2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/device.h>
#include <sys/kernel.h>
#include <sys/systm.h>

#include <machine/autoconf.h>
#include <machine/mpc860dev.h>

struct device *booted_device;	/* boot device */
int booted_partition;		/* ...and partition on that device */

char fec_enaddr[6], sce_enaddr[6];

#define M24C16W_ETHER_OFFSET	0x7d8
#define M24C16W_ADDR(off)	(0xa0 | ((off) & 0x700) >> 7)
#define M24C16W_OFF(off)	((off) & 0xff)

/*
 * Determine device configuration for a machine.
 */
void
cpu_configure()
{
	int msr;
	u_int64_t enaddr;

	timebase_init();

	i2c_read(M24C16W_ADDR(M24C16W_ETHER_OFFSET),
	    M24C16W_ETHER_OFFSET & 0xff, 6, &enaddr);
	hostid = enaddr >> 16 & 0xffffff;

	bcopy(&enaddr, fec_enaddr, 6);
	enaddr += 0x10000;
	bcopy(&enaddr, sce_enaddr, 6);

	if (config_rootfound("mainbus", NULL) == NULL)
		panic("configure: mainbus not configured");

	spl0();

	/* Enable hardware interrupts. */
	asm volatile ("mfmsr %0" : "=r"(msr));
	msr |= PSL_EE|PSL_RI;
	asm volatile ("mtmsr %0" :: "r"(msr));
}

void
cpu_rootconf()
{
	volatile struct mpc860dev *immr = get_immr();
	struct device *dv;
	char *bootdev;
	int dipsw;

	/*
	 * DIP-SW configuration
	 * netboot:
	 *   1 ON, 2 ON
	 * CF: (IDE)
	 *   1 ON, 2 OFF
	 * firmware update:
	 *   1 OFF, 2 ON
	 */
	dipsw = (immr->cpm_padat & 0xc000) >> 14;

	if (dipsw & 1)		/* SW1-2 = OFF */
		bootdev = "wd0";
	else			/* SW1-2 = ON */
		bootdev = "fec0";

	for (dv = TAILQ_FIRST(&alldevs); dv; dv = TAILQ_NEXT(dv, dv_list))
		if (strcmp(dv->dv_xname, bootdev) == 0)
			break;

	booted_device = dv;
	booted_partition = 0;

	printf("boot device: %s\n",
	    booted_device ? booted_device->dv_xname : "<unknown>");

	setroot(booted_device, booted_partition);
}
