/*	$Id: bus_space.c,v 1.5 2003/12/18 10:20:36 tsubai Exp $	*/

/*-
 * Copyright (c) 2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>

#include <machine/bus.h>
#include <machine/pio.h>

#define __BA(t, h, o) ((void *)((h) + (o)))
#define BUS_ADDR_MASK	0xffffffff
#define BUS_STRIDE_MASK	0x00000000

#ifdef BUS_SPACE_DEBUG
#define DPRINTF printf
#else
#define DPRINTF while (0) printf
#endif

void *mapiodev(paddr_t, psize_t);	/* XXX */

bus_space_tag_t
_make_bus_space_tag(base, stride)
	paddr_t base;
	int stride;
{
	bus_space_tag_t t;

	t = (base & BUS_ADDR_MASK) | stride;
	DPRINTF("_make_bus_space_tag: 0x%08x\n", t);
	return t;
}

int
bus_space_map(t, addr, size, flags, hp)
	bus_space_tag_t t;
	bus_addr_t addr;
	bus_size_t size;
	int flags;
	bus_space_handle_t *hp;
{
	paddr_t base = t & BUS_ADDR_MASK;
	int stride = t & BUS_STRIDE_MASK;

	addr <<= stride;
	size <<= stride;

	DPRINTF("bus_space_map: 0x%lx, 0x%lx\n", base + addr, size);
	*hp = (bus_space_handle_t)mapiodev(base + addr, size);
	return 0;
}

void
bus_space_unmap(t, h, size)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t size;
{
	DPRINTF("bus_space_unmap\n");
}

int
bus_space_subregion(t, h, off, size, nhp)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off, size;
	bus_space_handle_t *nhp;
{
	*nhp = (bus_space_tag_t)__BA(t, h, off);
	return 0;
}

#if 0
int
bus_space_alloc(t, rstart, rend, size, align, boundary, flags, addrp, hp)
	bus_space_tag_t t;
	bus_addr_t rstart, rend;
	bus_size_t size, align, boundary;
	int flags;
	bus_addr_t *addrp;
	bus_space_handle_t *hp;
{
	panic("bus_space_alloc: not implemented");
}
#endif

#if 0
void
bus_space_free(t, h, size)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t size;
{
	panic("bus_space_free: not implemented");
}
#endif

paddr_t
bus_space_mmap(t, addr, off, prot, flags)
	bus_space_tag_t t;
	bus_addr_t addr;
	off_t off;
	int prot, flags;
{
	paddr_t base = t & BUS_ADDR_MASK;
	int stride = t & BUS_STRIDE_MASK;

	if (flags & BUS_SPACE_MAP_LINEAR) {
		if (stride != 0)
			return -1;
	}
	return base + ((addr + off) << stride);
}

#if 1 /* 16bit big-endian bus (XXX for now) */

u_int8_t
bus_space_read_1(t, h, off)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
{
	return in16(__BA(t, h, off));
}

u_int16_t
bus_space_read_2(t, h, off)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
{
	return in16(__BA(t, h, off));
}

void
bus_space_read_multi_2(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int16_t *addr;
	bus_size_t count;
{
	volatile u_int16_t *s = __BA(t, h, off);

	while (count-- > 0)
		asm volatile ("lhzx %0,0,%1" : "=r"(*addr++) : "r"(s));
	asm volatile ("sync");
}

void
bus_space_read_multi_4(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int32_t *addr;
	bus_size_t count;
{
	volatile u_int32_t *s = __BA(t, h, off);

	while (count-- > 0)
		asm volatile ("lwzx %0,0,%1" : "=r"(*addr++) : "r"(s));
	asm volatile ("sync");
}

void
bus_space_read_multi_stream_2(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int16_t *addr;
	bus_size_t count;
{
	volatile u_int16_t *s = __BA(t, h, off);

	while (count-- > 0)
		*addr ++ = *s;
	asm volatile ("sync");
}

void
bus_space_read_multi_stream_4(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int32_t *addr;
	bus_size_t count;
{
	volatile u_int32_t *s = __BA(t, h, off);

	while (count-- > 0)
		*addr ++ = *s;
	asm volatile ("sync");
}

void
bus_space_write_1(t, h, off, val)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int8_t val;
{
	out16(__BA(t, h, off), val);
}

void
bus_space_write_multi_2(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int16_t *addr;
	bus_size_t count;
{
	volatile u_int16_t *d = __BA(t, h, off);

	while (count-- > 0)
		asm volatile ("sthx %0,0,%1" :: "r"(*addr++), "r"(d));
	asm volatile ("sync");
}

void
bus_space_write_multi_4(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int32_t *addr;
	bus_size_t count;
{
	volatile u_int32_t *d = __BA(t, h, off);

	while (count-- > 0)
		asm volatile ("stwx %0,0,%1" :: "r"(*addr++), "r"(d));
	asm volatile ("sync");
}

void
bus_space_write_multi_stream_2(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int16_t *addr;
	bus_size_t count;
{
	volatile u_int16_t *d = __BA(t, h, off);

	while (count-- > 0)
		*d = *addr++;
	asm volatile ("sync");
}

void
bus_space_write_multi_stream_4(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int32_t *addr;
	bus_size_t count;
{
	volatile u_int32_t *d = __BA(t, h, off);

	while (count-- > 0)
		*d = *addr++;
	asm volatile ("sync");
}

#else /* little-endian bus_space */
u_int8_t
bus_space_read_1(t, h, off)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
{
	return in8(__BA(t, h, off));
}

u_int16_t
bus_space_read_2(t, h, off)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
{
	return in16rb(__BA(t, h, off));
}

u_int32_t
bus_space_read_4(t, h, off)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
{
	return in32rb(__BA(t, h, off));
}

#if 0
u_int64_t
bus_space_read_8(t, h, off)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
{
	panic("bus_space_read_8: not implemented");
}
#endif

void
bus_space_write_1(t, h, off, val)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int8_t val;
{
	out8(__BA(t, h, off), val);
}

void
bus_space_write_2(t, h, off, val)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int16_t val;
{
	out16rb(__BA(t, h, off), val);
}

void
bus_space_write_4(t, h, off, val)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int32_t val;
{
	out32rb(__BA(t, h, off), val);
}

#if 0
void
bus_space_write_8(t, h, off, val)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int64_t val;
{
	panic("bus_space_write_8: not implemented");
}
#endif

void
bus_space_barrier(t, h, off, len, flags)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off, len;
	int flags;
{
	asm volatile ("sync");
}

void
bus_space_read_region_1(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int8_t *addr;
	bus_size_t count;
{
	int stride = 1 << (t & BUS_STRIDE_MASK);
	volatile u_int8_t *s = __BA(t, h, off);

	while (count-- > 0) {
		*addr++ = *s;
		s += stride;
	}
	asm volatile ("sync");
}

void
bus_space_read_region_2(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int16_t *addr;
	bus_size_t count;
{
	int stride = 1 << (t & BUS_STRIDE_MASK);
	vaddr_t s = (vaddr_t)__BA(t, h, off);

	if (stride < 2)
		stride = 2;
	while (count-- > 0) {
		asm volatile ("lhbrx %0,0,%1" : "=r"(*addr) : "r"(s));
		addr++;
		s += stride;
	}
	asm volatile ("sync");
}

void
bus_space_read_region_4(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int32_t *addr;
	bus_size_t count;
{
	int stride = 1 << (t & BUS_STRIDE_MASK);
	vaddr_t s = (vaddr_t)__BA(t, h, off);

	if (stride < 4)
		stride = 4;
	while (count-- > 0) {
		asm volatile ("lwbrx %0,0,%1" : "=r"(*addr) : "r"(s));
		addr++;
		s += stride;
	}
	asm volatile ("sync");
}

#if 0
void
bus_space_read_region_8(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int32_t *addr;
	bus_size_t count;
{
	panic("bus_space_read_region_8: not implemented");
}
#endif

void
bus_space_write_region_1(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int8_t *addr;
	bus_size_t count;
{
	int stride = 1 << (t & BUS_STRIDE_MASK);
	volatile u_int8_t *d = __BA(t, h, off);

	while (count-- > 0) {
		*d = *addr++;
		d += stride;
	}
	asm volatile ("sync");
}

void
bus_space_write_region_2(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int16_t *addr;
	bus_size_t count;
{
	int stride = 1 << (t & BUS_STRIDE_MASK);
	vaddr_t d = (vaddr_t)__BA(t, h, off);

	if (stride < 2)
		stride = 2;
	while (count-- > 0) {
		asm volatile ("sthbrx %0,0,%1" :: "r"(*addr), "r"(d));
		addr++;
		d += stride;
	}
	asm volatile ("sync");
}

void
bus_space_write_region_4(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int32_t *addr;
	bus_size_t count;
{
	int stride = 1 << (t & BUS_STRIDE_MASK);
	vaddr_t d = (vaddr_t)__BA(t, h, off);

	if (stride < 4)
		stride = 4;
	while (count-- > 0) {
		asm volatile ("stwbrx %0,0,%1" :: "r"(*addr), "r"(d));
		addr++;
		d += stride;
	}
	asm volatile ("sync");
}

#if 0
void
bus_space_write_region_8(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int32_t *addr;
	bus_size_t count;
{
	panic("bus_space_write_region_8: not implemented");
}
#endif

void
bus_space_set_region_1(t, h, off, val, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int8_t val;
	bus_size_t count;
{
	int stride = 1 << (t & BUS_STRIDE_MASK);
	volatile u_int8_t *d = __BA(t, h, off);

	while (count-- > 0) {
		*d = val;
		d += stride;
	}
	asm volatile ("sync");
}

void
bus_space_set_region_2(t, h, off, val, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int16_t val;
	bus_size_t count;
{
	int stride = 1 << (t & BUS_STRIDE_MASK);
	vaddr_t d = (vaddr_t)__BA(t, h, off);

	if (stride < 2)
		stride = 2;
	while (count-- > 0) {
		asm volatile ("sthbrx %0,0,%1" :: "r"(val), "r"(d));
		d += stride;
	}
	asm volatile ("sync");
}

void
bus_space_set_region_4(t, h, off, val, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int32_t val;
	bus_size_t count;
{
	int stride = 1 << (t & BUS_STRIDE_MASK);
	vaddr_t d = (vaddr_t)__BA(t, h, off);

	if (stride < 4)
		stride = 4;
	while (count-- > 0) {
		asm volatile ("stwbrx %0,0,%1" :: "r"(val), "r"(d));
		d += stride;
	}
	asm volatile ("sync");
}

#if 0
void
bus_space_set_region_8(t, h, off, val, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int32_t val;
	bus_size_t count;
{
	panic("bus_space_set_region_8: not implemented");
}
#endif

void
bus_space_read_multi_1(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int8_t *addr;
	bus_size_t count;
{
	volatile u_int8_t *s = __BA(t, h, off);

	while (count-- > 0)
		*addr++ = *s;
	asm volatile ("sync");
}

void
bus_space_read_multi_2(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int16_t *addr;
	bus_size_t count;
{
	volatile u_int16_t *s = __BA(t, h, off);

	while (count-- > 0)
		asm volatile ("lhbrx %0,0,%1" : "=r"(*addr++) : "r"(s));
	asm volatile ("sync");
}

void
bus_space_read_multi_4(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int32_t *addr;
	bus_size_t count;
{
	volatile u_int32_t *s = __BA(t, h, off);

	while (count-- > 0)
		asm volatile ("lwbrx %0,0,%1" : "=r"(*addr++) : "r"(s));
	asm volatile ("sync");
}

#if 0
void
bus_space_read_multi_8(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int64_t *addr;
	bus_size_t count;
{
	panic("bus_space_read_multi_8: not implemented");
}
#endif

void
bus_space_write_multi_1(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int8_t *addr;
	bus_size_t count;
{
	volatile u_int8_t *d = __BA(t, h, off);

	while (count-- > 0)
		*d = *addr++;
	asm volatile ("sync");
}

void
bus_space_write_multi_2(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int16_t *addr;
	bus_size_t count;
{
	volatile u_int16_t *d = __BA(t, h, off);

	while (count-- > 0)
		asm volatile ("sthbrx %0,0,%1" :: "r"(*addr++), "r"(d));
	asm volatile ("sync");
}

void
bus_space_write_multi_4(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int32_t *addr;
	bus_size_t count;
{
	volatile u_int32_t *d = __BA(t, h, off);

	while (count-- > 0)
		asm volatile ("stwbrx %0,0,%1" :: "r"(*addr++), "r"(d));
	asm volatile ("sync");
}

#if 0
void
bus_space_write_multi_8(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int64_t *addr;
	bus_size_t count;
{
	panic("bus_space_write_multi_8: not implemented");
}
#endif

u_int16_t
bus_space_read_stream_2(t, h, off)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
{
	return in16(__BA(t, h, off));
}

u_int32_t
bus_space_read_stream_4(t, h, off)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
{
	return in32(__BA(t, h, off));
}

#if 0
u_int64_t
bus_space_read_stream_8(t, h, off)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
{
	panic("bus_space_read_stream_8: not implemented");
}
#endif

void
bus_space_read_multi_stream_2(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int16_t *addr;
	bus_size_t count;
{
	volatile u_int16_t *s = __BA(t, h, off);

	while (count-- > 0)
		*addr ++ = *s;
	asm volatile ("sync");
}

void
bus_space_read_multi_stream_4(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int32_t *addr;
	bus_size_t count;
{
	volatile u_int32_t *s = __BA(t, h, off);

	while (count-- > 0)
		*addr ++ = *s;
	asm volatile ("sync");
}

#if 0
void
bus_space_read_multi_stream_8(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int64_t *addr;
	bus_size_t count;
{
	panic("bus_space_read_multi_stream_8: not implemented");
}
#endif

void
bus_space_read_region_stream_2(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int16_t *addr;
	bus_size_t count;
{
	int stride = 1 << (t & BUS_STRIDE_MASK);
	volatile u_int16_t *s = __BA(t, h, off);

	stride /= sizeof(u_int16_t);
	if (stride < 1)
		stride = 1;
	while (count-- > 0) {
		*addr++ = *s;
		s += stride;
	}
	asm volatile ("sync");
}

void
bus_space_read_region_stream_4(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int32_t *addr;
	bus_size_t count;
{
	int stride = 1 << (t & BUS_STRIDE_MASK);
	volatile u_int32_t *s = __BA(t, h, off);

	stride /= sizeof(u_int32_t);
	if (stride < 1)
		stride = 1;
	while (count-- > 0) {
		*addr++ = *s;
		s += stride;
	}
	asm volatile ("sync");
}

#if 0
void
bus_space_read_region_stream_8(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int32_t *addr;
	bus_size_t count;
{
	panic("bus_space_read_region_stream_8: not implemented");
}
#endif

void
bus_space_write_stream_2(t, h, off, val)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int16_t val;
{
	return out16(__BA(t, h, off), val);
}

void
bus_space_write_stream_4(t, h, off, val)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int32_t val;
{
	return out32(__BA(t, h, off), val);
}

#if 0
void
bus_space_write_stream_8(t, h, off, val)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	u_int64_t val;
{
	panic("bus_space_write_stream_8: not implemented");
}
#endif

void
bus_space_write_multi_stream_2(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int16_t *addr;
	bus_size_t count;
{
	volatile u_int16_t *d = __BA(t, h, off);

	while (count-- > 0)
		*d = *addr++;
	asm volatile ("sync");
}

void
bus_space_write_multi_stream_4(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int32_t *addr;
	bus_size_t count;
{
	volatile u_int32_t *d = __BA(t, h, off);

	while (count-- > 0)
		*d = *addr++;
	asm volatile ("sync");
}

#if 0
void
bus_space_write_multi_stream_8(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int64_t *addr;
	bus_size_t count;
{
	panic("bus_space_write_multi_stream_8: not implemented");
}
#endif

void
bus_space_write_region_stream_2(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int16_t *addr;
	bus_size_t count;
{
	int stride = 1 << (t & BUS_STRIDE_MASK);
	volatile u_int16_t *d = __BA(t, h, off);

	stride /= sizeof(u_int16_t);
	if (stride < 1)
		stride = 1;
	while (count-- > 0) {
		*d = *addr++;
		d += stride;
	}
	asm volatile ("sync");
}

void
bus_space_write_region_stream_4(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int32_t *addr;
	bus_size_t count;
{
	int stride = 1 << (t & BUS_STRIDE_MASK);
	volatile u_int32_t *d = __BA(t, h, off);

	stride /= sizeof(u_int32_t);
	if (stride < 1)
		stride = 1;
	while (count--)
		*d++ = *addr++;
	asm volatile ("sync");
}

#if 0
void
bus_space_write_region_stream_8(t, h, off, addr, count)
	bus_space_tag_t t;
	bus_space_handle_t h;
	bus_size_t off;
	const u_int32_t *addr;
	bus_size_t count;
{
	panic("bus_space_write_region_stream_8: not implemented");
}
#endif
#endif
