/*	$Id: locore.S,v 1.6 2003/01/25 11:33:55 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_ddb.h"
#include "opt_lockdebug.h"
#include "assym.h"

#include <sys/syscall.h>

#include <machine/param.h>
#include <machine/asm.h>
#include <machine/pte.h>
#include <machine/psl.h>
#include <machine/spr.h>
#include <machine/trap.h>

/*
 * Some instructions gas doesn't understand (yet?)
 */
#define	bdneq	bdnzf 2,

#define	INTSTK   8192			/* 8K interrupt stack */

/*
 * Globals
 */
	.data
GLOBAL(esym)
	.long	0			/* end of symbol table */
GLOBAL(proc0paddr)
	.long	0			/* proc0 p_addr */

GLOBAL(intrnames)
	.asciz	"irq0", "irq1", "irq2", "irq3"
	.asciz	"irq4", "irq5", "irq6", "irq7"
	.asciz	"irq8", "irq9", "irq10", "irq11"
	.asciz	"irq12", "irq13", "irq14", "irq15"
	.asciz	"clock"
GLOBAL(eintrnames)
	.align	4
GLOBAL(intrcnt)
	.long	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.long	0,0,0,0
GLOBAL(eintrcnt)

GLOBAL(powersave)
	.long	0

/*
 * File-scope for locore.S
 */
idle_u:
	.long	0			/* fake uarea during idle after exit */

/*
 * This symbol is here for the benefit of kvm_mkdb, and is supposed to
 * mark the start of kernel text.
 */
	.text
	.globl	_C_LABEL(kernel_text)
_C_LABEL(kernel_text):

	.globl	__start
__start:

	/* Invalidate all TLBs. */
	li	0,0
	lis	7,CTR_TWAM@h
	mtspr	MD_CTR,7
	mtspr	MI_CTR,0
	tlbia

	mtspr	M_CASID,0

	lis	7,0x5555
	ori	7,7,0x5555
	mtspr	MD_AP,7
	mtspr	MI_AP,7

	mr	30,3
	mr	31,4
	bl	inittlb
	mr	3,30
	mr	4,31

	bl	1f
1:	mflr	7
	oris	7,7,0x8000
	addi	7,7,2f-1b
	mtlr	7
	blr
2:

	li	7,PGOFSET
	add	4,4,7
	andc	4,4,7			/* esym = roundup(esym, NBPG) */

	lis	7,idle_u@ha
	stw	4,idle_u@l(7)
	addi	4,4,USPACE

	lis	7,_C_LABEL(proc0paddr)@ha
	stw	4,_C_LABEL(proc0paddr)@l(7)
	addi	4,4,USPACE

	addi	1,4,-FRAMELEN

	li	0,0
	stw	0,0(1)

	bl	_C_LABEL(initmpc860)
	bl	_C_LABEL(main)
1:
	b	1b


/*
 * No processes are runnable, so loop waiting for one.
 * Separate label here for accounting purposes.
 * When we get here, interrupts are off (MSR[EE]=0) and sched_lock is held.
 */
ASENTRY(Idle)
	lis	8,_C_LABEL(sched_whichqs)@ha
	lwz	9,_C_LABEL(sched_whichqs)@l(8)

	or.	9,9,9
	bne+	.Lsw1			/* at least one queue non-empty */

#if defined(MULTIPROCESSOR) || defined(LOCKDEBUG)
	bl	_C_LABEL(sched_unlock_idle)
#endif

	mfmsr	3
	ori	3,3,PSL_EE@l		/* reenable ints again */
	mtmsr	3
	isync

/* Check if we can use power saving mode */
	lis	8,_C_LABEL(powersave)@ha
	lwz	9,_C_LABEL(powersave)@l(8)

	or.	9,9,9
	beq	1f

	sync
	oris	3,3,PSL_POW@h		/* enter power saving mode */
	mtmsr	3
	isync
1:
	andi.	3,3,~PSL_EE@l		/* disable interrupts while
					   manipulating runque */
	mtmsr	3

#if defined(MULTIPROCESSOR) || defined(LOCKDEBUG)
	bl	_C_LABEL(sched_lock_idle)
#endif
	b	_ASM_LABEL(Idle)

/*
 * switchexit(struct lwp *l, void (*exit)(struct lwp *))
 * switchexit gets called from cpu_exit to complete the exit procedure.
 */
ENTRY(switchexit)
	mtctr	4
/* First switch to the idle pcb/kernel stack */
#if defined(MULTIPROCESSOR)
	GET_CPUINFO(7)
	lwz	6,CI_IDLE_PCB(7)
	stw	6,CI_CURPCB(7)
#else
	lis	6,idle_u@ha
	lwz	6,idle_u@l(6)
	lis	7,_C_LABEL(curpcb)@ha
	stw	6,_C_LABEL(curpcb)@l(7)
#endif
	addi	1,6,USPACE-16		/* 16 bytes are reserved at stack top */
	/*
	 * Schedule the vmspace and stack to be freed (the proc arg is
	 * already in r3).
	 */
	bctrl

#if defined(MULTIPROCESSOR) || defined(LOCKDEBUG)
	bl	_C_LABEL(sched_lock_idle)
#endif

/* Fall through to cpu_switch to actually select another proc */
	li	3,0			/* indicate exited process */

/*
 * void cpu_switch(struct lwp *l)
 * Find a runnable process and switch to it.
 */
/* XXX noprofile?  --thorpej@netbsd.org */
ENTRY(cpu_switch)
	mflr	0			/* save lr */
	stw	0,4(1)
	stwu	1,-16(1)
	stw	31,12(1)
	stw	30,8(1)

	mr	30,3
	lis	3,_C_LABEL(curlwp)@ha
	li	31,0
	stw	31,_C_LABEL(curlwp)@l(3) /* Zero to not accumulate cpu time */
	lis	3,_C_LABEL(curpcb)@ha
	lwz	31,_C_LABEL(curpcb)@l(3)

#if defined(MULTIPROCESSOR) || defined(LOCKDEBUG)
/* Release the sched_lock before processing interrupts. */
	bl	_C_LABEL(sched_unlock_idle)
#endif

	li	3,0
	bl	_C_LABEL(lcsplx)
	stw	3,PCB_SPL(31)		/* save spl */

/* Lock the scheduler. */
	mfmsr	3
	andi.	3,3,~PSL_EE@l		/* disable interrupts while
					   manipulating runque */
	mtmsr	3
	isync
#if defined(MULTIPROCESSOR) || defined(LOCKDEBUG)
	bl	_C_LABEL(sched_lock_idle)
#endif

/* Find a new process */
	lis	8,_C_LABEL(sched_whichqs)@ha
	lwz	9,_C_LABEL(sched_whichqs)@l(8)

	cmpwi	9,0
	beq-	_ASM_LABEL(Idle)	/* all queues empty */
.Lsw1:
	cntlzw	10,9
	lis	4,_C_LABEL(sched_qs)@ha
	addi	4,4,_C_LABEL(sched_qs)@l
	slwi	3,10,3
	add	3,3,4			/* select queue */

	lwz	31,L_FORW(3)		/* unlink first proc from queue */
	lwz	4,L_FORW(31)
	stw	4,L_FORW(3)
	stw	3,L_BACK(4)

	cmplw	3,4			/* queue empty? */
	bne	switch_common

	lis	3,0x80000000@h
	srw	3,3,10
	andc	9,9,3
	stw	9,_C_LABEL(sched_whichqs)@l(8) /* mark it empty */

switch_common:
	/* just did this resched thing */
	li	3,0
	lis	4,_C_LABEL(want_resched)@ha
	stw	3,_C_LABEL(want_resched)@l(4)

	stw	3,L_BACK(31)		/* probably superfluous */

#if defined(MULTIPROCESSOR) || defined(LOCKDEBUG)
	/* Unlock the sched_lock, but leave interrupts off, for now. */
	bl	_C_LABEL(sched_unlock_idle)
#endif

	/* Process now running on a processor. */
	li	3,LSONPROC		/* l->l_stat = LSONPROC */
	stw	3,L_STAT(31)

	/* record new process */
	lis	4,_C_LABEL(curlwp)@ha
	stw	31,_C_LABEL(curlwp)@l(4)

	lwz	4,L_ADDR(31)

	mfmsr	3
	ori	3,3,PSL_EE@l		/* Now we can interrupt again */
	mtmsr	3

	li	3,0
	cmplw	31,30			/* is it the same process? */
	beq	switch_return

	cmpwi	30,0			/* old process was exiting? */
	beq	switch_exited

	mfcr	11			/* save cr */
	mr	12,2			/* save r2 */
	stwu	1,-SFRAMELEN(1)		/* still running on old stack */
	stmw	10,8(1)
	lwz	3,L_ADDR(30)
	stw	1,PCB_SP(3)		/* save SP */

switch_exited:
	mfmsr	3
	andi.	3,3,~PSL_EE@l		/* disable interrupts while
					   actually switching */
	mtmsr	3

	/* indicate new pcb */
	lis	5,_C_LABEL(curpcb)@ha
	stw	4,_C_LABEL(curpcb)@l(5)

	/* save real pmap pointer for spill fill */
	lwz	5,PCB_PM(4)
	lis	6,_C_LABEL(curpm)@ha
	stwu	5,_C_LABEL(curpm)@l(6)
	stwcx.	5,0,6			/* clear possible reservation */

	lwz	6,PM_PDIRPA(5)
	stw	6,PCB_CR3(4)
	mtspr	M_TWB,6
	tlbia				/* XXX overkill */
	isync

	lwz	1,PCB_SP(4)		/* get new procs SP */

	ori	3,3,PSL_EE@l		/* interrupts are okay again */
	mtmsr	3

	lmw	10,8(1)			/* get other regs */
	lwz	1,0(1)			/* get saved SP */
	mr	2,12			/* get saved r2 */
	mtcr	11			/* get saved cr */
	isync
	li	3,1			/* switched lwps */

switch_return:
	mr	30,3
	lwz	3,PCB_SPL(4)
	bl	_C_LABEL(lcsplx)

	mr	3,30

	lwz	31,12(1)
	lwz	30,8(1)
	addi	1,1,16
	lwz	0,4(1)
	mtlr	0
	blr

/*
 * void
 * cpu_switchto(struct lwp *current, struct lwp *new)
 * Variant of cpu_switch, but instead of picking a new LWP from the run
 * queue, switch to the indicated new LWP.
 * 	r3 - current LWP
 * 	r4 - LWP to switch to
 */
ENTRY(cpu_switchto)
	mflr	0			/* save lr */
	stw	0,4(1)
	stwu	1,-16(1)
	stw	31,12(1)
	stw	30,8(1)

	mr	30,3			/* r30 = curlwp */
	mr	31,4			/* r31 = newlwp */

	stwu	1,-16(1)
	stw	29,8(1)

	li	0,0
	lis	3,_C_LABEL(curlwp)@ha
	stw	0,_C_LABEL(curlwp)@l(3)	/* Zero to not accumulate cpu time */
	lis	3,_C_LABEL(curpcb)@ha
	lwz	29,_C_LABEL(curpcb)@l(3)

	li	3,0
	bl	_C_LABEL(lcsplx)
	stw	3,PCB_SPL(29)		/* save spl */

	lwz	29,8(1)
	addi	1,1,16

	mfmsr	3
	andi.	3,3,~PSL_EE@l		/* disable interrupts while
					   manipulating runque */
	mtmsr	3
	isync

	b	switch_common

/*
 * Child comes here at the end of a fork.
 * Return to userspace via the trap return path.
 */
	.globl	_C_LABEL(fork_trampoline)
_C_LABEL(fork_trampoline):
#if defined(MULTIPROCESSOR)
	bl	_C_LABEL(proc_trampoline_mp)
#endif
	xor	3,3,3
	bl	_C_LABEL(lcsplx)
	mtlr	31
	mr	3,30
	blrl				/* jump indirect to r31 */
	b	trapexit

/*
 * Pull in common trap vector code.
 */
#include <mpc860/mpc860/trap_subr.S>

/*
 * int setfault()
 *
 * Similar to setjmp to setup for handling faults on accesses to user memory.
 * Any routine using this may only call bcopy, either the form below,
 * or the (currently used) C code optimized, so it doesn't use any non-volatile
 * registers.
 */
	.globl	_C_LABEL(setfault)
_C_LABEL(setfault):
	mflr	0
	mfcr	12
#if defined(MULTIPROCESSOR)
	GET_CPUINFO(4)
	lwz	4,CI_CURPCB(4)
#else
	lis	4,_C_LABEL(curpcb)@ha
	lwz	4,_C_LABEL(curpcb)@l(4)
#endif
	stw	3,PCB_FAULT(4)
	stw	0,0(3)
	stw	1,4(3)
	stw	2,8(3)
	stmw	12,12(3)
	xor	3,3,3
	blr

	.globl	_C_LABEL(dtlbmiss), _C_LABEL(dtlbmiss_end)
_C_LABEL(dtlbmiss):
#ifdef TLBMISS_DEBUG
	stmw	0,0x4010(0)
	mflr	28
	mfctr	29
	mfcr	30
	mfxer	31
	stmw	28,0x4000(0)

	lis	1,0x000d0000@ha
	addi	1,1,0x000d0000@l
	lis	3,dtlbmiss_hook@ha
	addi	3,3,dtlbmiss_hook@l
	mtlr	3
	blrl

	lmw	28,0x4000(0)
	mtlr	28
	mtctr	29
	mtcr	30
	mtxer	31
	lwz	0,0x4010(0)
	lmw	1,0x4014(0)
#endif

	mtspr	M_TW,1		# save r1
	mfspr	1,M_TWB		# L1 pte (ptp) address
	lwz	1,0(1)
	rlwinm	1,1,0,31,19	# l1pte &= 0xfffff001
	mtspr	MD_TWC,1
	mfspr	1,MD_TWC	# L2 pte address
	lwz	1,0(1)
	ori	1,1,0xf0	# PG_4K
	mtspr	MD_RPN,1	# write TLB entry
	mfspr	1,M_TW		# restore r1
	rfi
_C_LABEL(dtlbmiss_end):

	.globl	_C_LABEL(itlbmiss), _C_LABEL(itlbmiss_end)
_C_LABEL(itlbmiss):
	mtspr	M_TW,1		# save r1
	mfspr	1,MI_EPN	# miss EA
	mtspr	MD_EPN,1	# set miss EA
	mfspr	1,M_TWB		# L1 pte (ptp) address
	lwz	1,0(1)
	rlwinm	1,1,0,31,19	# l1pte &= 0xfffff001
	mtspr	MI_TWC,1
	mtspr	MD_TWC,1	# use MD_TWC to calculate L2 pte address
	mfspr	1,MD_TWC	# L2 pte address
	lwz	1,0(1)
	ori	1,1,0xf0	# PG_4K
	mtspr	MI_RPN,1	# write TLB entry
	mfspr	1,M_TW		# restore r1
	rfi
_C_LABEL(itlbmiss_end):

/*
 * Signal trampoline; copied to top of user stack.
 */
	.globl	_C_LABEL(sigcode),_C_LABEL(esigcode)
_C_LABEL(sigcode):
	addi	1,1,-16			/* reserved space for callee */
	blrl
	addi	3,1,16+8		/* compute &sf_sc */
	li	0,SYS___sigreturn14
	sc				/* sigreturn(scp) */
	li	0,SYS_exit
	sc				/* exit(errno) */
_C_LABEL(esigcode):

user_text = 0x01800000
