/*	$Id: machdep.c,v 1.22 2004/01/01 14:57:47 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_compat_netbsd.h"
#include "opt_ddb.h"
#include "opt_ipkdb.h"
#include "opt_multiprocessor.h"
#include "ksyms.h"

#include <sys/param.h>
#include <sys/boot_flag.h>
#include <sys/buf.h>
#include <sys/kernel.h>
#include <sys/ksyms.h>
#include <sys/mbuf.h>
#include <sys/mount.h>
#include <sys/msgbuf.h>
#include <sys/reboot.h>
#include <sys/sa.h>
#include <sys/sysctl.h>
#include <sys/systm.h>
#include <sys/user.h>

#include <net/netisr.h>

#include <machine/powerpc.h>
#include <machine/trap.h>

#include <dev/cons.h>

struct vm_map *exec_map;
struct vm_map *mb_map;
struct vm_map *phys_map;

/*
 * Global variables used here and there
 */
struct pcb *curpcb;
struct pmap *curpm;
struct lwp *fpulwp;

extern struct user *proc0paddr;

paddr_t msgbuf_paddr;
vaddr_t msgbuf_vaddr;

int msgbufmapped = 0;

int lcsplx(int);
void install_extint(void (*)(void));
static void enable_cache();

/*
 * This should probably be in autoconf!				XXX
 */
char cpu_model[80];
char machine[] = MACHINE;		/* from <machine/param.h> */
char machine_arch[] = MACHINE_ARCH;	/* from <machine/param.h> */

void
install_extint(handler)
	void (*handler) __P((void));
{
	extern extint, extsize;
	extern u_long extint_call;
	u_long offset = (u_long)handler - (u_long)&extint_call;
	int omsr, msr;

#ifdef DIAGNOSTIC
	if (offset > 0x1ffffff)
		panic("install_extint: too far away");
#endif
	asm volatile ("mfmsr %0; andi. %1,%0,%2; mtmsr %1"
		      : "=r"(omsr), "=r"(msr) : "K"((u_short)~PSL_EE));
	extint_call = (extint_call & 0xfc000003) | offset;
	bcopy(&extint, (void *)EXC_EXI, (size_t)&extsize);
	__syncicache((void *)&extint_call, sizeof extint_call);
	__syncicache((void *)EXC_EXI, (int)&extsize);
	asm volatile ("mtmsr %0" :: "r"(omsr));
}

/*
 * Machine dependent startup code.
 */
void
cpu_startup()
{
	int i;
	caddr_t v;
	vaddr_t minaddr, maxaddr;
	char pbuf[9];

	for (i = 0; i < btoc(MSGBUFSIZE); i++)
		pmap_kenter_pa(msgbuf_vaddr + i * PAGE_SIZE,
		    msgbuf_paddr + i * PAGE_SIZE, VM_PROT_READ|VM_PROT_WRITE);
	pmap_update(pmap_kernel());

	initmsgbuf((caddr_t)msgbuf_vaddr, round_page(MSGBUFSIZE));

	lwp0.l_addr = proc0paddr;
	v = (caddr_t)proc0paddr + USPACE;

	printf("%s", version);
	identifycpu(cpu_model);

	format_bytes(pbuf, sizeof(pbuf), ctob(physmem));
	printf("total memory = %s\n", pbuf);

	minaddr = 0;

	/*
	 * Allocate a submap for exec arguments.  This map effectively
	 * limits the number of processes exec'ing at any time.
	 */
	exec_map = uvm_km_suballoc(kernel_map, &minaddr, &maxaddr,
	    16*NCARGS, VM_MAP_PAGEABLE, FALSE, NULL);

	/*
	 * Allocate a submap for physio
	 */
	phys_map = uvm_km_suballoc(kernel_map, &minaddr, &maxaddr,
	    VM_PHYS_SIZE, 0, FALSE, NULL);

	/*
	 * Finally, allocate mbuf cluster submap.
	 */
	mb_map = uvm_km_suballoc(kernel_map, &minaddr, &maxaddr,
	    nmbclusters * mclbytes, VM_MAP_INTRSAFE, FALSE, NULL);

	/*
	 * No need to allocate an mbuf cluster submap.  Mbuf clusters
	 * are allocated via the pool allocator, and we use direct-mapped
	 * pool pages.
	 */

	format_bytes(pbuf, sizeof(pbuf), ptoa(uvmexp.free));
	printf("avail memory = %s\n", pbuf);
	format_bytes(pbuf, sizeof(pbuf), bufpages * NBPG);
	printf("using %d buffers containing %s of memory\n", nbuf, pbuf);
}

/*
 * consinit
 * Initialize system console.
 */
void
consinit()
{
	static int initted;

	if (initted)
		return;
	initted = 1;
	cninit();
}

/*
 * Machine dependent system variables.
 */
SYSCTL_SETUP(sysctl_machdep_setup, "sysctl machdep subtree setup")
{
	u_int pvr;

	sysctl_createv(SYSCTL_PERMANENT,
		       CTLTYPE_NODE, "machdep", NULL,
		       NULL, 0, NULL, 0,
		       CTL_MACHDEP, CTL_EOL);

	sysctl_createv(SYSCTL_PERMANENT|SYSCTL_IMMEDIATE,
		       CTLTYPE_INT, "cachelinesize", NULL,
		       NULL, CACHELINESIZE, NULL, 0,
		       CTL_MACHDEP, CPU_CACHELINE, CTL_EOL);
}

/*
 * Crash dump handling.
 */
void
dumpsys()
{
	printf("dumpsys: TBD\n");
}

/*
 * Soft networking interrupts.
 */
void
softnet()
{
	extern /*volatile*/ int netisr;
	int isr;

	isr = netisr;
	netisr = 0;

#define DONETISR(bit, fn) do {		\
	if (isr & (1 << bit))		\
		fn();			\
} while (0)

#include <net/netisr_dispatch.h>

#undef DONETISR

}

/*
 * Soft tty interrupts.
 */
void
softserial()
{
#if NZSC > 0
	zssoft();
#endif
#if NCOM > 0
	comsoft();
#endif
}

/*
 * Halt or reboot the machine after syncing/dumping according to howto.
 */
void
cpu_reboot(howto, what)
	int howto;
	char *what;
{
	static int syncing;
	static char str[256];
	char *ap = str, *ap1 = ap;

	boothowto = howto;
	if (!cold && !(howto & RB_NOSYNC) && !syncing) {
		syncing = 1;
		vfs_shutdown();		/* sync */
		resettodr();		/* set wall clock */
	}

	splhigh();

	if (!cold && (howto & RB_DUMP))
		dumpsys();

	doshutdownhooks();

	if (howto & RB_HALT) {
		printf("halted\n\n");

		/* flush cache for msgbuf */
		__syncicache((void *)msgbuf_vaddr, round_page(MSGBUFSIZE));

		ppc_exit();
	}

	printf("rebooting\n\n");
	if (what && *what) {
		if (strlen(what) > sizeof str - 5)
			printf("boot string too large, ignored\n");
		else {
			strcpy(str, what);
			ap1 = ap = str + strlen(str);
			*ap++ = ' ';
		}
	}
	*ap++ = '-';
	if (howto & RB_SINGLE)
		*ap++ = 's';
	if (howto & RB_KDB)
		*ap++ = 'd';
	*ap++ = 0;
	if (ap[-2] == '-')
		*ap1 = 0;

	/* flush cache for msgbuf */
	__syncicache((void *)msgbuf_vaddr, round_page(MSGBUFSIZE));

	ppc_boot(str);
}

void
ppc_exit()
{
	printf("\n");
	printf("The operating system has halted.\n");
	printf("Please type any character to reboot.\n\n");
	cnpollc(1);
	cngetc();
	cnpollc(0);

	ppc_boot(NULL);
}

void
ppc_boot(str)
	char *str;
{
	/* Cause watchdog reset. */
	for (;;);
}

int
lcsplx(ipl)
	int ipl;
{
	int s;

	if (ipl == 0)
		ipl = imask[IPL_NONE];
	s = spllower(ipl);
	return s;
}

void
cninit()
{
	struct consdev *cp;
	extern struct consdev consdev_scc;

	cp = &consdev_scc;
	(*cp->cn_probe)(cp);
	(*cp->cn_init)(cp);
	cn_tab = cp;
}




#include <machine/mpc860dev.h>

static inline void
eieio()
{
	asm volatile ("eieio; sync");
}

/* Service software watchdog timer. */
void
sswt()
{
	volatile struct mpc860dev *immr = get_immr();

	immr->siu_swsr = 0x556c;
	eieio();
	immr->siu_swsr = 0xaa39;
	eieio();
}

static inline void
enable_mmu()
{
	int msr;

	asm volatile ("sync; isync");
	asm volatile ("mfmsr %0" : "=r"(msr));
	msr |= PSL_DR|PSL_IR;
	asm volatile ("mtmsr %0" :: "r"(msr));
	asm volatile ("sync; isync");
}

#include <machine/spr.h>

extern char itlbmiss[], itlbmiss_end[];
extern char dtlbmiss[], dtlbmiss_end[];
extern char _edata[], _end[];

extern trapcode, trapsize;
extern alitrap, alisize;
extern dsitrap, dsisize;
extern isitrap, isisize;
extern decrint, decrsize;
extern tlbimiss, tlbimsize;
extern tlbdlmiss, tlbdlmsize;
extern tlbdsmiss, tlbdsmsize;
extern sctrap;
#ifdef DDB
extern ddblow, ddbsize;
#endif


void ext_intr(void);

extern paddr_t avail_start, avail_end;

void
initmpc860(ssym, esym)
	paddr_t ssym, esym;
{
	char *exc;

	/* Clear BSS. */
	bzero(_edata, _end - _edata);

	/* IMMR = 0xff000000; */

	consinit();

#if NKSYMS || defined(DDB) || defined(LKM)
	printf("ssym = %x, esym = %x\n", ssym, esym);
	ksyms_init(esym - ssym, (void *)ssym, (void *)esym);
#endif

	lwp0.l_addr = proc0paddr;
	bzero(lwp0.l_addr, sizeof *lwp0.l_addr);

	curpcb = &proc0paddr->u_pcb;

	curpm = curpcb->pcb_pm = pmap_kernel();

	/*
	 * Set up trap vectors
	 */
	for (exc = (char *)EXC_RSVD; exc <= (char *)EXC_LAST; exc += 0x100)
		bcopy(&trapcode, (void *)exc, (size_t)&trapsize);

	bcopy(&alitrap, (void *)EXC_ALI, (size_t)&alisize);
	bcopy(&dsitrap, (void *)EXC_DSI, (size_t)&dsisize);
	bcopy(&isitrap, (void *)EXC_ISI, (size_t)&isisize);
	bcopy(&decrint, (void *)EXC_DECR, (size_t)&decrsize);
	bcopy(itlbmiss, (char *)EXC_ITLBMISS, itlbmiss_end - itlbmiss);
	bcopy(dtlbmiss, (char *)EXC_DTLBMISS, dtlbmiss_end - dtlbmiss);
	bcopy(&isitrap, (void *)EXC_ITLBERR, (size_t)&isisize);
	bcopy(&dsitrap, (void *)EXC_DTLBERR, (size_t)&dsisize);
	bcopy(&sctrap, (void *)EXC_SC, 0x100);
#if defined(DDB)
	bcopy(&ddblow, (void *)EXC_PGM, (size_t)&ddbsize);
	bcopy(&ddblow, (void *)EXC_TRC, (size_t)&ddbsize);
	bcopy(&ddblow, (void *)EXC_BPT, (size_t)&ddbsize);
#endif

	/*
	 * external interrupt handler install
	 */
	install_extint(ext_intr);

	__syncicache((void *)EXC_RST, EXC_LAST - EXC_RST + 0x100);

#ifdef FORCE_RB_SINGLE
	boothowto = RB_SINGLE;
#endif

#ifdef DDB
	if (boothowto & RB_KDB)
		Debugger();
#endif

	pmap_bootstrap0(esym, 16*1024*1024);		/* XXX 16MB */
	/* asm volatile ("tlbie %0" :: "r"(0)); */

	uvm_setpagesize();

	uvm_page_physload(atop(avail_start), atop(avail_end),
	    atop(avail_start), atop(avail_end), VM_FREELIST_DEFAULT);

	pmap_bootstrap(0xc0000000);

	enable_cache();
}

#ifdef TLBMISS_DEBUG
void
dtlbmiss_hook()
{
	u_int srr0, srr1, epn, ctr;
	static int prev_epn, prev_ctr;

	asm ("mfsrr0 %0" : "=r"(srr0));
	asm ("mfsrr1 %0" : "=r"(srr1));
	printf("tlbmiss srr0 = 0x%x, ", srr0);

	asm ("mfspr %0,MD_EPN" : "=r"(epn));
	printf("MD_EPN = 0x%08x, ", epn);

	asm ("mfspr %0,MD_CTR" : "=r"(ctr));
	printf("MD_CTR = 0x%08x\n", ctr);

	if (1) {
		u_int *l1pt, *l2pt, pte;
		vaddr_t va = epn & PG_FRAME;

		asm volatile ("mfspr %0,M_TWB" : "=r"(l1pt));
		l1pt = (u_int *)((u_int)l1pt & PG_FRAME);

		l2pt = (void *)((u_int)l1pt[va >> 22] & PG_FRAME);
		pte =  l2pt[(va & 0x003ff000) >> 12];

		printf("va = 0x%x, pte = 0x%x\n", va, pte);

		printf("&l2pt = %p\n", &l1pt[va >> 22]);
		printf("&pte = %p\n", &l2pt[(va & 0x003ff000) >> 12]);
	}

	if (prev_epn == epn && prev_ctr - 0x100 == ctr)
		Debugger();
	prev_epn = epn;
	prev_ctr = ctr;
}
#endif /* TLBMISS_DEBUG */

int
pmap_bootstrap0(esym, emem)
	paddr_t esym, emem;
{
	pt_entry_t *l1pt, *pte;
	paddr_t avail;
	int i;

	avail = roundup(esym, NBPG);

	l1pt = (pt_entry_t *)avail;
	bzero(l1pt, NBPG);
	avail += NBPG;

	asm volatile ("mtspr M_TWB,%0" :: "r"(l1pt));

	curpcb->pcb_cr3 = (paddr_t)l1pt;

	pte = &l1pt[pdei(KERNBASE)];
	for (i = 0; i < 8; i++) {
		bzero((void *)avail, NBPG);
		*pte++ = (pt_entry_t)avail | PG_V | PG_KW | PG_M | PG_SH;
		avail += NBPG;
	}
	l1pt[PDSLOT_PTE] = (pt_entry_t)l1pt | PG_V | PG_KW | PG_M | PG_SH;

	avail_start = avail;
	avail_end = emem;

	physmem = btoc(avail_end);
	avail_end -= round_page(MSGBUFSIZE);
	msgbuf_paddr = avail_end;

	return 0;
}

void
inittlb()
{
	u_int ctr, epn, twc, rpn;
	paddr_t pa;
	vaddr_t va;

	/* Invalidate all tlb. */
	ctr = CTR_TWAM;
	asm volatile ("mtspr MD_CTR,%0" :: "r"(ctr));
	asm volatile ("mtspr MI_CTR,%0" :: "r"(ctr));
	asm volatile ("tlbia");

	/* Map 0x00000000-0x007fffff to 0x80000000. */
	pa = 0x00000000;
	va = 0x80000000;

	ctr = CTR_TWAM | 31 << 8;
	asm volatile ("mtspr MD_CTR,%0" :: "r"(ctr));
	asm volatile ("mtspr MI_CTR,%0" :: "r"(ctr));

	epn = va | EPN_EV;
	asm volatile ("mtspr MD_EPN,%0" :: "r"(epn));
	asm volatile ("mtspr MI_EPN,%0" :: "r"(epn));

	twc = TWC_PS_8M | TWC_V;
	asm volatile ("mtspr MD_TWC,%0" :: "r"(twc));
	asm volatile ("mtspr MI_TWC,%0" :: "r"(twc));

	rpn = pa | RPN_8M | RPN_M | RPN_SH | RPN_V;
	asm volatile ("mtspr MD_RPN,%0" :: "r"(rpn));
	asm volatile ("mtspr MI_RPN,%0" :: "r"(rpn));

	/* Map 0xff000000-0xff07ffff to 0xff000000. */
	pa = 0xff000000;
	va = 0xff000000;

	ctr = CTR_TWAM | 30 << 8;
	asm volatile ("mtspr MD_CTR,%0" :: "r"(ctr));
	asm volatile ("mtspr MI_CTR,%0" :: "r"(ctr));

	epn = va | EPN_EV;
	asm volatile ("mtspr MD_EPN,%0" :: "r"(epn));
	asm volatile ("mtspr MI_EPN,%0" :: "r"(epn));

	twc = TWC_PS_512K | TWC_V;
	asm volatile ("mtspr MD_TWC,%0" :: "r"(twc));
	asm volatile ("mtspr MI_TWC,%0" :: "r"(twc));

	rpn = pa | RPN_512K | RPN_M | RPN_SH | RPN_CI | RPN_V;
	asm volatile ("mtspr MD_RPN,%0" :: "r"(rpn));
	asm volatile ("mtspr MI_RPN,%0" :: "r"(rpn));

	/* Map 0x00000000-0x007fffff to 0x00000000. */
	pa = 0x00000000;
	va = 0x00000000;

	ctr = CTR_TWAM | 29 << 8;
	asm volatile ("mtspr MD_CTR,%0" :: "r"(ctr));
	asm volatile ("mtspr MI_CTR,%0" :: "r"(ctr));

	epn = va | EPN_EV;
	asm volatile ("mtspr MD_EPN,%0" :: "r"(epn));
	asm volatile ("mtspr MI_EPN,%0" :: "r"(epn));

	twc = TWC_PS_8M | TWC_V;
	asm volatile ("mtspr MD_TWC,%0" :: "r"(twc));
	asm volatile ("mtspr MI_TWC,%0" :: "r"(twc));

	rpn = pa | RPN_8M | RPN_M | RPN_SH | RPN_V;
	asm volatile ("mtspr MD_RPN,%0" :: "r"(rpn));
	asm volatile ("mtspr MI_RPN,%0" :: "r"(rpn));

	ctr = CTR_TWAM | CTR_RSV4;
	asm volatile ("mtspr MD_CTR,%0" :: "r"(ctr));
	asm volatile ("mtspr MI_CTR,%0" :: "r"(ctr));

	asm volatile ("tlbia; sync");
	enable_mmu();
}

#define IC_CST_DEN		0x80000000	/* I- cache enable (R) */
#define IC_CST_CMD		0x0e000000
#define IC_CST_CMD_CE		0x02000000	/* Cache enable */
#define IC_CST_CMD_CD		0x04000000	/* Cache disable */
#define IC_CST_CMD_LLCB		0x06000000	/* Load and lock cache block */
#define IC_CST_CMD_UNLOCK_CB	0x08000000	/* Unlock cache block */
#define IC_CST_CMD_UNLOCK_ALL	0x0a000000	/* Unlock all */
#define IC_CST_CMD_INV_ALL	0x0c000000	/* Invalidate all */
#define IC_CST_CCER1		0x00200000	/* Data cache error type 1 (R)*/
#define IC_CST_CCER2		0x00100000	/* Data cache error type 2 (R)*/
#define IC_CST_CCER3		0x00080000	/* Reserved */

#define DC_CST_DEN		0x80000000	/* Data cache enable */
#define DC_CST_DFWT		0x40000000	/* Force write thru */
#define DC_CST_LES		0x20000000	/* Little endian swap */
#define DC_CST_CMD		0x0f000000
#define DC_CST_CMD_FWT		0x01000000	/* Force write thru enable */
#define DC_CST_CMD_CE		0x02000000	/* Cache enable */
#define DC_CST_CMD_FWTD		0x03000000	/* Force write thru disable */
#define DC_CST_CMD_CD		0x04000000	/* Cache disable */
#define DC_CST_CMD_LES		0x05000000	/* Little endian enable */
#define DC_CST_CMD_LLCB		0x06000000	/* Load and lock cache block */
#define DC_CST_CMD_LESD		0x07000000	/* */
#define DC_CST_CMD_UCB		0x08000000	/* Unlock cache block */
#define DC_CST_CMD_UNLOCK_ALL	0x0a000000	/* Unlock all */
#define DC_CST_CMD_INV_ALL	0x0c000000	/* Invalidate all */
#define DC_CST_CMD_FCB		0x0e000000	/* Flush cache block */
#define DC_CST_CCER1		0x00200000	/* Data cache error type 1 */
#define DC_CST_CCER2		0x00100000	/* Data cache error type 2 */
#define DC_CST_CCER3		0x00080000	/* Reserved */

void
enable_cache()
{
	asm volatile ("mtspr IC_CST,%0" :: "r"(IC_CST_CMD_UNLOCK_ALL));
	asm volatile ("mtspr IC_CST,%0" :: "r"(IC_CST_CMD_INV_ALL));
	asm volatile ("mtspr IC_CST,%0" :: "r"(IC_CST_CMD_CE));

	asm volatile ("mtspr DC_CST,%0" :: "r"(DC_CST_CMD_UNLOCK_ALL));
	asm volatile ("mtspr DC_CST,%0" :: "r"(DC_CST_CMD_INV_ALL));
#if 1
	asm volatile ("mtspr DC_CST,%0" :: "r"(DC_CST_CMD_FWT));
#endif
	asm volatile ("mtspr DC_CST,%0" :: "r"(DC_CST_CMD_CE));
}

void
flushcache(from, len)
	void *from;
	int len;
{
	int l, off;
	char *p;

	off = (u_int)from & (CACHELINESIZE - 1);
	len += off;

	p = from;
	l = len;
	do {
		asm volatile ("dcbf 0,%0" :: "r"(p));
		p += CACHELINESIZE;
	} while ((l -= CACHELINESIZE) > 0);
	asm volatile ("sync");
}

void
invalidatecache(from, len)
	void *from;
	int len;
{
	int l, off;
	char *p;

	off = (u_int)from & (CACHELINESIZE - 1);
	len += off;

	p = from;
	l = len;
	do {
		asm volatile ("dcbi 0,%0" :: "r"(p));
		p += CACHELINESIZE;
	} while ((l -= CACHELINESIZE) > 0);
	asm volatile ("sync");
}

/*
 * Allocate vm space and mapin the I/O address
 */
void *
mapiodev(pa, len)
	paddr_t pa;
	psize_t len;
{
	paddr_t faddr;
	vaddr_t taddr, va;
	int off;

	faddr = trunc_page(pa);
	off = pa - faddr;
	len = round_page(off + len);
	va = taddr = uvm_km_valloc(kernel_map, len);

	if (va == 0)
		return NULL;

	for (; len > 0; len -= NBPG) {
		pmap_kenter_pa(taddr, faddr|PG_CI, VM_PROT_ALL|PMAP_CI);
		faddr += NBPG;
		taddr += NBPG;
	}
	pmap_update(pmap_kernel());
	return (void *)(va + off);
}
