/*	$Id: trap.c,v 1.19 2003/11/02 08:42:26 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_ddb.h"

#include <sys/param.h>
#include <sys/pool.h>
#include <sys/proc.h>
#include <sys/sa.h>
#include <sys/savar.h>
#include <sys/systm.h>
#include <sys/user.h>

#include <uvm/uvm_extern.h>

#include <machine/cpu.h>
#include <machine/frame.h>
#include <machine/pcb.h>
#include <machine/pmap.h>
#include <machine/psl.h>
#include <machine/trap.h>

volatile int astpending;
volatile int want_resched;

#ifdef TRAP_DEBUG
# define DPRINTF printf
#else
# define DPRINTF while (0) printf
#endif

void
trap(frame)
	struct trapframe *frame;
{
	struct lwp *l = curlwp;
	struct proc *p = l ? l->l_proc : NULL;
	struct pcb *pcb = curpcb;
	vaddr_t va = frame->dar;
	int type = frame->exc;
	int ftype, rv;
	ksiginfo_t ksi;

	uvmexp.traps++;

	if (frame->srr1 & PSL_PR)
		type |= EXC_USER;

	DPRINTF("trap: srr0 = 0x%x, srr1 = 0x%x", frame->srr0, frame->srr1);
	DPRINTF(", type = 0x%x\n", type);

	switch (type) {
	case EXC_TRC|EXC_USER:
		KERNEL_PROC_LOCK(l);
		frame->srr1 &= ~PSL_SE;
		KSI_INIT_TRAP(&ksi);
		ksi.ksi_signo = SIGTRAP;
		ksi.ksi_addr = (void *)frame->srr0;
		ksi.ksi_trap = EXC_TRC;
		ksi.ksi_code = TRAP_TRACE;
		trapsignal(l, &ksi);
		KERNEL_PROC_UNLOCK(l);
		break;
	case EXC_DSI:
		printf("EXC_DSI\n");
	case EXC_DTLBERR:
	{
		struct vm_map *map;
		faultbuf *fb;

		/* XXX CHECK intr_depth. */

		KERNEL_LOCK(LK_CANRECURSE|LK_EXCLUSIVE);
		map = kernel_map;
		/*
		 * It is only a kernel address space fault iff:
		 *	1. !user and
		 *	2. pcb_onfault not set or
		 *	3. pcb_onfault set but kernel space data fault
		 * The last can occur during an exec() copyin where the
		 * argument space is lazy-allocated.
		 */
/* XXX */	if ((type & EXC_USER) == 0 && (va >= 0x80000000 ||
		    pcb->pcb_onfault == 0))
			map = kernel_map;
		else
			map = &p->p_vmspace->vm_map;

		if (frame->dsisr & DSISR_STORE)
			ftype = VM_PROT_WRITE;
		else
			ftype = VM_PROT_READ;

		/* XXX */
		if (frame->dsisr & 0x08000000)
			printf("DSISR(1) 0x%08x\n", frame->dsisr);
		/* PG_M emulation */
		if ((frame->dsisr & DSISR_ALL) ==
		    (DSISR_PROT|DSISR_STORE) &&
		    pmap_emul_modify(map->pmap, va))
			return;

		rv = uvm_fault(map, trunc_page(va), 0, ftype);
		KERNEL_UNLOCK();
		if (rv == 0)
			return;
		if (rv == EACCES)
			rv = EFAULT;
		if (fb = pcb->pcb_onfault) {
			frame->srr0 = (*fb)[0];
			frame->fixreg[1] = (*fb)[1];
			frame->fixreg[2] = (*fb)[2];
			frame->fixreg[3] = rv;
			frame->cr = (*fb)[3];
			bcopy(&(*fb)[4], &frame->fixreg[13],
			    19 * sizeof(register_t));
			return;
		}
		goto brain_damage;
	}
	case EXC_DSI|EXC_USER:
		printf("EXC_DSI|EXC_USER\n");
	case EXC_DTLBERR|EXC_USER:
		/* XXX */
		if (frame->dsisr & 0x08000000)
			printf("DSISR(2) 0x%08x va 0x%08x\n", frame->dsisr, va);
		/* PG_M emulation */
		if ((frame->dsisr & DSISR_ALL) == (DSISR_PROT|DSISR_STORE) &&
		    pmap_emul_modify(p->p_vmspace->vm_map.pmap, va))
			return;

		KERNEL_PROC_LOCK(l);
		if (frame->dsisr & DSISR_STORE)
			ftype = VM_PROT_WRITE;
		else
			ftype = VM_PROT_READ;
		if (l->l_flag & L_SA) {
			KDASSERT(p != NULL && p->p_sa != NULL);
			p->p_sa->sa_vp_faultaddr = va;
			l->l_flag |= L_SA_PAGEFAULT;
		}
		rv = uvm_fault(&p->p_vmspace->vm_map, trunc_page(va), 0, ftype);
		if (rv == 0) {
			uvm_grow(p, va);
			l->l_flag &= ~L_SA_PAGEFAULT;
			KERNEL_PROC_UNLOCK(l);
			break;
		}
		KSI_INIT_TRAP(&ksi);
		ksi.ksi_addr = (void *)va;
		ksi.ksi_trap = EXC_DTLBERR;
		if (rv == EACCES)
			ksi.ksi_code = SEGV_ACCERR;
		else
			ksi.ksi_code = SEGV_MAPERR;
		if (rv == ENOMEM) {
			printf("UVM: pid %d (%s) lid %d, uid %d killed: "
			       "out of swap\n",
			       p->p_pid, p->p_comm, l->l_lid,
			       p->p_cred && p->p_ucred ?
			       p->p_ucred->cr_uid : -1);
			ksi.ksi_signo = SIGKILL;
			trapsignal(l, &ksi);
		} else {
			ksi.ksi_signo = SIGSEGV;
			trapsignal(l, &ksi);
		}
		l->l_flag &= ~L_SA_PAGEFAULT;
		KERNEL_PROC_UNLOCK(l);
		break;
	case EXC_ISI|EXC_USER:
	case EXC_ITLBERR|EXC_USER:
		KERNEL_PROC_LOCK(l);
		ftype = VM_PROT_READ | VM_PROT_EXECUTE;
		if (l->l_flag & L_SA) {
			KDASSERT(p != NULL && p->p_sa != NULL);
			p->p_sa->sa_vp_faultaddr = frame->srr0;
			l->l_flag |= L_SA_PAGEFAULT;
		}
		rv = uvm_fault(&p->p_vmspace->vm_map, trunc_page(frame->srr0),
		    0, ftype);
		if (rv == 0) {
			l->l_flag &= ~L_SA_PAGEFAULT;
			KERNEL_PROC_UNLOCK(l);
			break;
		}
		KSI_INIT_TRAP(&ksi);
		ksi.ksi_signo = SIGSEGV;
		ksi.ksi_addr = (void *)frame->srr0;
		ksi.ksi_trap = EXC_ITLBERR;
		ksi.ksi_code = SEGV_MAPERR;
		trapsignal(l, &ksi);
		l->l_flag &= ~L_SA_PAGEFAULT;
		KERNEL_PROC_UNLOCK(l);
		break;

	case EXC_AST|EXC_USER:
		astpending = 0;		/* we are about to do it */
		KERNEL_PROC_LOCK(l);
		uvmexp.softs++;
		if (p->p_flag & P_OWEUPC) {
			p->p_flag &= ~P_OWEUPC;
			ADDUPROF(p);
		}
		/* Check whether we are being preempted. */
		if (want_resched)
			preempt(0);
		KERNEL_PROC_UNLOCK(l);
		break;

	case EXC_ALI|EXC_USER:
		KERNEL_PROC_LOCK(l);
		KSI_INIT_TRAP(&ksi);
		ksi.ksi_signo = SIGBUS;
		ksi.ksi_addr = (void *)frame->srr0;
		ksi.ksi_trap = EXC_ALI;
		ksi.ksi_code = BUS_ADRALN;
		trapsignal(l, &ksi);
		KERNEL_PROC_UNLOCK(l);
		break;

	case EXC_PGM|EXC_USER:
/* XXX temporarily */
		KERNEL_PROC_LOCK(l);
		KSI_INIT_TRAP(&ksi);
		ksi.ksi_addr = (void *)frame->srr0;
		ksi.ksi_trap = EXC_PGM;
		if (frame->srr1 & 0x00020000) {
			ksi.ksi_signo = SIGTRAP;
			ksi.ksi_code = TRAP_BRKPT;
			trapsignal(l, &ksi);
		} else {
			ksi.ksi_signo = SIGILL;
			ksi.ksi_code = ILL_ILLOPC;
			trapsignal(l, &ksi);
		}
		KERNEL_PROC_UNLOCK(l);
		break;

	case EXC_MCHK:
		if (pcb->pcb_onfault) {
			faultbuf *fb = pcb->pcb_onfault;

			frame->srr0 = (*fb)[0];
			frame->fixreg[1] = (*fb)[1];
			frame->fixreg[2] = (*fb)[2];
			frame->fixreg[3] = EFAULT;
			frame->cr = (*fb)[3];
			bcopy(&(*fb)[4], &frame->fixreg[13],
			      19 * sizeof(register_t));
			return;
		}
		goto brain_damage;

	case EXC_SOFTEMU|EXC_USER:
		if (!(pcb->pcb_flags & PCB_FPU)) {
			bzero(&pcb->pcb_fpu, sizeof pcb->pcb_fpu);
			pcb->pcb_flags |= PCB_FPU;
		}

		rv = fpu_emulate(frame, (struct fpreg *)&pcb->pcb_fpu);
		if (rv) {
			KERNEL_PROC_LOCK(l);
			KSI_INIT_TRAP(&ksi);
			ksi.ksi_signo = rv;
			ksi.ksi_addr = (void *)frame->srr0;
			ksi.ksi_trap = EXC_SOFTEMU;
			switch (rv) {
			case SIGSEGV:
				ksi.ksi_code = SEGV_MAPERR;
				break;
			case SIGTRAP:
				ksi.ksi_code = TRAP_TRACE;
				break;
			case SIGILL:
				ksi.ksi_code = ILL_ILLOPC;
				break;
			case SIGFPE:
			default:
				ksi.ksi_code = FPE_FLTINV;	/* XXX */
				break;
			}
			trapsignal(l, &ksi);
			KERNEL_PROC_UNLOCK(l);
		}
		break;

	default:
brain_damage:
		printf("trap type %x at %x\n", type, frame->srr0);
#ifdef DDB
		if (kdb_trap(type, frame))
			return;
#endif
#ifdef TRAP_PANICWAIT
		printf("Press a key to panic.\n");
		cngetc();
#endif
		panic("trap");
	}

	userret(l, frame);
}

int
copyin(udaddr, kaddr, len)
	const void *udaddr;
	void *kaddr;
	size_t len;
{
	const char *up = udaddr;
	char *kp = kaddr;
	char *p;
	int rv;
	size_t l;
	faultbuf env;

	if ((rv = setfault(env)) != 0) {
		curpcb->pcb_onfault = 0;
		return rv;
	}
	bcopy(udaddr, kaddr, len);
	asm volatile ("sync");
	curpcb->pcb_onfault = 0;
	return 0;
}

int
copyout(kaddr, udaddr, len)
	const void *kaddr;
	void *udaddr;
	size_t len;
{
	const char *kp = kaddr;
	char *up = udaddr;
	char *p;
	int rv;
	size_t l;
	faultbuf env;

	if ((rv = setfault(env)) != 0) {
		curpcb->pcb_onfault = 0;
		return rv;
	}
	bcopy(kaddr, udaddr, len);
	asm volatile ("sync");
	curpcb->pcb_onfault = 0;
	return 0;
}

/*
 * kcopy(const void *src, void *dst, size_t len);
 *
 * Copy len bytes from src to dst, aborting if we encounter a fatal
 * page fault.
 *
 * kcopy() _must_ save and restore the old fault handler since it is
 * called by uiomove(), which may be in the path of servicing a non-fatal
 * page fault.
 */
int
kcopy(src, dst, len)
	const void *src;
	void *dst;
	size_t len;
{
	faultbuf env, *oldfault;
	int rv;

	oldfault = curpcb->pcb_onfault;
	if ((rv = setfault(env)) != 0) {
		curpcb->pcb_onfault = oldfault;
		return rv;
	}

	bcopy(src, dst, len);

	asm volatile ("sync");
	curpcb->pcb_onfault = oldfault;
	return 0;
}

int
badaddr(addr, size)
	void *addr;
	size_t size;
{
	return badaddr_read(addr, size, NULL);
}

int
badaddr_read(addr, size, rptr)
	void *addr;
	size_t size;
	int *rptr;
{
	faultbuf env;
	int x;

	/* Get rid of any stale machine checks that have been waiting.  */
	asm volatile ("sync; isync");

	if (setfault(env)) {
		curpcb->pcb_onfault = 0;
		asm volatile ("sync");
		return 1;
	}

	asm volatile ("sync");

	switch (size) {
	case 1:
		x = *(volatile int8_t *)addr;
		break;
	case 2:
		x = *(volatile int16_t *)addr;
		break;
	case 4:
		x = *(volatile int32_t *)addr;
		break;
	default:
		panic("badaddr: invalid size (%d)", size);
	}

	/* Make sure we took the machine check, if we caused one. */
	asm volatile ("sync; isync");

	curpcb->pcb_onfault = 0;
	asm volatile ("sync");	/* To be sure. */

	/* Use the value to avoid reorder. */
	if (rptr)
		*rptr = x;

	return 0;
}

/*
 * Start a new LWP.
 */
void
startlwp(arg)
	void *arg;
{
	int err;
	ucontext_t *uc = arg;
	struct lwp *l = curlwp;

	err = cpu_setmcontext(l, &uc->uc_mcontext, uc->uc_flags);
#if DIAGNOSTIC
	if (err)
		printf("Error %d from cpu_setmcontext.", err);
#endif
	pool_put(&lwp_uc_pool, uc);

	upcallret(l);
}

/*
 * XXX This is a terrible name.
 */
void
upcallret(l)
	struct lwp *l;
{
	struct trapframe *frame = trapframe(l);

	userret(l, frame);
}

#include <sys/userret.h>

void
userret(l, frame)
	struct lwp *l;
	struct trapframe *frame;
{
	mi_userret(l);

	curcpu()->ci_schedstate.spc_curpriority = l->l_priority = l->l_usrpri;
}
