/*	$Id: boot.c,v 1.4 2003/01/11 16:06:08 tsubai Exp $	*/

/*-
 * Copyright (c) 2001,2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>

#include <machine/cpu.h>
#include <machine/cpufunc.h>
#include <machine/mpc860dev.h>
#include <machine/sysconf.h>

#include <lib/libsa/stand.h>
#include <lib/libsa/loadfile.h>
#include <lib/libkern/libkern.h>

#define syncicache(addr, len)	__syncicache(addr, len)

int boothowto;

void scc_uart_init(void);
void fecinit(void);
void wdcinit(void);
void delay(u_int);

extern char _start[], _end[], _edata[];

void
startup(start)
	char *start;
{
	bcopy(start, _start , _edata - _start);
	bzero(_edata, _end - _edata);

	syncicache(_start, _end - _start);

	asm volatile ("lis 3,main@ha; addi 3,3,main@l; mtlr 3; blr");

	for (;;);
}

__dead void
_rtt()
{
	for (;;);
}

extern char bootprog_name[], bootprog_rev[], bootprog_maker[], bootprog_date[];

void
main()
{
	volatile struct mpc860dev *immr = get_immr();
	u_long marks[MARK_MAX];
	void (*entry)();
	char *ssym, *esym, *kernels[2];
	int msr, dipsw;

	/* Disable external interrupt. */
	asm ("mfmsr %0" : "=r"(msr));
	msr &= ~PSL_EE;
	asm volatile ("mtmsr %0" :: "r"(msr));

	/* Start timebase. */
	immr->siu_tbscr = 0x0001;

	/* Initialize SCC (console). */
	scc_uart_init();

	strcpy(bootprog_name, "OpenBlockS Boot");	/* XXX */

	printf("\n");
	printf(">> %s, Revision %s\n", bootprog_name, bootprog_rev);
	printf(">> (%s, %s)\n", bootprog_maker, bootprog_date);

	dipsw = (immr->cpm_padat & 0xc000) >> 14;
	if (dipsw & 1) {
		kernels[0] = "disk:/netbsd";
		kernels[1] = "net:/netbsd";
	} else {
		kernels[0] = "net:/netbsd";
		kernels[1] = "disk:/netbsd";
	}

	wdcinit();
	fecinit();

	marks[MARK_START] = -0x80000000;
	if (loadfile(kernels[0], marks, LOAD_KERNEL) < 0)
		if (loadfile(kernels[1], marks, LOAD_KERNEL) < 0)
			for (;;);

	entry = (void *)marks[MARK_ENTRY];
	ssym = (void *)marks[MARK_SYM];
	esym = (void *)marks[MARK_END];
	
	printf(" start=%p\n", entry);
	syncicache(entry, ssym - (char *)entry);

	(*entry)(ssym, esym);

	for (;;);
}

void
wdcinit()
{
	volatile struct mpc860dev *immr = get_immr();

	/* Configure IDE reset pin. */
	immr->cpm_pcpar &= ~0x0008;
	immr->cpm_pcdir |= 0x0008;
	immr->cpm_pcso &= ~0x0008;

	/* Reset IDE. */
	immr->cpm_pcdat &= ~0x0008;	/* assert */
	delay(100);
	immr->cpm_pcdat |= 0x0008;	/* negate */
	delay(10000);

#if 0
	/* Wait for reset done. */
	while (wdread(cmdreg, STATUS) & BSY);
#endif
}

void
delay(n)
	u_int n;
{
	const int ns_per_tick = 1000000000 / (TIMEBASE_FREQ / 4);
	u_quad_t tb;
	u_long tbh, tbl, scratch;

	tb = mftb();
	tb += (n * 1000 + ns_per_tick - 1) / ns_per_tick;
	tbh = tb >> 32;
	tbl = tb;
	asm volatile ("1: mftbu %0; cmplw %0,%1; blt 1b; bgt 2f;"
		      "mftb %0; cmplw %0,%2; blt 1b; 2:"
		      : "=r"(scratch) : "r"(tbh), "r"(tbl));
}

/* Service software watchdog timer. */
void
sswt()
{
	volatile struct mpc860dev *immr = get_immr();

	immr->siu_swsr = 0x556c;
	asm volatile ("eieio; sync");
	immr->siu_swsr = 0xaa39;
	asm volatile ("eieio; sync");
}

void
__syncicache(from, len)
	void *from;
	int len;
{
	int l, off;
	char *p;

	off = (u_int)from & (CACHELINESIZE - 1);
	len += off;

	p = from;
	l = len;
	do {
		asm volatile ("dcbst 0,%0" :: "r"(p));
		p += CACHELINESIZE;
	} while ((l -= CACHELINESIZE) > 0);
	asm volatile ("sync");

	p = from;
	l = len;
	do {
		asm volatile ("icbi 0,%0" :: "r"(p));
		p += CACHELINESIZE;
	} while ((l -= CACHELINESIZE) > 0);
	asm volatile ("sync");	/* required on 745x */
	asm volatile ("isync");
}
