/*	$Id: devopen.c,v 1.4 2003/03/23 04:15:41 tsubai Exp $	*/

/*-
 * Copyright (c) 2001,2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <netinet/in.h>

#include <lib/libkern/libkern.h>

#include <lib/libsa/stand.h>
#include <lib/libsa/nfs.h>
#include <lib/libsa/ufs.h>

#include <machine/pio.h>

#ifdef DEBUG
# define DPRINTF printf
#else
# define DPRINTF while (0) printf
#endif

void delay(u_int);
void sswt(void);
int net_open(void *);

int dkopen(struct open_file *, ...);
int dkclose(struct open_file *);
int dkstrategy(void *, int, daddr_t, size_t, void *, size_t *);
static int readsector(u_int, void *);

struct devsw devsw[] = {
	{ "dk", dkstrategy, dkopen, dkclose, noioctl }
};
int ndevs = sizeof(devsw) / sizeof(devsw[0]);

static struct fs_ops file_system_ufs = {
	ufs_open, ufs_close, ufs_read, ufs_write, ufs_seek, ufs_stat
};
static struct fs_ops file_system_nfs = {
	nfs_open, nfs_close, nfs_read, nfs_write, nfs_seek, nfs_stat
};

struct fs_ops file_system[3];
int nfsys;

static struct {
	int dummy;
} devdata;

char opened_name[256];

int
devopen(of, name, file)
	struct open_file *of;
	const char *name;
	char **file;
{
	char *cp;
	int error = 0;

	DPRINTF("devopen: name=%s\n", name);

	if (of->f_flags != F_READ)
		return EPERM;

	cp = strchr(name, ':');
	if (cp == NULL)
		return ENOENT;

	strcpy(opened_name, cp + 1);
	*file = opened_name;

	DPRINTF("file=%s\n", opened_name);

	of->f_dev = devsw;
	of->f_devdata = &devdata;

	if (strncmp(name, "net:", 4) == 0) {
		file_system[0] = file_system_nfs;
		nfsys = 1;
		error = net_open(&devdata);
		return error;
	}
	if (strncmp(name, "disk:", 5) == 0) {
		file_system[0] = file_system_ufs;
		nfsys = 1;
		return 0;
	}

	return ENOENT;
}

/*
 * IDE Υ쥸 0x80000010 ȡ
 * 0x8020000c ¸ߤ롣
 * եåȤ2ܤ16ӥåȥɬפ롣
 */
#define cmdreg 0x80000010
#define ctlreg 0x8020000c

#define BSY 0x80
#define DRQ 0x08

#define DEV 0x10
#define LBA 0x40

static u_int
wdread(h, off)
	paddr_t h;
	u_int off;
{
	return in16(h + off * 2);
}

static void
wdwrite(h, off, val)
	paddr_t h;
	u_int off, val;
{
	out16(h + off * 2, val);
}

#define CONTROL	0	/* device control(w) / alt status(r) */

#define DATA	0	/* data */
#define ERROR	1	/* features(w) / error(r) */
#define SECCNT	2	/* sector count */
#define SECNUM	3	/* sector number */
#define CYL_L	4	/* cylinder low */
#define CYL_H	5	/* cylinder high */
#define DEVHEAD	6	/* device / head */
#define COMMAND	7	/* command(w) / status(r) */
#define STATUS	7

static int
readsector(lba, buf)
	u_int lba;
	void *buf;
{
	u_int16_t *p = buf;
	int i, status;

	/*
	 * Device selection
	 */
	do {
		status = wdread(cmdreg, STATUS);
	} while (status & (BSY|DRQ));

	wdwrite(cmdreg, DEVHEAD, LBA);
	delay(1);

	do {
		status = wdread(cmdreg, STATUS);
	} while (status & (BSY|DRQ));

	/*
	 * Read one sector.
	 */
	wdwrite(ctlreg, 0, 0x02);
	wdwrite(cmdreg, SECCNT, 1);
	wdwrite(cmdreg, SECNUM, lba & 0xff);
	wdwrite(cmdreg, CYL_L, lba >> 8 & 0xff);
	wdwrite(cmdreg, CYL_H, lba >> 16 & 0xff);
	wdwrite(cmdreg, DEVHEAD, 0x40 | 0x00 | (lba >> 24 & 0x0f));
	wdwrite(cmdreg, COMMAND, 0x20);
	delay(1);

	wdread(ctlreg, 0);

	do {
		status = wdread(cmdreg, STATUS);
	} while ((status & (BSY|DRQ)) != DRQ);

	for (i = 0; i < 256; i++)
		*p++ = wdread(cmdreg, DATA);

	wdread(ctlreg, 0);
	status = wdread(cmdreg, STATUS);

	/* Service software watchdog timer. */
	sswt();

	return status;		/* status & (BSY|DF|ERR) */
}

int
dkopen(struct open_file *f, ...)
{
	DPRINTF("dkopen\n");
	return 0;
}

int
dkclose(f)
	struct open_file *f;
{
	DPRINTF("dkclose\n");
	return 0;
}

int
dkstrategy(devdata, rw, block, size, buf, rsize)
	void *devdata;
	int rw;
	daddr_t block;
	size_t size;
	void *buf;
	size_t *rsize;
{
	int len = size;
	static char wdbuf[512];

	DPRINTF("dkstrategy buf=%p block=%lld, size=%d\n", buf, block, size);

	while (len > 0) {
		readsector(block, wdbuf);
		bcopy(wdbuf, buf, min(sizeof wdbuf, len));
		len -= sizeof wdbuf;
		block++;
		buf = (char *)buf + sizeof wdbuf;
	}
	*rsize = size;
	return 0;
}
