/*	$Id: fec.c,v 1.2 2002/07/14 14:20:37 tsubai Exp $	*/

/*-
 * Copyright (c) 2001,2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>

#include <lib/libsa/stand.h>
#include <machine/mpc860dev.h>

#define M24C16W_ETHER_OFFSET	0x7d8
#define M24C16W_ADDR(off)	(0xa0 | ((off) & 0x700) >> 7)
#define M24C16W_OFF(off)	((off) & 0xff)

char fec_enaddr[6];

int i2c_read(int, int, int, void *);		/* XXX */
int getsecs(void);
void sswt(void);

static inline void
eieio()
{
	asm volatile ("eieio; sync");
}

void
fecinit()
{
	volatile struct mpc860dev *immr = get_immr();
	struct cpm_bdesc *TxBD, *RxBD;
	int enaddr[2];
	char *buf;
	u_char *ea = fec_enaddr;

	i2c_read(M24C16W_ADDR(M24C16W_ETHER_OFFSET),
	    M24C16W_ETHER_OFFSET & 0xff, 6, enaddr);
	bcopy(enaddr, fec_enaddr, 6);

	buf = alloc(sizeof(struct cpm_bdesc) * 4 + 16);
	TxBD = (void *)roundup((u_int)buf, 16);

	buf = alloc(0x800 * 4 + 16);
	buf = (void *)roundup((u_int)buf, 16);
	TxBD[0].bd_flags = 0;
	TxBD[0].bd_addr = buf;
	buf += 0x800;
	TxBD[1].bd_flags = 0;
	TxBD[1].bd_addr = buf;

	buf += 0x800;
	RxBD = &TxBD[2];
	RxBD[0].bd_flags = RxBD_E;
	RxBD[0].bd_addr = buf;
	buf += 0x800;
	RxBD[1].bd_flags = RxBD_E|RxBD_W;
	RxBD[1].bd_addr = buf;

	immr->fec_imask = 0;			/* Mask all interrupts. */
	immr->fec_ievent = 0xffff;		/* Clear all events. */
	immr->fec_ivec = 0;
	immr->fec_addr_high = ea[0] << 24 | ea[1] << 16 | ea[2] << 8 | ea[3];
	immr->fec_addr_low  = ea[4] << 24 | ea[5] << 16;
	immr->fec_hash_table_high = 0;
	immr->fec_hash_table_low = 0;
	immr->fec_r_buff_size = 0x5f0;
	immr->fec_r_des_start = RxBD;
	immr->fec_x_des_start = TxBD;
	immr->fec_r_cntrl = 0x6;		/* XXX half-duplex */
	immr->fec_x_cntrl = 0x0;
	immr->fec_fun_code = 0x7f000000;

	immr->cpm_pdpar = 0x1fff;
	immr->cpm_pddir = 0x1fff;
	eieio();

	/* XXX Should we reset FEC here? */

	immr->fec_ecntrl = 0x04 | 0x02;		/* Enable ethernet */
	immr->fec_r_des_active = 0x100;		/* Enable receiver. */
	eieio();
}

int
fecsend(buf, len)
	char *buf;
	int len;
{
	volatile struct mpc860dev *immr = get_immr();
	volatile static struct cpm_bdesc *TxBD;

	if (len > 1520) {
		printf("fecsend: len > 1520 (%d)\n", len);
		len = 1520;
	}

	if (TxBD == NULL)
		TxBD = immr->fec_x_des_start;

	/* Wait for transmission done. */
	while (TxBD->bd_flags & TxBD_R)
		sswt();

	bcopy(buf, TxBD->bd_addr, len);
	if (TxBD == immr->fec_x_des_start) {
		TxBD->bd_flags = TxBD_R|TxBD_L|TxBD_TC;
		TxBD->bd_len = len;
		TxBD++;
		TxBD->bd_flags = TxBD_W;
	} else {
		TxBD->bd_flags = TxBD_R|TxBD_L|TxBD_TC|TxBD_W;
		TxBD->bd_len = len;
		TxBD = immr->fec_x_des_start;
		TxBD->bd_flags = 0;
	}

	eieio();

	/* Start transmission. */
	immr->fec_x_des_active = 0x0100;
	eieio();

	return len;
}

int
fecrecv(buf, timo)
	char *buf;
	int timo;
{
	volatile struct mpc860dev *immr = get_immr();
	volatile struct cpm_bdesc *RxBD;
	int len;
	int t0;

	timo++;
	t0 = getsecs();

	for (;;) {
		RxBD = immr->fec_r_des_start;
		if ((RxBD->bd_flags & RxBD_E) == 0) {
			RxBD->bd_flags = RxBD_E;
			break;
		}
		RxBD++;
		if ((RxBD->bd_flags & RxBD_E) == 0) {
			RxBD->bd_flags = RxBD_E|RxBD_W;
			break;
		}

		if (getsecs() - t0 > timo)
			return -1;
		sswt();
	}

	len = RxBD->bd_len;
	if (len > 1520) {
		printf("fecrecv: len > 1520 (%d)\n", len);
		return -1;
	}
	bcopy(RxBD->bd_addr, buf, len);
	eieio();

	immr->fec_r_des_active = 0x100;		/* XXX needed? */

	return len;
}
