/*	$Id: scc.c,v 1.2 2002/07/14 14:20:37 tsubai Exp $	*/

/*-
 * Copyright (c) 2001 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <lib/libsa/stand.h>
#include <machine/mpc860dev.h>

static inline void
eieio()
{
	asm volatile ("eieio; sync");
}


void
scc_uart_init()
{
	volatile struct mpc860dev *immr = get_immr();
	struct cpm_bdesc *txbd, *rxbd;
	struct scc_uart *scc = (void *)immr->cpm_pram1;
	int i;

	immr->cpm_papar |=  0x0003;	/* PA14 = TXD1, PA15 = RXD1 */
	immr->cpm_padir &= ~0x0003;	/* Select func 0 on pin 14 and 15. */
	immr->cpm_paodr &= ~0x0002;	/* TXD1 is NOT open-drain */

	immr->cpm_pbpar |=  0x1000;	/* PB19 = RTS1 */
	immr->cpm_pbdir |=  0x1000;	/* Select function 1 on pin 19. */
	immr->cpm_pbodr &= ~0x1000;

	immr->cpm_pcpar &= ~0x0030;	/* PC10 = CD1, PC11 = CTS1 */
	immr->cpm_pcdir &= ~0x0030;
	immr->cpm_pcso  |=  0x0030;

	immr->cpm_brgc2 = 0x1028a;	/* Enable BRG2, div = 325 (9600bps) */
	immr->cpm_sicr &= ~0xff;
	immr->cpm_sicr |=  0x09;	/* Connect BRG2 to SCC1 */

	/* siu_sdcr = 1; */

	rxbd = (void *)0xff002820;
	txbd = (void *)0xff002830;
	scc->rbase = (int)rxbd;
	scc->tbase = (int)txbd;

	rxbd->bd_flags = 0;
	rxbd->bd_addr = (char *)0xff002840;
	rxbd++;
	rxbd->bd_flags = 0;
	rxbd->bd_addr = (char *)0xff002842;

	txbd->bd_flags = 0;
	txbd->bd_addr = (char *)0xff002844;
	txbd++;
	txbd->bd_flags = 0;
	txbd->bd_addr = (char *)0xff002846;

	immr->cpm_cpcr = 0x0001;	/* Initialize SCC1 */

	scc->rfcr = 0x10;
	scc->tfcr = 0x10;
	scc->mrblr = 1;		/* maximum receive length */
	scc->max_idl = 0;
	scc->brkcr = 1;
	scc->parec = 0;
	scc->frmec = 0;
	scc->brkec = 0;
	scc->uaddr1 = 0;
	scc->uaddr2 = 0;
	scc->toseq = 0;

	for (i = 0; i < 8; i++)
		scc->character[i] = 0x8000;

	scc->rccm = 0xc0ff;

	rxbd = (void *)0xff002820;
	txbd = (void *)0xff002830;

	rxbd->bd_flags = 0x8000;
	rxbd++;
	rxbd->bd_flags = 0xa000;

	txbd->bd_flags = 0x2000;
	txbd->bd_len = 1;

	immr->cpm_scce1 = 0xffff;	/* Clear any previous events. */
	immr->cpm_sccm1 = 0;		/* No interrupts. */

	immr->cpm_gsmr_h1 = 0x00000020;
	immr->cpm_gsmr_l1 = 0x00028004;

	immr->cpm_psmr1 = 0xb000;

	immr->cpm_gsmr_l1 = 0x00028034;	/* Enable Tx and Rx */
}

void
putchar(c)
	int c;
{
	volatile struct cpm_bdesc *txbd = (void *)0xff002830;

	if (c == '\n')
		putchar('\r');

	while (txbd->bd_flags & 0x8000);    
	eieio();

	*(txbd->bd_addr) = c;
	txbd->bd_len = 1;

	eieio();
	txbd->bd_flags = 0xa000;
}
