/*	$Id: bat.h,v 1.8 2002/11/17 09:14:39 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _POWERPC_BAT_H_
#define _POWERPC_BAT_H_

#ifndef _LOCORE
struct bat {
	u_int32_t batu;
	u_int32_t batl;
};

#ifdef _KERNEL
extern struct bat battable[16];
#endif
#endif

#define BATU(va, len)	(((va) & BATU_BEPI) | ((len) & BATU_BL) | BATU_Vs)
#define BATL(pa, wimg)	(((pa) & BATL_BRPN) | (wimg) | BATL_RW)

#define BATU_BEPI	0xfffe0000	/* Block Effective Page Index */
#define BATU_BL		0x0001fffc	/* Block Length */
#define BATU_Vs		0x00000002	/* Valid in supervisor mode */
#define BATU_Vp		0x00000001	/* Valid in user mode */

#define BATL_BRPN	0xfffe0000	/* Block Real Page Number */
#define BATL_WIMG	0x00000078	/* Access modes */
#define BATL_PP		0x00000003	/* Protection */

/* BATU_BL definitions */
#define BATU_128K	0x00000000
#define BATU_256K	0x00000004
#define BATU_512K	0x0000000c
#define BATU_1M		0x0000001c
#define BATU_2M		0x0000003c
#define BATU_4M		0x0000007c
#define BATU_8M		0x000000fc
#define BATU_16M	0x000001fc
#define BATU_32M	0x000003fc
#define BATU_64M	0x000007fc
#define BATU_128M	0x00000ffc
#define BATU_256M	0x00001ffc
#define BATU_512M	0x00003ffc	/* HIGH_BAT_EN */
#define BATU_1G		0x00007ffc	/* HIGH_BAT_EN */
#define BATU_2G		0x0000fffc	/* HIGH_BAT_EN */
#define BATU_4G		0x0001fffc	/* HIGH_BAT_EN */

/* BATL_WIMG definitions */
#define BATL_W		0x40
#define BATL_I		0x20
#define BATL_M		0x10
#define BATL_G		0x08		/* (D-BAT only ) */

/* BATL_PP definitions */
#define BATL_NONE	0x00
#define BATL_RO		0x01
#define BATL_RW		0x02

#endif /* _POWERPC_BAT_H_ */
