/*	$Id: cpu.h,v 1.23 2003/09/30 10:03:36 tsubai Exp $	*/

/*
 * Copyright (C) 1999 Wolfgang Solfrank.
 * Copyright (C) 1999 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _POWERPC_CPU_H_
#define _POWERPC_CPU_H_

/*
 * CTL_MACHDEP definitions.
 */
#define CPU_CACHELINE	1
#define CPU_TIMEBASE	2
#define CPU_CPUTEMP	3
#define CPU_BOOTED_KERNEL 4
#define CPU_CACHEINFO	5
#define CPU_ALTIVEC	6
#define CPU_MAXID	7

#define CTL_MACHDEP_NAMES { \
	{ 0, 0 }, \
	{ "cachelinesize", CTLTYPE_INT }, \
	{ "timebase", CTLTYPE_INT }, \
	{ "cputemperature", CTLTYPE_INT }, \
	{ "booted_kernel", CTLTYPE_STRING }, \
	{ "cacheinfo", CTLTYPE_STRUCT }, \
	{ "altivec", CTLTYPE_INT }, \
}

struct cache_info {
	int dcache_size;
	int dcache_line_size;
	int icache_size;
	int icache_line_size;
};

void __syncicache(void *, int);

#ifdef _KERNEL

#define MPC601		1
#define MPC602		5
#define MPC603		3
#define MPC603e		6
#define MPC603ev	7
#define MPC604		4
#define MPC604ev	9
#define MPC620		20
#define MPC750		8
#define PPC750FX	0x7000
#define MPC7400		12
#define MPC7410		0x800c
#define MPC7450		0x8000
#define MPC7455		0x8001
#define MPC7457		0x8002

#ifdef MULTIPROCESSOR

struct cpu_hatch_data {
	struct device *dev;
	int running;
	int pir;
	int hid0, hid1;
	int l2cr, l3cr;
	int msscr0;
	int sdr1;
	int sr[16];
	int tbu, tbl;
};

#endif

struct trapframe;
void trap(struct trapframe *);
void userret(struct lwp *, struct trapframe *);
int badaddr(void *, size_t);
void ofbcopy(const void *, void *, size_t);
int setfault(int *);				/* XXX setfault(faultbuf); */
void switchexit(struct lwp *, void (*)(struct lwp *));
void mpc7455_xbsen(void);
void mpc7450_print_hid(char *);
void mpc7450_enable_L1cache(void);
void mpc7450_enable_L2cache(void);
void mpc7450_enable_L3cache(u_int);
u_int mpc7450_corefreq(u_int);
void ppc750fx_clearbat(void);
void ppc750fx_config_l2cr(void);
void ppc750fx_print_hid(char *);

#ifdef MULTIPROCESSOR
void mpc7450_initialize(void);
#endif

#endif /* _KERNEL */

#endif /* _POWERPC_CPU_H_ */
