/*	$Id: altivec.c,v 1.8 2003/01/24 17:52:35 tsubai Exp $	*/

/*-
 * Copyright (c) 2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/proc.h>
#include <sys/systm.h>
#include <sys/user.h>

#include <machine/altivec.h>
#include <machine/psl.h>
#include <machine/spr.h>

void
enable_vec(l)
	struct lwp *l;
{
	struct pcb *pcb = &l->l_addr->u_pcb;
	struct trapframe *tf = trapframe(l);
	u_int msr, vrsave;

	tf->srr1 |= PSL_VEC;
	if (!(pcb->pcb_flags & PCB_VEC)) {
		bzero(&pcb->pcb_vec, sizeof pcb->pcb_vec);
		pcb->pcb_vec.vscr[3] |= VSCR_NJ;
		pcb->pcb_flags |= PCB_VEC;
#ifdef DIAGNOSTIC
		if ((u_int)&pcb->pcb_vec & 0x0f)
			panic("pcb_vec is not aligned");
#endif
	}
	asm volatile ("mfmsr %0" : "=r"(msr));
	asm volatile ("mtmsr %0" :: "r"(msr | PSL_VEC));
	asm volatile ("isync");

	vrsave = pcb->pcb_vec.vrsave;
	asm volatile ("mtspr VRSAVE,%0" :: "r"(vrsave));
	asm volatile ("lvx 0,0,%0; mtvscr 0" :: "r"(&pcb->pcb_vec.vscr));

	/* XXX Load all of the vector registers for security reason. */
	asm volatile ("lvx 0,0,%0" :: "r"(pcb->pcb_vec.vreg[0]));
	asm volatile ("lvx 1,0,%0" :: "r"(pcb->pcb_vec.vreg[1]));
	asm volatile ("lvx 2,0,%0" :: "r"(pcb->pcb_vec.vreg[2]));
	asm volatile ("lvx 3,0,%0" :: "r"(pcb->pcb_vec.vreg[3]));
	asm volatile ("lvx 4,0,%0" :: "r"(pcb->pcb_vec.vreg[4]));
	asm volatile ("lvx 5,0,%0" :: "r"(pcb->pcb_vec.vreg[5]));
	asm volatile ("lvx 6,0,%0" :: "r"(pcb->pcb_vec.vreg[6]));
	asm volatile ("lvx 7,0,%0" :: "r"(pcb->pcb_vec.vreg[7]));
	asm volatile ("lvx 8,0,%0" :: "r"(pcb->pcb_vec.vreg[8]));
	asm volatile ("lvx 9,0,%0" :: "r"(pcb->pcb_vec.vreg[9]));
	asm volatile ("lvx 10,0,%0" :: "r"(pcb->pcb_vec.vreg[10]));
	asm volatile ("lvx 11,0,%0" :: "r"(pcb->pcb_vec.vreg[11]));
	asm volatile ("lvx 12,0,%0" :: "r"(pcb->pcb_vec.vreg[12]));
	asm volatile ("lvx 13,0,%0" :: "r"(pcb->pcb_vec.vreg[13]));
	asm volatile ("lvx 14,0,%0" :: "r"(pcb->pcb_vec.vreg[14]));
	asm volatile ("lvx 15,0,%0" :: "r"(pcb->pcb_vec.vreg[15]));
	asm volatile ("lvx 16,0,%0" :: "r"(pcb->pcb_vec.vreg[16]));
	asm volatile ("lvx 17,0,%0" :: "r"(pcb->pcb_vec.vreg[17]));
	asm volatile ("lvx 18,0,%0" :: "r"(pcb->pcb_vec.vreg[18]));
	asm volatile ("lvx 19,0,%0" :: "r"(pcb->pcb_vec.vreg[19]));
	asm volatile ("lvx 20,0,%0" :: "r"(pcb->pcb_vec.vreg[20]));
	asm volatile ("lvx 21,0,%0" :: "r"(pcb->pcb_vec.vreg[21]));
	asm volatile ("lvx 22,0,%0" :: "r"(pcb->pcb_vec.vreg[22]));
	asm volatile ("lvx 23,0,%0" :: "r"(pcb->pcb_vec.vreg[23]));
	asm volatile ("lvx 24,0,%0" :: "r"(pcb->pcb_vec.vreg[24]));
	asm volatile ("lvx 25,0,%0" :: "r"(pcb->pcb_vec.vreg[25]));
	asm volatile ("lvx 26,0,%0" :: "r"(pcb->pcb_vec.vreg[26]));
	asm volatile ("lvx 27,0,%0" :: "r"(pcb->pcb_vec.vreg[27]));
	asm volatile ("lvx 28,0,%0" :: "r"(pcb->pcb_vec.vreg[28]));
	asm volatile ("lvx 29,0,%0" :: "r"(pcb->pcb_vec.vreg[29]));
	asm volatile ("lvx 30,0,%0" :: "r"(pcb->pcb_vec.vreg[30]));
	asm volatile ("lvx 31,0,%0" :: "r"(pcb->pcb_vec.vreg[31]));

	asm volatile ("mtmsr %0; isync" :: "r"(msr));
}

void
save_vec(l)
	struct lwp *l;
{
	struct pcb *pcb = &l->l_addr->u_pcb;
	u_int msr, vrsave;

	asm volatile ("mfmsr %0" : "=r"(msr));
	asm volatile ("mtmsr %0" :: "r"(msr | PSL_VEC));
	asm volatile ("isync");

	asm volatile ("mfspr %0,VRSAVE" : "=r"(vrsave));
	pcb->pcb_vec.vrsave = vrsave;

	if (vrsave & (0x80000000 >> 0))
		asm volatile ("stvx 0,0,%0" :: "r"(pcb->pcb_vec.vreg[0]));
	if (vrsave & (0x80000000 >> 1))
		asm volatile ("stvx 1,0,%0" :: "r"(pcb->pcb_vec.vreg[1]));
	if (vrsave & (0x80000000 >> 2))
		asm volatile ("stvx 2,0,%0" :: "r"(pcb->pcb_vec.vreg[2]));
	if (vrsave & (0x80000000 >> 3))
		asm volatile ("stvx 3,0,%0" :: "r"(pcb->pcb_vec.vreg[3]));
	if (vrsave & (0x80000000 >> 4))
		asm volatile ("stvx 4,0,%0" :: "r"(pcb->pcb_vec.vreg[4]));
	if (vrsave & (0x80000000 >> 5))
		asm volatile ("stvx 5,0,%0" :: "r"(pcb->pcb_vec.vreg[5]));
	if (vrsave & (0x80000000 >> 6))
		asm volatile ("stvx 6,0,%0" :: "r"(pcb->pcb_vec.vreg[6]));
	if (vrsave & (0x80000000 >> 7))
		asm volatile ("stvx 7,0,%0" :: "r"(pcb->pcb_vec.vreg[7]));
	if (vrsave & (0x80000000 >> 8))
		asm volatile ("stvx 8,0,%0" :: "r"(pcb->pcb_vec.vreg[8]));
	if (vrsave & (0x80000000 >> 9))
		asm volatile ("stvx 9,0,%0" :: "r"(pcb->pcb_vec.vreg[9]));
	if (vrsave & (0x80000000 >> 10))
		asm volatile ("stvx 10,0,%0" :: "r"(pcb->pcb_vec.vreg[10]));
	if (vrsave & (0x80000000 >> 11))
		asm volatile ("stvx 11,0,%0" :: "r"(pcb->pcb_vec.vreg[11]));
	if (vrsave & (0x80000000 >> 12))
		asm volatile ("stvx 12,0,%0" :: "r"(pcb->pcb_vec.vreg[12]));
	if (vrsave & (0x80000000 >> 13))
		asm volatile ("stvx 13,0,%0" :: "r"(pcb->pcb_vec.vreg[13]));
	if (vrsave & (0x80000000 >> 14))
		asm volatile ("stvx 14,0,%0" :: "r"(pcb->pcb_vec.vreg[14]));
	if (vrsave & (0x80000000 >> 15))
		asm volatile ("stvx 15,0,%0" :: "r"(pcb->pcb_vec.vreg[15]));
	if (vrsave & (0x80000000 >> 16))
		asm volatile ("stvx 16,0,%0" :: "r"(pcb->pcb_vec.vreg[16]));
	if (vrsave & (0x80000000 >> 17))
		asm volatile ("stvx 17,0,%0" :: "r"(pcb->pcb_vec.vreg[17]));
	if (vrsave & (0x80000000 >> 18))
		asm volatile ("stvx 18,0,%0" :: "r"(pcb->pcb_vec.vreg[18]));
	if (vrsave & (0x80000000 >> 19))
		asm volatile ("stvx 19,0,%0" :: "r"(pcb->pcb_vec.vreg[19]));
	if (vrsave & (0x80000000 >> 20))
		asm volatile ("stvx 20,0,%0" :: "r"(pcb->pcb_vec.vreg[20]));
	if (vrsave & (0x80000000 >> 21))
		asm volatile ("stvx 21,0,%0" :: "r"(pcb->pcb_vec.vreg[21]));
	if (vrsave & (0x80000000 >> 22))
		asm volatile ("stvx 22,0,%0" :: "r"(pcb->pcb_vec.vreg[22]));
	if (vrsave & (0x80000000 >> 23))
		asm volatile ("stvx 23,0,%0" :: "r"(pcb->pcb_vec.vreg[23]));
	if (vrsave & (0x80000000 >> 24))
		asm volatile ("stvx 24,0,%0" :: "r"(pcb->pcb_vec.vreg[24]));
	if (vrsave & (0x80000000 >> 25))
		asm volatile ("stvx 25,0,%0" :: "r"(pcb->pcb_vec.vreg[25]));
	if (vrsave & (0x80000000 >> 26))
		asm volatile ("stvx 26,0,%0" :: "r"(pcb->pcb_vec.vreg[26]));
	if (vrsave & (0x80000000 >> 27))
		asm volatile ("stvx 27,0,%0" :: "r"(pcb->pcb_vec.vreg[27]));
	if (vrsave & (0x80000000 >> 28))
		asm volatile ("stvx 28,0,%0" :: "r"(pcb->pcb_vec.vreg[28]));
	if (vrsave & (0x80000000 >> 29))
		asm volatile ("stvx 29,0,%0" :: "r"(pcb->pcb_vec.vreg[29]));
	if (vrsave & (0x80000000 >> 30))
		asm volatile ("stvx 30,0,%0" :: "r"(pcb->pcb_vec.vreg[30]));
	if (vrsave & (0x80000000 >> 31))
		asm volatile ("stvx 31,0,%0" :: "r"(pcb->pcb_vec.vreg[31]));

	asm volatile ("mfvscr 0; stvx 0,0,%0" :: "r"(&pcb->pcb_vec.vscr));

	asm volatile ("mtmsr %0; isync" :: "r"(msr));
	pcb->pcb_veccpu = NULL;
	veclwp = NULL;
}
