/*	$Id: cpu_7450.c,v 1.18 2003/09/30 10:58:51 tsubai Exp $	*/

/*-
 * Copyright (c) 2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_multiprocessor.h"

#include <sys/param.h>
#include <sys/systm.h>

#include <machine/bat.h>
#include <machine/cpufunc.h>
#include <machine/spr.h>

void config_l3cr(void);

/*
 * 7450/7455 specific part.
 */
#define HID0_TBEN    0x04000000	/* Timebase enable */
#define HID0_STEN    0x01000000	/* Software table search enable */
#define HID0_HBATEN  0x00800000	/* Enable BAT[4-7] (7455) */
#define HID0_NAP     0x00400000	/* Enable nap mode */
#define HID0_SLEEP   0x00200000	/* Enable sleep mode */
#define HID0_DPM     0x00100000	/* Enable Dynamic power management */
#define HID0_BHTCLR  0x00080000	/* Clear branch history table */
#define HID0_XAEN    0x00040000	/* Enable extended Addressing */
#define HID0_NHR     0x00010000	/* Not hard reset */
#define HID0_ICE     0x00008000	/* Enable i-cache */
#define HID0_DCE     0x00004000	/* Enable d-cache */
#define HID0_ILOCK   0x00002000	/* i-cache lock */
#define HID0_DLOCK   0x00001000	/* d-cache lock */
#define HID0_ICFI    0x00000800	/* i-cache flush invalidate */
#define HID0_DCFI    0x00000400	/* d-cache flush invalidate */
#define HID0_SPD     0x00000200	/* Disable speculative cache access */
#define HID0_XBSEN   0x00000100	/* Enable extended BAT size (7455) */
#define HID0_SGE     0x00000080	/* Enable store gathering */
#define HID0_BTIC    0x00000020	/* Enable BTIC */
#define HID0_LRSTK   0x00000010	/* Enable link register stack */
#define HID0_FOLD    0x00000008	/* Enable branch folding */
#define HID0_BHT     0x00000004	/* Enable branch history table */
#define HID0_NOPDST  0x00000002	/* No-op dst, dstt, dstst, and dststt */
#define HID0_NOPTI   0x00000001	/* No-op the dcbt(st) */

#define HID0_BITMASK "\020" \
    "\033TBEN\031STEN\030HBATEN" \
    "\027NAP\026SLEEP\025DPM\023BHTCLR\022XAEN\021NHR" \
    "\020ICE\017DCE\016ILOCK\015DLOCK\014ICFI\013DCFI\012SPD\011XBSEN" \
    "\010SGE\006BTIC\005LRSTK\004FOLD\003BHT\002NOPDST\001NOPTI"

#define HID1_EMCP    0x80000000	/* Enable MCP */
#define HID1_EBA     0x20000000	/* Enable bus addr parity checking */
#define HID1_EBD     0x10000000	/* Enable bus data parity checking */
#define HID1_BCLK    0x08000000	/* CLK_OUT clock type selection */
#define HID1_ECLK    0x02000000	/* CLK_OUT clock type selection */
#define HID1_PAR     0x01000000	/* Disable precharge of ARTRY */
#define HID1_PC0     0x00010000	/* PLL configuration bit 0 (RO) */
#define HID1_PC1     0x00008000	/* PLL configuration bit 1 (RO) */
#define HID1_PC2     0x00004000	/* PLL configuration bit 2 (RO) */
#define HID1_PC3     0x00002000	/* PLL configuration bit 3 (RO) */
#define HID1_PC4     0x00001000	/* PLL configuration bit 4 (RO) */
#define HID1_SYNCBE  0x00000800	/* Enable address broadcast for sync, eieio */
#define HID1_ABE     0x00000400	/* Enable address broadcast for dcbf, dcbst,
                                   dcbi, icbi, tlbie, and tlbsync */

#define HID1_BITMASK "\020" \
    "\040EMCP\036EBA\035EBD\034BCLK\032ECLK\031PAR" \
    "\021PC0\020PC1" \
    "\017PC2\016PC3\015PC4\014SYNCBE\013ABE"

#define L2CR_L2E     0x80000000	/* 0: L2 enable */
#define L2CR_L2PE    0x40000000	/* 1: L2 data parity checking enable */
				/* 2-3: reserved (must be 0b00) */
#define L2CR_L2I     0x00200000	/* 10: L2 global invalidate */
#define L2CR_L2IO    0x00100000	/* 11: L2 instruction only */
#define L2CR_L3OH0   0x00080000	/* 12: L3 output hold 0 */
#define L2CR_L2DO    0x00010000	/* 15: L2 data only */
#define L2CR_L2REP   0x00001000	/* 19: L2 replacement algorithm */
#define L2CR_L2HWF   0x00000800	/* 20: L2 hardware flush */

#define L3CR_L3E     0x80000000	/* 0: L3 enable */
#define L3CR_L3PE    0x40000000	/* 1: L3 data parity checking enable */
#define L3CR_L3APE   0x20000000	/* 2: L3 address parity checking enable */
#define L3CR_L3SIZ   0x10000000	/* 3: L3 size (0:1MB, 1:2MB) */
#define L3CR_L3CLKEN 0x08000000	/* 4: enable L3_CLK signals */
				/* 5: reserved (must be 1) */
#define L3CR_L3CLK   0x03800000	/* 6-8: L3 clock ratio */
#define  L3CLK_6	0x00000000	/* 1 / 6   (!7457) */
#define  L3CLK_2	0x01000000	/* 1 / 2   (!7457) */
#define  L3CLK_25	0x01800000	/* 1 / 2.5 (!7457) */
#define  L3CLK_3	0x02000000	/* 1 / 3   (!7457) */
#define  L3CLK_35	0x02800000	/* 1 / 3.5 (!7457) */
#define  L3CLK_4	0x03000000	/* 1 / 4   (!7457) */
#define  L3CLK_5	0x03800000	/* 1 / 5   (!7457) */
#define  L3CLK_7	0x00000000	/* 1 / 7   (7457) */
#define  L3CLK_8	0x00800000	/* 1 / 8   (7457) */
#define  L3CLK_45	0x01000000	/* 1 / 4.5 (7457) */
#define  L3CLK_55	0x01800000	/* 1 / 5.5 (7457) */
#define  L3CLK_65	0x02000000	/* 1 / 6.5 (7457) */
#define  L3CLK_75	0x02800000	/* 1 / 7.5 (7457) */
#define L3CR_L3IO    0x00400000	/* 9: L3 instruction only */
#define L3CR_L3CLKEXT 0x00100000 /* 10: L3 clock ratio extension (7457) */
#define L3CR_L3CKSPEXT 0x00100000 /* 11: L3 clock sample point ext. (7457) */
#define L3CR_L3OH1   0x00080000	/* 12: L3 output hold 1 (7455) */
#define L3CR_L3SPO   0x00040000	/* 13: L3 sample point override */
#define L3CR_L3CKSP  0x00030000	/* 14-15: L3 clock sample point */
#define L3CR_L3PSP   0x0000e000	/* 16-18: L3 P-clock sample point */
#define L3CR_L3REP   0x00001000	/* 19: L3 replacement algorithm */
#define L3CR_L3HWF   0x00000800	/* 20: L3 hardware flush */
#define L3CR_L3I     0x00000400	/* 21: L3 global invalidate */
#define L3CR_L3RT    0x00000300	/* 22-23: L3 SRAM type */
#define  L3RT_DDR	0x00000000	/* MSUG2 DDR */
#define  L3RT_LATE	0x00000100	/* pipelined synchronous late-write */
#define  L3RT_PB2	0x00000300	/* PB2 */
#define L3CR_L3NIRCA 0x00000080	/* 24: L3 SRAM clock control */
#define L3CR_L3DO    0x00000040	/* 25: L3 data only */
#define L3CR_PMEN    0x00000004	/* 29: private memory enable */
#define L3CR_PMSIZ   0x00000003	/* 30-31: private memory size */
#define  L3CR_PMSIZ_1M  0x00000000	/* 1MB */
#define  L3CR_PMSIZ_2M  0x00000001	/* 2MB */
#define  L3CR_PMSIZ_4M  0x00000002	/* 4MB (7457) */

/*
 * Enable extended BAT size to use BAT_BL_1G to map physical memory.
 */
void
mpc7455_xbsen()
{
	u_int hid0;

	asm ("mfspr %0,HID0" : "=r"(hid0));
	hid0 |= HID0_XBSEN;
	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
	asm volatile ("isync");
}

/*
 * Print HID[01] configuration.
 */
void
mpc7450_print_hid(name)
	char *name;
{
	u_int hid0, hid1;
	char buf[128];

	asm ("mfspr %0,HID0" : "=r"(hid0));
	asm ("mfspr %0,HID1" : "=r"(hid1));

	bitmask_snprintf(hid0, HID0_BITMASK, buf, sizeof buf);
	printf("%s: HID0 %s\n", name, buf);
	bitmask_snprintf(hid1, HID1_BITMASK, buf, sizeof buf);
	printf("%s: HID1 %s\n", name, buf);
}

void
mpc7450_enable_L1cache()
{
	u_int hid0;

	/* Enable L1 caches. */
	asm volatile ("mfspr %0,HID0" : "=r"(hid0));
	hid0 |= HID0_ICFI|HID0_ICE;
	asm volatile ("isync");
	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
	asm volatile ("isync");

	asm volatile ("mfspr %0,HID0" : "=r"(hid0));
	hid0 |= HID0_DCFI|HID0_DCE;
	asm volatile ("sync");
	asm volatile ("mtspr HID0,%0" :: "r"(hid0));

	/* XXX ICTRL[EICP] */
}

/*
 * Invalidate and enable the L2 cache.
 */
void
mpc7450_enable_L2cache()
{
	u_int l2cr;

	/* Invalidate and enable L2 cache. */
	asm volatile ("mfspr %0,L2CR" : "=r"(l2cr));

	/* Invalidate the entire L2 cache. */
	/* asm volatile ("dssall; sync"); */
	l2cr |= L2CR_L2I;
	asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));

	/* Wait for the invalidation done. */
	while (l2cr & L2CR_L2I)
		asm volatile ("mfspr %0,L2CR" : "=r"(l2cr));

	/* Enable L2 cache. */
	l2cr = L2CR_L2E;
	asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));

	/* XXX L2PE */
}

/*
 * Invalidate and enable the L3 cache.
 */
void
mpc7450_enable_L3cache(l3param)
	u_int l3param;
{
	u_int l3cr;

	l3cr = l3param & L3CR_L3CLK;
	asm volatile ("mtspr L3CR,%0" :: "r"(l3cr));

	l3cr = l3param & ~(L3CR_L3E|L3CR_L3I|L3CR_L3PE|L3CR_L3CLKEN);
	asm volatile ("mtspr L3CR,%0" :: "r"(l3cr));

	/* XXX Make sure L3CR[5] == 1. */

	l3cr |= L3CR_L3CLKEN;
	asm volatile ("mtspr L3CR,%0" :: "r"(l3cr));
	delay(2);

	/* Invalidate the entire L3 cache. */
	/* asm volatile ("dssall; sync"); */
	l3cr |= L3CR_L3I;
	asm volatile ("mtspr L3CR,%0" :: "r"(l3cr));

	/* Wait for the invalidation done. */
	while (l3cr & L3CR_L3I)
		asm volatile ("mfspr %0,L3CR" : "=r"(l3cr));

	l3cr &= ~L3CR_L3CLKEN;
	asm volatile ("mtspr L3CR,%0" :: "r"(l3cr));
	asm volatile ("sync");
	delay(2);

	l3cr |= L3CR_L3E | L3CR_L3CLKEN;
	asm volatile ("mtspr L3CR,%0" :: "r"(l3cr));
	asm volatile ("sync");
	delay(2);
}

u_int
mpc7450_corefreq(busfreq)
	u_int busfreq;
{
	u_int hid1;

	/* Bus to Core Multiplier * 2 */
	static char pll_mul2[] = {
		23, 34, 15, 30, 14, 36,  2, 40,
		 4, 42, 13, 26, 17, 48, 19, 18,
		 6, 21, 11, 22,  8, 20, 10, 24,
		16, 28, 12, 32, 27, 56,  0, 25
	};

	asm ("mfspr %0,HID1" : "=r"(hid1));

	hid1 &= HID1_PC0|HID1_PC1|HID1_PC2|HID1_PC3|HID1_PC4;
	hid1 >>= 12;

	return busfreq * pll_mul2[hid1] / 2;
}

void
config_l3cr()
{
	u_int l2cr, l3cr;

	if (mfpvr() >> 16 == MPC7457)
		printf(": 512KB on-chip L2 cache");
	else
		printf(": 256KB on-chip L2 cache");

	asm ("mfspr %0,L2CR" : "=r"(l2cr));
	asm ("mfspr %0,L3CR" : "=r"(l3cr));

#if __notyet
	/*
	 * Configure L3 cache if not enabled.
	 */
	if ((l3cr & L3CR_L3E) == 0 && l2cr_config != 0) {
		l2cr = l2cr_config;
		asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));

		/* Wait for L2 clock to be stable (640 L2 clocks). */
		delay(100);

		/* Invalidate all L2 contents. */
		l2cr |= L2CR_L2I;
		asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));
		do {
			asm volatile ("mfspr %0,L2CR" : "=r"(x));
		} while (x & L2CR_L2IP);

		/* Enable L2 cache. */
		l2cr &= ~L2CR_L2I;
		l2cr |= L2CR_L2E;
		asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));
	}
#endif

	if (l3cr & L3CR_L3E) {
		if (l3cr & L3CR_L3SIZ)
			printf(", 2MB");
		else
			printf(", 1MB");
		printf(" backside L3 cache");
	}

	printf("\n");
}

#ifdef MULTIPROCESSOR

extern volatile struct cpu_hatch_data *cpu_hatch_data;

void
mpc7450_initialize()
{
	int hid0, l2cr, l3cr;
	volatile struct cpu_hatch_data *h = cpu_hatch_data;

	/* Enable timebase */
	asm volatile ("mfspr %0,HID0" : "=r"(hid0));
	hid0 |= HID0_TBEN;
	asm volatile ("mtspr HID0,%0" :: "r"(hid0));

	asm volatile ("mtspr HID1,%0" :: "r"(h->hid1));
	asm volatile ("mtspr MSSCR0,%0" :: "r"(h->msscr0));

	/* Enable the L1 caches. */
	mpc7450_enable_L1cache();

	/* Enable the L2 cache. */
	asm volatile ("mfspr %0,L2CR" : "=r"(l2cr));
	if ((l2cr & L2CR_L2E) == 0)
		mpc7450_enable_L2cache();

	/* Enable the L3 cache. */
	if (h->l3cr & L3CR_L3E) {
		asm volatile ("mfspr %0,L3CR" : "=r"(l3cr));
		if ((l3cr & L3CR_L3E) == 0)
			mpc7450_enable_L3cache(h->l3cr);
	}
}

#endif
