/*	$Id: cpu_750fx.c,v 1.3 2003/09/20 16:01:27 tsubai Exp $	*/

/*-
 * Copyright (c) 2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>

#include <machine/bat.h>
#include <machine/spr.h>

/*
 * 750FX specific part.
 */
#define HID0_EMCP    0x80000000	/* Enable MCP */
#define HID0_DBP     0x40000000	/* Disable 60x bus parity generation */
#define HID0_EBA     0x20000000	/* Enable 60x bus address parity checking */
#define HID0_EBD     0x10000000	/* Enable 60x bus data parity checking */
#define HID0_PAR     0x01000000	/* Disable precharge of ARTRY */
#define HID0_DOZE    0x00800000	/* Enable doze mode */
#define HID0_NAP     0x00400000	/* Enable nap mode */
#define HID0_SLEEP   0x00200000	/* Enable sleep mode */
#define HID0_DPM     0x00100000	/* Enable Dynamic power management */
#define HID0_RISEG   0x00080000	/* Read I-SEG */
#define HID0_NHR     0x00010000	/* Not hard reset */
#define HID0_ICE     0x00008000	/* Enable i-cache */
#define HID0_DCE     0x00004000	/* Enable d-cache */
#define HID0_ILOCK   0x00002000	/* i-cache lock */
#define HID0_DLOCK   0x00001000	/* d-cache lock */
#define HID0_ICFI    0x00000800	/* i-cache flush invalidate */
#define HID0_DCFI    0x00000400	/* d-cache flush invalidate */
#define HID0_SPD     0x00000200	/* Disable speculative cache access */
#define HID0_IFEM    0x00000100	/* Enable M-bit for I-fetch */
#define HID0_SGE     0x00000080	/* Enable store gathering */
#define HID0_DCFA    0x00000040	/* Data cache flush assist */
#define HID0_BTIC    0x00000020	/* Enable BTIC */
#define HID0_ABE     0x00000008	/* Enable address broadcast */
#define HID0_BHT     0x00000004	/* Enable branch history table */
#define HID0_NOOPTI  0x00000001	/* No-op the dcbt(st) */

#define HID0_BITMASK "\020" \
    "\040EMCP\037DBP\036EBA\035EBD\031PAR" \
    "\030DOZE\027NAP\026SLEEP\025DPM\024RISEG\021NHR" \
    "\020ICE\017DCE\016ILOCK\015DLOCK\014ICFI\013DCFI\012SPD\011IFEM" \
    "\010SGE\007DCFA\006BTIC\004ABE\003BHT\001NOOPTI"

#define HID1_PCE0    0x80000000	/* PLL external configuration bit 0 (RO) */
#define HID1_PCE1    0x40000000	/* PLL external configuration bit 1 (RO) */
#define HID1_PCE2    0x20000000	/* PLL external configuration bit 2 (RO) */
#define HID1_PCE3    0x10000000	/* PLL external configuration bit 3 (RO) */
#define HID1_PCE4    0x08000000	/* PLL external configuration bit 4 (RO) */
#define HID1_PRE0    0x04000000	/* PLL external range bit 0 (RO) */
#define HID1_PRE1    0x02000000	/* PLL external range bit 1 (RO) */
#define HID1_PSS     0x01000000	/* Processor switch status */
#define HID1_ECLK    0x00800000	/* Enable CLKOUT pin */
#define HID1_CLKSEL  0x00700000	/* CLKOUT selection */

#define HID1_PI0     0x00020000	/* PLL0 internal configuration select */
#define HID1_PS      0x00010000	/* PLL select */
#define HID1_PC00    0x00008000	/* PLL0 configuration bit 0 */
#define HID1_PC01    0x00004000	/* PLL0 configuration bit 1 */
#define HID1_PC02    0x00002000	/* PLL0 configuration bit 2 */
#define HID1_PC03    0x00001000	/* PLL0 configuration bit 3 */
#define HID1_PC04    0x00000800	/* PLL0 configuration bit 4 */
#define HID1_PR00    0x00000400	/* PLL0 range selection bit 0 */
#define HID1_PR01    0x00000200	/* PLL0 range selection bit 1 */
#define HID1_PC10    0x00000080	/* PLL1 configuration bit 0 */
#define HID1_PC11    0x00000040	/* PLL1 configuration bit 1 */
#define HID1_PC12    0x00000020	/* PLL1 configuration bit 2 */
#define HID1_PC13    0x00000010	/* PLL1 configuration bit 3 */
#define HID1_PC14    0x00000008	/* PLL1 configuration bit 4 */
#define HID1_PR10    0x00000004	/* PLL1 range selection bit 0 */
#define HID1_PR11    0x00000002	/* PLL1 range selection bit 1 */

#define HID1_BITMASK "\020" \
    "\040PCE0\037PCE1\036PCE2\035PCE3\034PCE4\033PRE0\032PRE1\031PSS" \
    "\030ECLK\022PI0\021PS" \
    "\020PC00\017PC01\016PC02\015PC03\014PC04\013PR00\012PR01" \
    "\010PC10\007PC11\006PC12\006PC13\005PC13\004PC14\003PR10\002PR11"

#define L2CR_L2E     0x80000000	/* 0: L2 enable */
#define L2CR_L2CE    0x40000000	/* 1: L2 checkstop enable */
#define L2CR_L2DO    0x00400000	/* 9: L2 data only */
#define L2CR_L2I     0x00200000	/* 10: L2 global invalidate */
#define L2CR_L2WT    0x00080000	/* 12: write-through */
#define L2CR_L2TS    0x00040000	/* 13: test support */
#define L2CR_L2LOCK  0x00000c00	/* 20-21: L2 lock */
#define L2CR_L2SHEE  0x00000200	/* 22: snoop hit in locked line error enable */
#define L2CR_L2SHERR 0x00000100	/* 23: snoop hit in locked line error */
#define L2CR_L2IP    0x00000001	/* 31: L2 global invalidate in progress */

#define L2CR_BITMASK "\020" \
    "\040L2E\037L2CE\027L2DO\026L2I\024L2WT\023L2TS" \
    "\014L2LOCK0\013L2LOCK1\012L2SHEE\011L2SHERR\001L2IP"

asm ("IBAT4U = 560");
asm ("IBAT4U = 560");
asm ("IBAT5U = 562");
asm ("IBAT6U = 564");
asm ("IBAT7U = 566");
asm ("DBAT4U = 568");
asm ("DBAT5U = 570");
asm ("DBAT6U = 572");
asm ("DBAT7U = 574");

/*
 * Invalidate bat0 - bat7.
 */
void
ppc750fx_clearbat()
{

	asm volatile ("mtibatu 0,%0" :: "r"(0));
	asm volatile ("mtibatu 1,%0" :: "r"(0));
	asm volatile ("mtibatu 2,%0" :: "r"(0));
	asm volatile ("mtibatu 3,%0" :: "r"(0));
#if 0 /* XXX Avoid assembler error. */
	asm volatile ("mtibatu 4,%0" :: "r"(0));
	asm volatile ("mtibatu 5,%0" :: "r"(0));
	asm volatile ("mtibatu 6,%0" :: "r"(0));
	asm volatile ("mtibatu 7,%0" :: "r"(0));
#else
	asm volatile ("mtspr IBAT4U,%0" :: "r"(0));
	asm volatile ("mtspr IBAT5U,%0" :: "r"(0));
	asm volatile ("mtspr IBAT6U,%0" :: "r"(0));
	asm volatile ("mtspr IBAT7U,%0" :: "r"(0));
#endif

	asm volatile ("mtdbatu 0,%0" :: "r"(0));
	asm volatile ("mtdbatu 1,%0" :: "r"(0));
	asm volatile ("mtdbatu 2,%0" :: "r"(0));
	asm volatile ("mtdbatu 3,%0" :: "r"(0));
#if 0 /* XXX Avoid assembler error. */
	asm volatile ("mtdbatu 4,%0" :: "r"(0));
	asm volatile ("mtdbatu 5,%0" :: "r"(0));
	asm volatile ("mtdbatu 6,%0" :: "r"(0));
	asm volatile ("mtdbatu 7,%0" :: "r"(0));
#else
	asm volatile ("mtspr DBAT4U,%0" :: "r"(0));
	asm volatile ("mtspr DBAT5U,%0" :: "r"(0));
	asm volatile ("mtspr DBAT6U,%0" :: "r"(0));
	asm volatile ("mtspr DBAT7U,%0" :: "r"(0));
#endif
}

/*
 * Print HID[01] configuration.
 */
void
ppc750fx_print_hid(name)
	char *name;
{
	u_int hid0, hid1;
	char buf[128];

	asm ("mfspr %0,HID0" : "=r"(hid0));
	asm ("mfspr %0,HID1" : "=r"(hid1));

	bitmask_snprintf(hid0, HID0_BITMASK, buf, sizeof buf);
	printf("%s: HID0 %s\n", name, buf);
	bitmask_snprintf(hid1, HID1_BITMASK, buf, sizeof buf);
	printf("%s: HID1 %s\n", name, buf);
}

void
ppc750fx_config_l2cr()
{
	u_int l2cr;

	asm ("mfspr %0,L2CR" : "=r"(l2cr));

	if ((l2cr & L2CR_L2E) == 0) {
		printf(": on-chip L2 cache disabled\n");
		/* XXX enable here */
		return;
	}

	printf(": 512KB on-chip L2 cache\n");
	l2cr &= L2CR_L2E|L2CR_L2CE;
	asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));
	asm volatile ("sync");
}
