/*	$Id: cpu_common.c,v 1.1 2003/07/04 07:55:42 tsubai Exp $	*/

/*-
 * Copyright (c) 2002,2003 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_l2cr_config.h"

#include <sys/param.h>
#include <sys/reboot.h>
#include <sys/systm.h>

#include <machine/cpufunc.h>
#include <machine/hid.h>
#include <machine/pio.h>
#include <machine/spr.h>

void config_l2cr(void);
void enable_l2cache(void);

#define L2CR_L2E	0x80000000   /* 0: L2 enable */
#define L2CR_L2PE	0x40000000   /* 1: L2 data parity enable */
#define L2CR_L2SIZ	0x30000000   /* 2-3: L2 size */
#define   L2SIZ_2M	  0x00000000 /* (!750) */
#define   L2SIZ_256K	  0x10000000
#define   L2SIZ_512K	  0x20000000
#define   L2SIZ_1M	  0x30000000
#define L2CR_L2CLK	0x0e000000   /* 4-6: L2 clock ratio */
#define   L2CLK_DIS	  0x00000000 /* disable L2 clock */
#define   L2CLK_10	  0x02000000 /* core clock / 1   */
#define   L2CLK_15	  0x04000000 /*            / 1.5 */
#define   L2CLK_35	  0x06000000 /*            / 3.5 (!750) */
#define   L2CLK_20	  0x08000000 /*            / 2   */
#define   L2CLK_25	  0x0a000000 /*            / 2.5 */
#define   L2CLK_30	  0x0c000000 /*            / 3   */
#define   L2CLK_40	  0x0e000000 /*            / 4   (!750) */
#define L2CR_L2RAM	0x01800000   /* 7-8: L2 RAM type */
#define   L2RAM_FLOWTHRU  0x00000000 /* (!7410) */
#define   L2RAM_PB3	  0x00800000 /* (7410) */
#define   L2RAM_PB2	  0x01000000
#define   L2RAM_LATE	  0x01800000
#define L2CR_L2DO	0x00400000   /* 9: L2 data-only. */
#define L2CR_L2I	0x00200000   /* 10: L2 global invalidate. */
#define L2CR_L2CTL	0x00100000   /* 11: L2 RAM control (ZZ enable). */
#define L2CR_L2WT	0x00080000   /* 12: L2 write-through. */
#define L2CR_L2TS	0x00040000   /* 13: L2 test support. */
#define L2CR_L2OH	0x00030000   /* 14-15: L2 output hold. */
#define L2CR_L2SL	0x00008000   /* 16: L2 DLL slow. */
#define L2CR_L2DF	0x00004000   /* 17: L2 differential clock. */
#define L2CR_L2BYP	0x00002000   /* 18: L2 DLL bypass. */
#define L2CR_L2FA	0x00001000   /* 19: L2 flush assist. (!750) */
#define L2CR_L2HWF	0x00000800   /* 20: L2 hardware flush. (!750) */
#define L2CR_L2IO	0x00000400   /* 21: L2 instruction-only. (!750) */
#define L2CR_L2CLKSTP	0x00000200   /* 22: L2 clock stop. (!750) */
#define L2CR_DRO	0x00000100   /* 23: L2DLL rollover checkstop (!750) */
#define L2CR_L2IP	0x00000001   /* 31: L2 global invalidate in progress
					    (read only). */

#ifdef L2CR_CONFIG
u_int l2cr_config = L2CR_CONFIG;
#else
u_int l2cr_config = 0;
#endif

void
config_l2cr()
{
	u_int l2cr;

	asm ("mfspr %0,L2CR" : "=r"(l2cr));

	/*
	 * Configure L2 cache if not enabled.
	 */
	if ((l2cr & L2CR_L2E) == 0 && l2cr_config != 0) {
		enable_l2cache();
		asm volatile ("mfspr %0,L2CR" : "=r"(l2cr));
	}

	if ((l2cr & L2CR_L2E) == 0) {
		printf(": L2 cache not enabled\n");
		return;
	}

	switch (l2cr & L2CR_L2SIZ) {
	case L2SIZ_256K:
		printf(": 256KB");
		break;
	case L2SIZ_512K:
		printf(": 512KB");
		break;
	case L2SIZ_1M:
		printf(": 1MB");
		break;
	case L2SIZ_2M:
		printf(": 2MB");
		break;
	default:
		printf(": unknown size");
	}
	if (bootverbose) {

		switch (l2cr & L2CR_L2CLK) {
		case L2CLK_DIS:
			printf(", (disabled?)"); break;
		case L2CLK_10:
			printf(", (1:1)"); break;
		case L2CLK_15:
			printf(", (1.5:1)"); break;
		case L2CLK_20:
			printf(", (2:1)"); break;
		case L2CLK_25:
			printf(", (2.5:1)"); break;
		case L2CLK_30:
			printf(", (3:1)"); break;
		case L2CLK_35:
			printf(", (3.5:1)"); break;
		case L2CLK_40:
			printf(", (4:1)"); break;
		}

		switch (l2cr & L2CR_L2RAM) {
		case L2RAM_FLOWTHRU:
			printf(", Flow-through synchronous burst SRAM");
			break;
		case L2RAM_PB2:
			printf(", Pipelined synchronous burst (PB2) SRAM");
			break;
		case L2RAM_PB3:
			printf(", Pipelined synchronous burst (PB3) SRAM");
			break;
		case L2RAM_LATE:
			printf(", Pipelined synchronous late-write SRAM");
			break;
		}

		if (l2cr & L2CR_L2PE)
			printf(", with parity");
	}

	printf(" backside L2 cache\n");
}

void
enable_l2cache()
{
	int i, x, size, maxsize;
	u_int msr, l2cr, hid0;
	const int vers = mfpvr() >> 16;
	volatile char *addr = (char *)0x80000000; /* XXX scratch address */

#ifdef DIAGNOSTIC
	asm ("mfspr %0,L2CR" : "=r"(l2cr));
	if (l2cr & L2CR_L2E)
		panic("L2 cache is already enabled");
#endif

	l2cr = l2cr_config & ~(L2CR_L2SIZ | L2CR_L2E);

	switch (vers) {
	case MPC750:
		maxsize = 1024*1024;
		l2cr |= L2SIZ_1M;
		break;
	case MPC7400:
	case MPC7410:
		maxsize = 2048*1024;
		l2cr |= L2SIZ_2M;
		break;
	default:
		return;
	}

	/* XXX Flush L1 cache. */
	for (i = 0; i < 1024*2*4; i++)
		in32(i * 32);

	/* Start L2 clock. */
	asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));

	/* Wait for L2 clock to be stable (640 L2 clocks). */
	delay(1000);

	/* Invalidate all L2 contents. */
	asm volatile ("mtspr L2CR,%0" :: "r"(l2cr | L2CR_L2I));
	do {
		asm volatile ("mfspr %0, L2CR" : "=r"(x));
	} while (x & L2CR_L2IP);

	/*
	 * Probe L2 cache size.
	 */
	/* Disable data address translation. */
	asm volatile ("sync; isync");
	msr = mfmsr();
	mtmsr(msr & ~PSL_DR);

	/* Enable L2 test mode. */
	asm volatile ("mtspr L2CR,%0"
	    :: "r"(l2cr | L2CR_L2E | L2CR_L2DO | L2CR_L2TS));
	asm volatile ("sync; isync");

	/* Zero-fill the entire L2 cache. */
	for (i = 0; i < maxsize; i += CACHELINESIZE)
		asm volatile ("dcbz 0,%0; dcbf 0,%0" :: "r"(addr + i));

	/* Invalidate and lock the L1 d-cache. */
	asm ("mfspr %0,HID0" : "=r"(hid0));
#if 1 /* XXX */
	asm volatile ("mtspr HID0,%0" :: "r"(hid0 | HID0_DLOCK));
#else
	asm volatile ("mtspr HID0,%0" :: "r"(hid0 | HID0_DLOCK | HID0_DCFI));
	delay(1000);
#endif

	/* Test cache RAM. */
	size = 0;
	for (i = 0; i < maxsize; i += CACHELINESIZE) {
		if (addr[i] == 0)
			size += CACHELINESIZE;
		addr[i] = 0xff;
	}
	asm volatile ("sync");

	/* Restore original state. */
	asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));
	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
	mtmsr(msr);

	/*
	 * Initialize L2 cache.
	 */
	l2cr &= ~L2CR_L2SIZ;

	switch (size) {
	case 256*1024:
		l2cr |= L2SIZ_256K; break;
	case 512*1024:
		l2cr |= L2SIZ_512K; break;
	case 1024*1024:
		l2cr |= L2SIZ_1M; break;
	case 2048*1024:
		l2cr |= L2SIZ_2M; break;
	default:
		size = 0;
		l2cr |= L2SIZ_256K;
	}

	/* Invalidate the entire L2 cache. */
	asm volatile ("mtspr L2CR,%0" :: "r"(l2cr | L2CR_L2I));

	/* Wait for the invalidation done. */
	do {
		asm volatile ("mfspr %0,L2CR" : "=r"(x));
	} while (x & L2CR_L2IP);

	if (size != 0) {
		l2cr |= L2CR_L2E;

		/* Enable L2 cache. */
		asm volatile ("mtspr L2CR,%0" :: "r"(l2cr));
	}
}
