/*	$Id: locore_subr.S,v 1.21 2003/10/11 18:47:32 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * NOTICE: This is not a standalone file.  to use it, #include it in
 * your port's locore.S, like so:
 *
 *	#include <powerpc/powerpc/locore_subr.S>
 */

	.data
GLOBAL(powersave)
	.long	0

	.text
/*
 * No processes are runnable, so loop waiting for one.
 * Separate label here for accounting purposes.
 * When we get here, interrupts are off (MSR[EE]=0) and sched_lock is held.
 */
ASENTRY(idle)
#if defined(MULTIPROCESSOR)
	/* Switch to the idle PCB unless we're already running on it. */
	cmpwi	30,0			/* old process was exiting? */
	beq	1f

	mfsr	10,USER_SR		/* save USER_SR for copyin/copyout */
	mfcr	11			/* save cr */
	mr	12,2			/* save r2 */
	stwu	1,-SFRAMELEN(1)		/* still running on old stack */
	stmw	10,8(1)
	lwz	3,L_ADDR(30)
	stw	1,PCB_SP(3)		/* save SP */

	GET_CPUINFO(7)
	lwz	6,CI_IDLE_PCB(7)
	addi	1,6,USPACE-16		/* 16 bytes are reserved at stack top */
	li	30,0			/* we are now on idle stack */
1:
#endif

#if defined(MULTIPROCESSOR) || defined(LOCKDEBUG)
	bl	_C_LABEL(sched_unlock_idle)
#endif

	mfmsr	31
	ori	31,31,PSL_EE@l		/* reenable ints again */
	mtmsr	31
	isync

#if defined(MULTIPROCESSOR)
	li	3,0
	bl	_C_LABEL(lcsplx)	/* splx(0) */
#endif

idle_loop:
/* Check if we can use power saving mode */
	lis	8,_C_LABEL(powersave)@ha
	lwz	9,_C_LABEL(powersave)@l(8)

	cmpwi	9,0
	beq	1f

	sync
	oris	31,31,PSL_POW@h		/* enter power saving mode */
	mtmsr	31
	isync
1:
	andi.	31,31,~PSL_EE@l		/* disable interrupts while
					   manipulating runque */
	mtmsr	31
	isync

#if defined(MULTIPROCESSOR) || defined(LOCKDEBUG)
	bl	_C_LABEL(sched_lock_idle)
#endif

	lis	8,_C_LABEL(sched_whichqs)@ha
	lwz	9,_C_LABEL(sched_whichqs)@l(8)

	cmpwi	9,0
	bne+	.Lsw1			/* at least one queue non-empty */

#if defined(MULTIPROCESSOR) || defined(LOCKDEBUG)
	bl	_C_LABEL(sched_unlock_idle)
#endif

	ori	31,31,PSL_EE@l		/* reenable ints again */
	mtmsr	31
	isync
	b	idle_loop

/*
 * switchexit(struct lwp *l, void (*exit)(struct lwp *))
 * switchexit gets called from cpu_exit to complete the exit procedure.
 */
ENTRY(switchexit)
	mtctr	4
/* First switch to the idle pcb/kernel stack */
#if defined(MULTIPROCESSOR)
	GET_CPUINFO(7)
	lwz	6,CI_IDLE_PCB(7)
	stw	6,CI_CURPCB(7)
#else
	lis	6,idle_u@ha
	lwz	6,idle_u@l(6)
	lis	7,_C_LABEL(curpcb)@ha
	stw	6,_C_LABEL(curpcb)@l(7)
#endif
	addi	1,6,USPACE-16		/* 16 bytes are reserved at stack top */
	/*
	 * Schedule the vmspace and stack to be freed (the proc arg is
	 * already in r3).
	 */
	bctrl

#if defined(MULTIPROCESSOR) || defined(LOCKDEBUG)
	bl	_C_LABEL(sched_lock_idle)
#endif

/* Fall through to cpu_switch to actually select another proc */
	li	3,0			/* indicate exited process */

/*
 * void cpu_switch(struct lwp *l)
 * Find a runnable process and switch to it.
 */
/* XXX noprofile?  --thorpej@netbsd.org */
ENTRY(cpu_switch)
	mflr	0			/* save lr */
	stw	0,4(1)
	stwu	1,-16(1)
	stw	31,12(1)
	stw	30,8(1)

	mr	30,3

#if defined(MULTIPROCESSOR)
	GET_CPUINFO(7)
	li	0,0
	stw	0,CI_CURLWP(7)		/* Zero to not accumulate cpu time */
	lwz	31,CI_CURPCB(7)

	lwz	3,CI_CPL(7)
	stw	3,PCB_SPL(31)		/* save spl */
#else
	li	0,0
	lis	3,_C_LABEL(curlwp)@ha
	stw	0,_C_LABEL(curlwp)@l(3)	/* Zero to not accumulate cpu time */
	lis	3,_C_LABEL(curpcb)@ha
	lwz	31,_C_LABEL(curpcb)@l(3)

#ifdef LOCKDEBUG
	lis	4,_C_LABEL(cpl)@ha
	lwz	3,_C_LABEL(cpl)@l(4)
	stw	3,PCB_SPL(31)		/* save spl */
#endif
#endif

#if !(defined(MULTIPROCESSOR) || defined(LOCKDEBUG))
	li	3,0
	bl	_C_LABEL(lcsplx)
	stw	3,PCB_SPL(31)		/* save spl */
#endif

/* Find a new process */
	mfmsr	3
	andi.	3,3,~PSL_EE@l		/* disable interrupts while
					   manipulating runque */
	mtmsr	3
	isync

	lis	8,_C_LABEL(sched_whichqs)@ha
	lwz	9,_C_LABEL(sched_whichqs)@l(8)

	cmpwi	9,0
	beq-	_ASM_LABEL(idle)	/* all queues empty */
.Lsw1:
	cntlzw	10,9
	lis	4,_C_LABEL(sched_qs)@ha
	addi	4,4,_C_LABEL(sched_qs)@l
	slwi	3,10,3
	add	3,3,4			/* select queue */

	lwz	31,L_FORW(3)		/* unlink first proc from queue */
	lwz	4,L_FORW(31)
	stw	4,L_FORW(3)
	stw	3,L_BACK(4)

	cmpw	3,4			/* queue empty? */
	bne	switch_common

	lis	3,0x80000000@h
	srw	3,3,10
	andc	9,9,3
	stw	9,_C_LABEL(sched_whichqs)@l(8) /* mark it empty */

switch_common:
	/* just did this resched thing */
	li	0,0
#if defined(MULTIPROCESSOR)
	GET_CPUINFO(4)
	stw	0,CI_WANT_RESCHED(4)
#else
	lis	4,_C_LABEL(want_resched)@ha
	stw	0,_C_LABEL(want_resched)@l(4)
#endif

	stw	0,L_BACK(31)		/* probably superfluous */

#if defined(MULTIPROCESSOR)
	GET_CPUINFO(4)
	stw	4,L_CPU(31)		/* l->l_cpu = curcpu() */
#endif

	/* Process now running on a processor. */
	li	3,LSONPROC		/* l->l_stat = LSONPROC */
	stw	3,L_STAT(31)

	/* record new process */
#if defined(MULTIPROCESSOR)
	stw	31,CI_CURLWP(4)
#else
	lis	4,_C_LABEL(curlwp)@ha
	stw	31,_C_LABEL(curlwp)@l(4)
#endif

	cmpw	31,30			/* is it the same lwp? */
	beq	switch_return

	cmpwi	30,0			/* old lwp was exiting? */
	beq	switch_exited

	mfsr	10,USER_SR		/* save USER_SR for copyin/copyout */
	mfcr	11			/* save cr */
	mr	12,2			/* save r2 */
	stwu	1,-SFRAMELEN(1)		/* still running on old stack */
	stmw	10,8(1)
	lwz	3,L_ADDR(30)
	stw	1,PCB_SP(3)		/* save SP */

switch_exited:
#if defined(MULTIPROCESSOR) || defined(LOCKDEBUG)
	bl	_C_LABEL(sched_unlock_idle)
#endif

	/* indicate new pcb */
	lwz	4,L_ADDR(31)
#if defined(MULTIPROCESSOR)
	GET_CPUINFO(6)
	stw	4,CI_CURPCB(6)
#else
	lis	5,_C_LABEL(curpcb)@ha
	stw	4,_C_LABEL(curpcb)@l(5)
#endif

	/* save real pmap pointer for spill fill */
	lwz	5,PCB_PMR(4)
#if defined(MULTIPROCESSOR)
	stwu	5,CI_CURPM(6)
#else
	lis	6,_C_LABEL(curpm)@ha
	stwu	5,_C_LABEL(curpm)@l(6)
#endif
	stwcx.	5,0,6			/* clear possible reservation */

	addic	5,5,64			/* &pm_sr[16] */
	li	6,0
	mfsr	8,KERNEL_SR		/* save kernel SR */
1:
	addis	6,6,-0x10000000@ha	/* set new procs segment registers */
	cmpwi	6,0			/* This is done from the real
					   address pmap */
	lwzu	7,-4(5)			/* so we don't have to worry */
	mtsrin	7,6			/* about accessibility */
	bne	1b
	mtsr	KERNEL_SR,8		/* restore kernel SR */
	isync

	lwz	1,PCB_SP(4)		/* get new procs SP */

	mfmsr	3
	ori	3,3,PSL_EE@l		/* interrupts are okay again */
	mtmsr	3

	lmw	10,8(1)			/* get other regs */
	lwz	1,0(1)			/* get saved SP */
	mr	2,12			/* get saved r2 */
	mtcr	11			/* get saved cr */
	isync
	mtsr	USER_SR,10		/* get saved USER_SR */
	isync

	li	30,1			/* switched lwps */
	b	1f

switch_return:
	li	30,0			/* same lwp; return 0 */
#if defined(MULTIPROCESSOR) || defined(LOCKDEBUG)
	bl	_C_LABEL(sched_unlock_idle)
#endif

	mfmsr	3
	ori	3,3,PSL_EE@l		/* Now we can interrupt again */
	mtmsr	3

	lwz	4,L_ADDR(31)
1:
	lwz	3,PCB_SPL(4)
	bl	_C_LABEL(lcsplx)

	mr	3,30			/* return value */
	lwz	31,12(1)
	lwz	30,8(1)
	addi	1,1,16
	lwz	0,4(1)
	mtlr	0
	blr

/*
 * void
 * cpu_switchto(struct lwp *current, struct lwp *new)
 * Variant of cpu_switch, but instead of picking a new LWP from the run
 * queue, switch to the indicated new LWP.
 * 	r3 - current LWP
 * 	r4 - LWP to switch to
 */
ENTRY(cpu_switchto)
	mflr	0			/* save lr */
	stw	0,4(1)
	stwu	1,-16(1)
	stw	31,12(1)
	stw	30,8(1)

	mr	30,3			/* r30 = curlwp */
	mr	31,4			/* r31 = newlwp */

#if defined(MULTIPROCESSOR)
	GET_CPUINFO(7)
	li	0,0
	stw	0,CI_CURLWP(7)		/* Zero to not accumulate cpu time */
	lwz	5,CI_CURPCB(7)

	lwz	3,CI_CPL(7)
	stw	3,PCB_SPL(5)		/* save spl */
#else
	li	0,0
	lis	3,_C_LABEL(curlwp)@ha
	stw	0,_C_LABEL(curlwp)@l(3)	/* Zero to not accumulate cpu time */
	lis	3,_C_LABEL(curpcb)@ha
	lwz	5,_C_LABEL(curpcb)@l(3)

	lis	4,_C_LABEL(cpl)@ha
	lwz	3,_C_LABEL(cpl)@l(4)
	stw	3,PCB_SPL(5)		/* save spl */
#endif

#if !(defined(MULTIPROCESSOR) || defined(LOCKDEBUG))
	li	3,0
	bl	_C_LABEL(lcsplx)
#endif

	mfmsr	3
	andi.	3,3,~PSL_EE@l		/* disable interrupts while
					   manipulating runque */
	mtmsr	3
	isync

	b	switch_common

/*
 * Child comes here at the end of a fork.
 * Return to userspace via the trap return path.
 */
	.globl	_C_LABEL(fork_trampoline)
_C_LABEL(fork_trampoline):
#if defined(MULTIPROCESSOR)
	bl	_C_LABEL(proc_trampoline_mp)
#endif
	li	3,0
	bl	_C_LABEL(lcsplx)
	mtlr	31
	mr	3,30
	blrl				/* jump indirect to r31 */
	b	trapexit

/*
 * int setfault()
 *
 * Similar to setjmp to setup for handling faults on accesses to user memory.
 * Any routine using this may only call bcopy, either the form below,
 * or the (currently used) C code optimized, so it doesn't use any non-volatile
 * registers.
 */
ENTRY(setfault)
	mflr	0
	mfcr	12
#if defined(MULTIPROCESSOR)
	GET_CPUINFO(4)
	lwz	4,CI_CURPCB(4)
#else
	lis	4,_C_LABEL(curpcb)@ha
	lwz	4,_C_LABEL(curpcb)@l(4)
#endif
	stw	3,PCB_FAULT(4)
	stw	0,0(3)
	stw	1,4(3)
	stw	2,8(3)
	stmw	12,12(3)
	sync
	isync
	li	3,0
	blr

/*
 * The following code gets copied to the top of the user stack on process
 * execution.  It does signal trampolining on signal delivery.
 *
 * On entry r1 points to a struct sigcontext at bottom of current stack.
 * All other registers are unchanged.
 */
	.globl	_C_LABEL(sigcode),_C_LABEL(esigcode)
_C_LABEL(sigcode):
	addi	1,1,-16			/* reserved space for callee */
	mr	31,5			/* save scp */
	blrl
	mr	3,31			/* restore scp */
	li	0,SYS_compat_16___sigreturn14
	sc				/* sigreturn(scp) */
	li	0,SYS_exit
	sc				/* exit(errno) */
_C_LABEL(esigcode):
