/*	$Id: syscall.c,v 1.13 2003/11/01 08:17:03 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_altivec.h"
#include "opt_compat_linux.h"
#include "opt_ktrace.h"
#include "opt_multiprocessor.h"
#include "opt_systrace.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/ktrace.h>
#include <sys/proc.h>
#include <sys/sa.h>
#include <sys/savar.h>
#include <sys/syscall.h>
#include <sys/systrace.h>
#include <sys/user.h>

#include <uvm/uvm_extern.h>

void syscall_intern(struct proc *);
void syscall_plain(struct trapframe *);
void syscall_fancy(struct trapframe *);

/* These definitions should probably be somewhere else			XXX */
#define FIRSTARG	3		/* first argument is in reg 3 */
#define NARGREG		8		/* 8 args are in registers */
#define MOREARGS(sp)	((caddr_t)((int)(sp) + 8)) /* more args go here */

void
syscall_intern(p)
	struct proc *p;
{

#ifdef KTRACE
	if (p->p_traceflag & (KTRFAC_SYSCALL | KTRFAC_SYSRET)) {
		p->p_md.md_syscall = syscall_fancy;
		return;
	}
#endif
#ifdef SYSTRACE
	if (ISSET(p->p_flag, P_SYSTRACE)) {
		p->p_md.md_syscall = syscall_fancy;
		return;
	}
#endif
	p->p_md.md_syscall = syscall_plain;
}

void
syscall_plain(frame)
	struct trapframe *frame;
{
	struct lwp *l = curlwp;
	struct proc *p = l->l_proc;
	const struct sysent *callp;
	size_t argsize;
	register_t code;
	register_t *params, rval[2];
	int n, error;
	register_t args[10];

	KERNEL_PROC_LOCK(l);

	uvmexp.syscalls++;

	code = frame->fixreg[0];
	callp = p->p_emul->e_sysent;
	params = frame->fixreg + FIRSTARG;
	n = NARGREG;

	switch (code) {
	case SYS_syscall:
		/*
		 * code is first argument,
		 * followed by actual args.
		 */
		code = *params++;
		n -= 1;
		break;
	case SYS___syscall:
		params++;
		code = *params++;
		n -= 2;
		break;
	default:
		break;
	}

	code &= (SYS_NSYSENT - 1);
	callp += code;
	argsize = callp->sy_argsize;

	if (argsize > n * sizeof(register_t)) {
		memcpy(args, params, n * sizeof(register_t));
		error = copyin(MOREARGS(frame->fixreg[1]),
			       args + n,
			       argsize - n * sizeof(register_t));
		if (error)
			goto syscall_bad;
		params = args;
	}

	rval[0] = 0;
	rval[1] = 0;

	error = (*callp->sy_call)(l, params, rval);
	switch (error) {
	case 0:
		frame->fixreg[FIRSTARG] = rval[0];
		frame->fixreg[FIRSTARG + 1] = rval[1];
		frame->cr &= ~0x10000000;
		break;
	case ERESTART:
		/*
		 * Set user's pc back to redo the system call.
		 */
		frame->srr0 -= 4;
		break;
	case EJUSTRETURN:
		/* nothing to do */
		break;
	default:
syscall_bad:
		if (p->p_emul->e_errno)
			error = p->p_emul->e_errno[error];
		frame->fixreg[FIRSTARG] = error;
		frame->cr |= 0x10000000;
		break;
	}

	KERNEL_PROC_UNLOCK(l);

	userret(l, frame);
}

void
syscall_fancy(frame)
	struct trapframe *frame;
{
	struct lwp *l = curlwp;
	struct proc *p = l->l_proc;
	const struct sysent *callp;
	size_t argsize;
	register_t code;
	register_t *params, rval[2];
	int n, error;
	register_t args[10];

	KERNEL_PROC_LOCK(l);

	uvmexp.syscalls++;

	code = frame->fixreg[0];
	callp = p->p_emul->e_sysent;
	params = frame->fixreg + FIRSTARG;
	n = NARGREG;

	switch (code) {
	case SYS_syscall:
		/*
		 * code is first argument,
		 * followed by actual args.
		 */
		code = *params++;
		n -= 1;
		break;
	case SYS___syscall:
		params++;
		code = *params++;
		n -= 2;
		break;
	default:
		break;
	}

	code &= (SYS_NSYSENT - 1);
	callp += code;
	argsize = callp->sy_argsize;

	if (argsize > n * sizeof(register_t)) {
		memcpy(args, params, n * sizeof(register_t));
		error = copyin(MOREARGS(frame->fixreg[1]),
			       args + n,
			       argsize - n * sizeof(register_t));
		if (error)
			goto syscall_bad;
		params = args;
	}

	if ((error = trace_enter(l, code, code, NULL, params)) != 0)
		goto syscall_bad;

	rval[0] = 0;
	rval[1] = 0;

	error = (*callp->sy_call)(l, params, rval);
	switch (error) {
	case 0:
		frame->fixreg[FIRSTARG] = rval[0];
		frame->fixreg[FIRSTARG + 1] = rval[1];
		frame->cr &= ~0x10000000;
		break;
	case ERESTART:
		/*
		 * Set user's pc back to redo the system call.
		 */
		frame->srr0 -= 4;
		break;
	case EJUSTRETURN:
		/* nothing to do */
		break;
	default:
syscall_bad:
		if (p->p_emul->e_errno)
			error = p->p_emul->e_errno[error];
		frame->fixreg[FIRSTARG] = error;
		frame->cr |= 0x10000000;
		break;
	}

	trace_exit(l, code, params, rval, error);

	KERNEL_PROC_UNLOCK(l);

	userret(l, frame);
}

void
child_return(arg)
	void *arg;
{
	struct lwp *l = arg;
	struct trapframe *tf = trapframe(l);

	KERNEL_PROC_UNLOCK(l);

	tf->fixreg[FIRSTARG] = 0;
	tf->fixreg[FIRSTARG + 1] = 1;
	tf->cr &= ~0x10000000;
	tf->srr1 &= ~(PSL_FP|PSL_VEC);	/* Disable FP & AltiVec, as we can't
					   be them. */
	l->l_addr->u_pcb.pcb_fpcpu = NULL;
#ifdef KTRACE
	if (KTRPOINT(l->l_proc, KTR_SYSRET)) {
		KERNEL_PROC_LOCK(l);
		ktrsysret(l->l_proc, SYS_fork, 0, 0);
		KERNEL_PROC_UNLOCK(l);
	}
#endif
	/* Profiling?							XXX */
	curcpu()->ci_schedstate.spc_curpriority = l->l_priority;
}

#ifdef COMPAT_LINUX
void linux_syscall_intern(struct proc *p);

void
linux_syscall_intern(p)
	struct proc *p;
{

	syscall_intern(p);
}
#endif
