/*	$Id: hfs.h,v 1.21 2004/01/08 11:26:58 tsubai Exp $	*/

/*-
 * Copyright (c) 2004 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

struct hfs_args {
	char *fspec;		/* blocks special holding the fs to mount */
	struct export_args export; /* network export information */
	uid_t uid;		/* uid that owns hfs files */
	gid_t gid;		/* gid that owns hfs files */
	mode_t mask;		/* mask to be applied for hfs perms */
};

#ifdef _KERNEL
#include <miscfs/genfs/genfs_node.h>

struct hfsnode {
	struct genfs_node gnode;
	time_t ctime;
	time_t mtime;
	u_int16_t abnum[3];
	u_int16_t ablen[3];
	struct hfsmount *hmp;	/* our mount struct */
	struct vnode *vp;	/* our vnode */
	u_int32_t fsize;	/* logical file size in bytes */
	u_int32_t psize;	/* physical file size in bytes */
	u_int32_t cnid;		/* catalog node ID */
	u_int32_t parent;	/* parent CNID */
	uid_t uid;
	gid_t gid;
	int flags;
};
#define VTOH(vp) ((struct hfsnode *)(vp)->v_data)

/*
 * mount data
 */
struct hfsmount {
	struct mount *mp;	/* mount struct for this fs */
	u_int16_t magic;
	u_long bsize;		/* size of blocks */
	uid_t uid;		/* uid of mounting user */
	gid_t gid;		/* gid of mounting user */
	u_long mask;		/* mode mask */
	struct vnode *devvp;	/* block device mounted on */
	struct vnode *rootvp;	/* out root vnode */
	struct netexport export; /* export information */
	u_long *bitmap;		/* allocation bitmap */
	u_long numblks;		/* number of usable blocks */
	u_long freeblks;	/* number of free blocks */

	u_int ablkoff;		/* allocation block offset */
	u_int ablksize;		/* allocation block size */
	u_int32_t flnode;	/* first leaf node */
	u_int catsize;		/* catalog file size */
	u_int16_t catABN[3];	/* catalog file extent */
	u_int16_t catABlen[3];
};

#define VFSTOHFS(mp) ((struct hfsmount *)(mp)->mnt_data)

struct buf *read_catalog(struct hfsmount *, u_int);
int getpstr(u_char *, u_char *, int);
void hfs_dumpnode(struct buf *);
void hfs_dumprecord(char *);
int hfs_lookup __P((void *));

extern struct pool hfs_node_pool;
extern int (**hfs_vnodeop_p) __P((void *));

static u_int GET8(void *, int);
static u_int GET16(void *, int);
static u_int GET32(void *, int);
static u_int GET8B(struct buf *, int);
static u_int GET16B(struct buf *, int);
static u_int GET32B(struct buf *, int);

/* XXX byte order, alignment */
static inline u_int
GET8(base, off)
	void *base;
	int off;
{
	u_int8_t *p = (void *)((char *)base + off);
	return *p;
}

static inline u_int
GET16(base, off)
	void *base;
	int off;
{
	u_int16_t *p = (void *)((char *)base + off);
	return *p;
}

static inline u_int
GET32(base, off)
	void *base;
	int off;
{
	u_int32_t *p = (void *)((char *)base + off);
	return *p;
}

static inline u_int
GET8B(bp, off)
	struct buf *bp;
	int off;
{
	u_int8_t *p = (void *)(bp->b_data + off);
	return *p;
}

static inline u_int
GET16B(bp, off)
	struct buf *bp;
	int off;
{
	u_int16_t *p = (void *)(bp->b_data + off);
	return *p;
}

static inline u_int
GET32B(bp, off)
	struct buf *bp;
	int off;
{
	u_int32_t *p = (void *)(bp->b_data + off);
	return *p;
}

#endif /* _KERNEL */

/* MDB offsets */
#define drSigWord  0x00
#define drCrDate   0x02
#define drLsMod    0x06
#define drAtrb     0x0a
#define drNmFls    0x0c
#define drVBMSt    0x0e
#define drAllocPtr 0x10
#define drNmAlBlks 0x12
#define drAlBlkSiz 0x14
#define drClpSiz   0x18
#define drAlBlSt   0x1c
#define drNxtCNID  0x1e
#define drFreeBks  0x22
#define drVN       0x24
#define drVolBkUp  0x40
#define drVSeqNum  0x44
#define drWrCnt    0x46
#define drXTClpSiz 0x4a
#define drCTClpSiz 0x4e
#define drNmRtDirs 0x52
#define drFilCnt   0x54
#define drDirCnt   0x58
#define drFndrInfo 0x5c
#define drVCSize   0x7c
#define drVBMCSize 0x7e
#define drCtlCSize 0x80
#define drXTFlSize 0x82
#define drXTExtRec 0x86
#define drCTFlSize 0x92
#define drCTExtRec 0x96

/*
 * Node descriptor offsets
 */
#define ndFLink     0x00	/* Forward link */
#define ndBLink     0x04	/* Backward link */
#define ndType      0x08	/* Node type */
#define ndNHeight   0x09	/* Node height (depth) */
#define ndNRecs     0x0a	/* Number of records in this node */

/*
 * B*-tree header offsets
 */
#define bthDepth    0x00
#define bthRoot     0x02	/* Root node number */
#define bthNRecs    0x06
#define bthFNode    0x0a	/* First leaf node */
#define bthLNode    0x0e	/* Last leaf node */
#define bthNodeSize 0x12	/* Node size (always 512) */
#define bthKeyLen   0x14
#define bthNNodes   0x16
#define bthFree     0x1a

/*
 * Catalog key offsets
 */
#define ckrKeyLen   0x00	/* Key length - 1 */
#define ckrParID    0x02	/* Parent directory CNID */
#define ckrCName    0x06	/* name[32] */

/*
 * Catalog data offsets
 */
#define cdrType     0x00
#define dirFlags    0x02
#define dirVal      0x04
#define dirDirID    0x06
#define dirCrDat    0x0a
#define dirMdDat    0x0e
#define dirBkDat    0x12
#define dirUsrInfo  0x16
#define dirFndrInfo 0x26
#define filFlags    0x02
#define filTyp      0x03
#define filUsrWds   0x04
#define filFlNum    0x14
#define filStBlk    0x18
#define filLgLen    0x1a
#define filPyLen    0x1e
#define filRStBlk   0x22
#define filRLgLen   0x24
#define filRPyLen   0x28
#define filCrDat    0x2c
#define filMdDat    0x30
#define filBkDat    0x34
#define filFndrInfo 0x38
#define filClpSize  0x48
#define filExtRec   0x4a
#define filRExtRec  0x56
#define thdParID    0x0a
#define thdCName    0x0e
#define fthdParID   0x0a
#define fthdCName   0x0e

#define HFS_MAGIC 0x4244
#define NODESIZE 512

/* XXX for now */
#define MOUNT_HFS "hfs"
#define VT_HFS 21
