/*	$Id: hfs_util.c,v 1.8 2004/01/08 11:08:25 tsubai Exp $	*/

/*-
 * Copyright (c) 2004 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/buf.h>
#include <sys/vnode.h>
#include <fs/hfs/hfs.h>

/*
 * Pascal string  C string Ѵ
 * dest  len Хȳݤ뤳
 * XXX NUL  / ޤǤɤ?
 */
int
getpstr(pstr, dest, len)
	u_char *pstr, *dest;
	int len;
{
	bzero(dest, len);
	if (*pstr >= len)
		return -1;
	bcopy(pstr + 1, dest, *pstr);
	return 0;
}

/*
 * catalog ɤࡣnode ϥΡֹ档(!= CNID XXX)
 * Ρɤ512ХȤθĹ
 */
struct buf *
read_catalog(hmp, node)
	struct hfsmount *hmp;
	u_int node;
{
	int i;

//printf("read_catalog (node = 0x%x)\n", node);
	if (node >= hmp->catsize / 512)
		return NULL;

	/* XXX extent overflow file */

	for (i = 0; i < 3; i++) {
		if (hmp->catABN[i] == 0)
			break;
		if (node < hmp->catABlen[i] * (hmp->ablksize / NODESIZE)) {
			daddr_t pos = (daddr_t)hmp->catABN[i] * hmp->ablksize;
			struct buf *bp;
			int error;

			pos += (daddr_t)node * NODESIZE;
			pos += hmp->ablkoff * NODESIZE;
			error = bread(hmp->devvp, pos / DEV_BSIZE, hmp->bsize,
			    NOCRED, &bp);
			if (error) {
				brelse(bp);
				return NULL;
			}
#if 0
printf("ndFLink = 0x%x, ", GET32B(bp, ndFLink));
printf("ndBLink = 0x%x, ", GET32B(bp, ndBLink));
printf("ndType = %d, ", GET8B(bp, ndType));
printf("ndNHeight = %d, ", GET8B(bp, ndNHeight));
printf("ndNRecs = %d\n", GET16B(bp, ndNRecs));
#endif

			return bp;
		}
		node -= hmp->catABlen[i] * (hmp->ablksize / NODESIZE);
	}

	return NULL;
}

void
hfs_dumpnode(bp)
	struct buf *bp;
{
	u_int nrec, recoff;
	int i;

	nrec = GET16B(bp, ndNRecs);
	printf("%d record(s)\n", nrec);

	for (i = 0; i < nrec; i++) {
		recoff = GET16B(bp, NODESIZE - 2 - i * 2);
		hfs_dumprecord(bp->b_data + recoff);
	}
}

void
hfs_dumprecord(rec)
	char *rec;
{
	u_int len, parent, type, id;
	char *data;
	char name[32];

	len = GET8(rec, ckrKeyLen);
	parent = GET32(rec, ckrParID);
	getpstr(rec + ckrCName, name, sizeof name);
	printf("Catalog key: parent = 0x%x, name = [%s]\n", parent, name);

//	dataoff = (recoff + len + 2) & ~1;
	len++;
	len++;
	len &= ~1;
	data = rec + len;

	type = GET8(data, cdrType);
	printf("Catalog data: cdrType = %d ", type);

	switch (type) {

	case 1: /* dir */
		printf("dir\n    ");
		id = GET32(data, dirDirID);
		printf("dirDirID = 0x%x, ", id);
		printf("dirCrDat = 0x%x, ", GET32(data, dirCrDat));
		printf("dirMdDat = 0x%x\n", GET32(data, dirMdDat));
		break;

	case 2: /* file */
		printf("file\n    ");
		id = GET32(data, filFlNum);
		printf("filFlNum = 0x%x, ", id);
		printf("filStBlk = 0x%x, ", GET16(data, filStBlk));
		printf("filCrDat = 0x%x, ", GET32(data, filCrDat));
		printf("filMdDat = 0x%x\n", GET32(data, filMdDat));
		break;

	case 3: /* directory thread */
		printf("dir thread\n");
		break;

	case 4: /* file thread */
		printf("file thread\n");
		break;

	default:
		printf("(?)\n");
	}
}
