/*	$Id: hfs_vnops.c,v 1.23 2004/01/08 11:08:25 tsubai Exp $	*/

/*-
 * Copyright (c) 2004 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/buf.h>
#include <sys/dirent.h>
#include <sys/malloc.h>
#include <sys/namei.h>
#include <sys/stat.h>
#include <sys/unistd.h>
#include <sys/vnode.h>
#include <miscfs/genfs/genfs.h>
#include <miscfs/specfs/specdev.h>
#include <fs/hfs/hfs.h>

int hfs_access __P((void *));
int hfs_getattr __P((void *));
int hfs_read __P((void *));
int hfs_write __P((void *));
int hfs_link __P((void *));
int hfs_symlink __P((void *));
int hfs_readdir __P((void *));
int hfs_readlink __P((void *));
int hfs_inactive __P((void *));
int hfs_reclaim __P((void *));
int hfs_bmap __P((void *));
int hfs_strategy __P((void *));
int hfs_print __P((void *));
int hfs_pathconf __P((void *));
int hfs_advlock __P((void *));
int hfs_reallocblks __P((void *));

int (**hfs_vnodeop_p) __P((void *));

const struct vnodeopv_entry_desc hfs_vnodeop_entries[] = {
	{ &vop_default_desc, vn_default_error },
	{ &vop_lookup_desc, hfs_lookup },		/* lookup */
	{ &vop_create_desc, genfs_eopnotsupp },		/* create */
	{ &vop_mknod_desc, genfs_eopnotsupp },		/* mknod */
	{ &vop_open_desc, genfs_nullop },		/* open */
	{ &vop_close_desc, genfs_nullop },		/* close */
	{ &vop_access_desc, hfs_access },		/* access */
	{ &vop_getattr_desc, hfs_getattr },		/* getattr */
	{ &vop_setattr_desc, genfs_eopnotsupp },	/* setattr */
	{ &vop_read_desc, hfs_read },			/* read */
	{ &vop_write_desc, hfs_write },			/* write */
	{ &vop_lease_desc, genfs_lease_check },		/* lease */
	{ &vop_ioctl_desc, genfs_enoioctl },		/* ioctl */
	{ &vop_fcntl_desc, genfs_fcntl },		/* fcntl */
	{ &vop_poll_desc, genfs_poll },			/* poll */
	{ &vop_kqfilter_desc, genfs_kqfilter },		/* kqfilter */
	{ &vop_revoke_desc, genfs_revoke },		/* revoke */
	{ &vop_mmap_desc, genfs_mmap },			/* mmap */
	{ &vop_fsync_desc, genfs_nullop },		/* fsync */
	{ &vop_seek_desc, genfs_seek },			/* seek */
	{ &vop_remove_desc, genfs_eopnotsupp },		/* remove */
	{ &vop_link_desc, hfs_link },			/* link */
	{ &vop_rename_desc, genfs_eopnotsupp },		/* rename */
	{ &vop_mkdir_desc, genfs_eopnotsupp },		/* mkdir */
	{ &vop_rmdir_desc, genfs_eopnotsupp },		/* rmdir */
	{ &vop_symlink_desc, hfs_symlink },		/* symlink */
	{ &vop_readdir_desc, hfs_readdir },		/* readdir */
	{ &vop_readlink_desc, hfs_readlink },		/* readlink */
	{ &vop_abortop_desc, genfs_abortop },		/* abortop */
	{ &vop_inactive_desc, hfs_inactive },		/* inactive */
	{ &vop_reclaim_desc, hfs_reclaim },		/* reclaim */
	{ &vop_lock_desc, genfs_lock },			/* lock */
	{ &vop_unlock_desc, genfs_unlock },		/* unlock */
	{ &vop_bmap_desc, hfs_bmap },			/* bmap */
	{ &vop_strategy_desc, hfs_strategy },		/* strategy */
	{ &vop_print_desc, hfs_print },			/* print */
	{ &vop_islocked_desc, genfs_islocked },		/* islocked */
	{ &vop_pathconf_desc, hfs_pathconf },		/* pathconf */
	{ &vop_advlock_desc, hfs_advlock },		/* advlock */
	{ &vop_blkatoff_desc, genfs_eopnotsupp },	/* blkatoff */
	{ &vop_valloc_desc, genfs_eopnotsupp },		/* valloc */
	{ &vop_reallocblks_desc, hfs_reallocblks },	/* reallocblks */
	{ &vop_vfree_desc, genfs_nullop },		/* vfree */
	{ &vop_truncate_desc, genfs_eopnotsupp },	/* truncate */
	{ &vop_update_desc, genfs_nullop },		/* update */
	{ &vop_bwrite_desc, genfs_eopnotsupp },		/* bwrite */
	{ &vop_getpages_desc, genfs_getpages },		/* getpages */
	{ &vop_putpages_desc, genfs_putpages },		/* putpages */
	{ NULL, NULL }
};

const struct vnodeopv_desc hfs_vnodeop_opv_desc =
	{ &hfs_vnodeop_p, hfs_vnodeop_entries };

int
hfs_access(v)
	void *v;
{
	struct vop_access_args /* {
		struct vnode *a_vp;
		int a_mode;
		struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap = v;
	struct vnode *vp = ap->a_vp;
	struct hfsnode *hp = VTOH(vp);
	mode_t mode = ap->a_mode;

	/*
	 * Disallow write attempts on read-only file systems;
	 * unless the file is a socket, fifo, or a block or
	 * character device resident on the file system.
	 */
	if (mode & VWRITE) {
		switch (vp->v_type) {
		case VDIR:
		case VLNK:
		case VREG:
			if (vp->v_mount->mnt_flag & MNT_RDONLY)
				return EROFS;
		default:
			break;
		}
	}

	return vaccess(ap->a_vp->v_type, 0755 /* XXX */ & hp->hmp->mask,
	    hp->uid, hp->gid, mode, ap->a_cred);
}

int
hfs_getattr(v)
	void *v;
{
	struct vop_getattr_args /* {
		struct vnode *a_vp;
		struct vattr *a_vap;
		struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap = v;
	struct vattr *vap = ap->a_vap;
	struct hfsnode *hp = VTOH(ap->a_vp);
	struct hfsmount *hmp = hp->hmp;

	vattr_null(vap);
	vap->va_uid = hp->uid;
	vap->va_gid = hp->gid;
	vap->va_fsid = ap->a_vp->v_rdev;
	vap->va_atime.tv_sec = vap->va_mtime.tv_sec = hp->mtime;
	vap->va_ctime.tv_sec = hp->ctime;	/* XXX va_birthtime? */
	vap->va_atime.tv_nsec = vap->va_mtime.tv_nsec =
	    vap->va_ctime.tv_nsec = 0;
	vap->va_gen = 0;
	vap->va_flags = 0;
	vap->va_rdev = 0;
	vap->va_fileid = hp->cnid;
	vap->va_type = ap->a_vp->v_type;
	/* vap->va_mode = S_IRUSR|S_IXUSR|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH; */
	/* vap->va_mode = S_IRWXU|S_IRWXG|S_IRWXO; */
	vap->va_mode = S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH;
	if (ap->a_vp->v_type == VDIR) {
		vap->va_nlink = 1;	/* XXX dirVal */
		vap->va_bytes = hmp->bsize;
		vap->va_size = hmp->bsize;
	} else {
		vap->va_nlink = 1;
		vap->va_bytes = hp->psize;
		vap->va_size = hp->fsize;
		vap->va_blocksize = hmp->ablksize;
	}
	return 0;
}

int
hfs_read(v)
	void *v;
{
	struct vop_read_args /* {
		struct vnode *a_vp;
		struct uio *a_uio;
		int a_ioflag;
		struct ucred *a_cred;
	} */ *ap = v;
	struct vnode *vp = ap->a_vp;
	struct hfsnode *hp = VTOH(ap->a_vp);
	struct uio *uio = ap->a_uio;
	vsize_t bytelen;
	void *win;
	int error = 0;

	if (vp->v_type != VREG)
		return EIO;

	if (uio->uio_resid == 0)
		return 0;
	if (uio->uio_offset < 0)
		return EINVAL;

	if (uio->uio_offset >= hp->fsize)
		return 0;

	while (uio->uio_resid > 0) {
		bytelen = MIN(hp->fsize - uio->uio_offset, uio->uio_resid);
		if (bytelen == 0)
			break;
		win = ubc_alloc(&vp->v_uobj, uio->uio_offset,
				&bytelen, UBC_READ);
		error = uiomove(win, bytelen, uio);
		ubc_release(win, 0);
		if (error)
			break;
	}

	return error;
}

int
hfs_write(v)
	void *v;
{
#if 0
	struct vop_write_args /* {
		struct vnode *a_vp;
		struct uio *a_uio;
		int a_ioflag;
		struct ucred *a_cred;
	} */ *ap = v;
#endif

	return EOPNOTSUPP;
}

int
hfs_link(v)
	void *v;
{
	struct vop_link_args /* {
		struct vnode *a_dvp;
		struct vnode *a_vp;
		struct componentname *a_cnp;
	} */ *ap = v;

	VOP_ABORTOP(ap->a_dvp, ap->a_cnp);
	vput(ap->a_dvp);
	return EOPNOTSUPP;
}

int
hfs_symlink(v)
	void *v;
{
	struct vop_symlink_args /* {
		struct vnode *a_dvp;
		struct vnode **a_vpp;
		struct componentname *a_cnp;
		struct vattr *a_vap;
		char *a_target;
	} */ *ap = v;

	VOP_ABORTOP(ap->a_dvp, ap->a_cnp);
	vput(ap->a_dvp);
	return EOPNOTSUPP;
}

struct hdirent {
	u_int32_t fileno;
	u_int16_t reclen;
	char type;
	char namlen;
	char name[32];
};

int
hfs_readdir(v)
	void *v;
{
	struct vop_readdir_args /* {
		struct vnode *a_vp;
		struct uio *a_uio;
		struct ucred *a_cred;
		int *a_eofflag;
		off_t **a_cookies;
		int *a_ncookies;
	} */ *ap = v;
	struct hfsnode *hp = VTOH(ap->a_vp);
	struct hfsmount *hmp = hp->hmp;
	struct uio *uio = ap->a_uio;
	int error, uavail;
	struct hdirent hdir;
	off_t uoff, *cookies = NULL;
	int ncookies = 0;
	struct buf *bp;
	int i;
	u_int node, cnid;

	if (ap->a_vp->v_type != VDIR) {
		error = ENOTDIR;
		goto reterr;
	}

	uoff = uio->uio_offset;
	if (uoff < 0) {
		error = EINVAL;
		goto reterr;
	}

	cnid = hp->cnid;
	error = 0;
	uavail = uio->uio_resid / sizeof hdir;

	/*
	 * if no slots available or offset requested is not on a slot boundry
	 */
	if (uavail < 1 || uoff % sizeof hdir) {
		error = EINVAL;
		goto reterr;
	}

	if (ap->a_ncookies) {
		ncookies = 0;
		cookies = malloc(sizeof(off_t) * uavail, M_TEMP, M_WAITOK);
		*ap->a_cookies = cookies;
	}

/* XXX XXX XXX FOR NOW XXX XXX XXX*/
if (uoff != 0) goto reterr;

	for (i = 0; i < 2; i++) {
		static char dot[2][2] = { ".", ".." };

		bzero(&hdir, sizeof hdir);
		bcopy(dot[i], hdir.name, 2);
		hdir.fileno = 2;
		hdir.type = DT_DIR;
		hdir.reclen = sizeof hdir;
		hdir.namlen = strlen(hdir.name);

		error = uiomove(&hdir, sizeof hdir, uio);
		if (error)
			goto reterr;
		if (ap->a_ncookies) {
			*cookies++ = uoff;
			ncookies++;
		}
		uoff += sizeof hdir;
		uavail--;
		if (uavail == 0)
			goto reterr;
	}

	node = hmp->flnode;	/* ǽ leaf node */
	while (node) {
		int len, type, id;
		u_int nrec, recoff, dataoff;

		bp = read_catalog(hmp, node);
		if (bp == NULL) {
			error = EINVAL;
			goto reterr;
		}
		nrec = GET16B(bp, ndNRecs);

		for (i = 0; i < nrec; i++) {
			recoff = GET16B(bp, 510 - i * 2);	/* XXX HFS */
			len = GET8B(bp, recoff + ckrKeyLen);
			/* XXX recoff, len ǧȤ */

			/* Catalog key ˿Ƥ̾ */

			/* Ƥ㤦ʤ鼡 */
			if (GET32B(bp, recoff + ckrParID) != cnid)
				continue;

			len++;
			len++;
			len &= ~1;
			dataoff = recoff + len;

			/* Catalog data */
			type = GET8B(bp, dataoff + cdrType);
			switch (type) {

			case 1: /* dir */
				id = GET32B(bp, dataoff + dirDirID);
				type = DT_DIR;
				break;

			case 2: /* file */
				id = GET32B(bp, dataoff + filFlNum);
				type = DT_REG;
				break;

			default:
				continue;
			}

			bzero(&hdir, sizeof hdir);
			if (getpstr(bp->b_data + recoff + ckrCName, hdir.name,
			    sizeof hdir.name))
				continue;	/* XXX ɤ? */
			hdir.fileno = id;
			hdir.type = type;
			hdir.reclen = sizeof hdir;
			hdir.namlen = strlen(hdir.name);

			error = uiomove(&hdir, sizeof hdir, uio);
			if (error) {
				brelse(bp);
				goto reterr;
			}
			if (ap->a_ncookies) {
				*cookies++ = uoff;
				ncookies++;
			}
			uoff += sizeof hdir;
			uavail--;
			if (uavail == 0) {
				brelse(bp);
				goto reterr;
			}
		}
		node = GET32B(bp, ndFLink);
		brelse(bp);
	}

	/* ޤǤǤ٤Ƥ record  */
	*ap->a_eofflag = 1;

reterr:
	if (ap->a_ncookies) {
		if (error) {
			free(*ap->a_cookies, M_TEMP);
			*ap->a_ncookies = 0;
			*ap->a_cookies = NULL;
		} else
			*ap->a_ncookies = ncookies;
	}
	return error;
}

int
hfs_readlink(v)
	void *v;
{
#if 0
	struct vop_readlink_args /* {
		struct vnode *a_vp;
		struct uio *a_uio;
		struct ucred *a_cred;
	} */ *ap = v;
#endif

	return EINVAL;
}

int
hfs_inactive(v)
	void *v;
{
	struct vop_inactive_args /* {
		struct vnode *a_vp;
		struct proc *a_p;
	} */ *ap = v;
	struct vnode *vp = ap->a_vp;
	struct proc *p = ap->a_p;

	VOP_UNLOCK(vp, 0);
	/* XXX this needs to check if file was deleted */
	vrecycle(vp, NULL, p);

	return 0;
}

int
hfs_reclaim(v)
	void *v;
{
	struct vop_reclaim_args /* {
		struct vnode *a_vp;
		struct proc *a_p;
	} */ *ap = v;
	struct vnode *vp = ap->a_vp;
	struct hfsnode *hp = VTOH(vp);

	cache_purge(vp);
	pool_put(&hfs_node_pool, hp);
	vp->v_data = NULL;
	return 0;
}

int
hfs_bmap(v)
	void *v;
{
	struct vop_bmap_args /* {
		struct vnode *a_vp;
		daddr_t a_bn;
		struct vnode **a_vpp;
		daddr_t *a_bnp;
		int *a_runp;
	} */ *ap = v;
	struct hfsnode *hp = VTOH(ap->a_vp);
	struct hfsmount *hmp = hp->hmp;
	daddr_t *bnp;
	daddr_t bn;
	int i;

	bn = ap->a_bn;
	bnp = ap->a_bnp;
	if (ap->a_runp)
		*ap->a_runp = 0;

	if (ap->a_vpp != NULL)
		*ap->a_vpp = hp->hmp->devvp;
	if (bnp == NULL)
		return 0;
	if (bn < 0)
		return EFBIG;

	if (ap->a_vp->v_type != VREG)
		return EINVAL;

	for (i = 0; i < 3; i++) {
		if (bn / hmp->ablksize < hp->ablen[i])
			goto found;
		bn -= (daddr_t)hmp->ablksize * hp->ablen[i];
	}
	printf("extent overflow\n");
	return EINVAL;

found:
	/* XXX blocksize == 512 */
	*bnp = hmp->ablkoff + (daddr_t)hp->abnum[i] * (hmp->ablksize / 512) +
	    bn;

	return 0;
}

int
hfs_strategy(v)
	void *v;
{
	struct vop_strategy_args /* {
		struct buf *a_bp;
	} */ *ap = v;
	struct buf *bp = ap->a_bp;
	struct hfsnode *hp;
	struct vnode *vp;
	int error;

	if (bp->b_vp == NULL) {
		bp->b_flags |= B_ERROR;
		biodone(bp);
		return EIO;
	}
	vp = bp->b_vp;
	hp = VTOH(vp);
	if (bp->b_blkno == bp->b_lblkno) {
		error = VOP_BMAP(vp, bp->b_lblkno, NULL, &bp->b_blkno, NULL);
		if (error) {
			bp->b_flags |= B_ERROR;
			biodone(bp);
			return error;
		}
	}
	if (bp->b_blkno == -1) {
		biodone(bp);
		return 0;
	}
	vp = hp->hmp->devvp;
	bp->b_dev = vp->v_rdev;
	VOCALL(vp->v_op, VOFFSET(vop_strategy), ap);
	return 0;
}

/*
 * Print out the contents of a hfs vnode.
 */
int
hfs_print(v)
	void *v;
{
#if 0
	struct vop_print_args /* {
		struct vnode *a_vp;
	} */ *ap = v;
#endif

	return 0;
}

int
hfs_pathconf(v)
	void *v;
{
	struct vop_pathconf_args /* {
		struct vnode *a_vp;
		int a_name;
		register_t *a_retval;
	} */ *ap = v;

	switch (ap->a_name) {
	case _PC_LINK_MAX:
		*ap->a_retval = 1;
		return 0;
	case _PC_NAME_MAX:
		*ap->a_retval = 31;
		return 0;
	case _PC_CHOWN_RESTRICTED:
		*ap->a_retval = 1;
		return 0;
	case _PC_SYNC_IO:
		*ap->a_retval = 1;
		return 0;
	case _PC_FILESIZEBITS:
		*ap->a_retval = 32;
		return 0;
	}

	return EINVAL;
}

int
hfs_advlock(v)
	void *v;
{
#if 0
	struct vop_advlock_args /* {
		struct vnode *a_vp;
		caddr_t a_id;
		int a_op;
		struct flock *a_fl;
		int a_flags;
	} */ *ap = v;
#endif

	return EINVAL;
}

int
hfs_reallocblks(v)
	void *v;
{
#if 0
	struct vop_reallocblks_args /* {
		struct vnode *a_vp;
		struct cluster_save *a_buflist;
	} */ *ap = v;
#endif

	/* Currently no support for clustering */		/* XXX */
	return ENOSPC;
}
